/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



/** 
    Base class for objects that are to be aligned. Since all those objects are a sequence of 
    some sort, they have a length and fragments can be specified by two integer values. On request,
    they export const pointers to their member data for aligning.
    
    This class is responsible for mapping exporting only a segment of the sequence, if such is desired. 
    The alignment-algorithms always assume that an Alignandum object reaches from 1 to length. 
    
    @author Andreas Heger
    @version $Id: Alignandum.i,v 1.2 2003/08/16 09:06:44 aheger Exp $
    @short protocol class of alignable objects
*/

%module Alignandum
%{
#include "../../src/alignlib.h"

using namespace AlignLib;

#include "../../src/Alignandum.h"
#include "../../src/HelpersSequence.h"
#include "../../src/HelpersProfile.h"
%}

class Alignandum {
 public:
    /* constructors and desctructors------------------------------------------------------- */
    
    /** desctructor */
    virtual ~Alignandum();

    /** accessors ------------------------------------------------------------------------- */

    /** return an identical copy of this object */
    virtual Alignandum * getClone() const = 0;

    /** get length of window */
    virtual ALIGNANDUM_POSITION	getLength() const = 0;

    /** use full length of sequence for exporting. Basically extends the region to the
	true begining and end of object */
    virtual void useFullLength() = 0;
    
    /** use a segment for exporting and set segment to from and to 
	@param from	where segment starts
	@param to		where segment ends
    */
    virtual void useSegment( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) = 0;

    /** return first residue number in segment */
    virtual ALIGNANDUM_POSITION getFrom() const = 0;

    /** return last residue number in segment */
    virtual ALIGNANDUM_POSITION getTo() const = 0;

    /** return true if object is prepared for alignment (for cacheable types ) */
    virtual bool isPrepared() const = 0;	

    /** returns something, that can be used to access the member data. Fortunately it is now
	part of the standard, to change the return type of an inherited member function.
    */
    virtual const AlignandumData & getData() const = 0;
    
    /** get internal representation of residue in position pos */
    virtual RESIDUE asResidue( ALIGNANDUM_POSITION pos ) const = 0;

    /** get internal representation of residue in position pos */
    virtual char asChar( ALIGNANDUM_POSITION pos ) const = 0;

    /** returns a string representation of the object */
    virtual std::string asString() const = 0;
    
    /** mask column at position x */
    %name(MaskResidue) virtual void Mask ( ALIGNANDUM_POSITION x) = 0;

    /** mask column in a segment */
    virtual void Mask ( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) = 0;

    /** shuffle sequence */
    virtual void Shuffle( unsigned int num_iterations = 1,
			  ALIGNANDUM_POSITION window_size = 0) = 0;
    /* Mutators ------------------------------------------------------------------------------ */

    /** load data into cache, if cacheable type */
    virtual void Prepare() const = 0;						
    
    /** discard cache, if cacheable type */
    virtual void Release() const = 0;					       

    /** write important member data in a minimally formated way to a stream. Important
	in this sense means data that the user is interested, not internal state variables,
	that would be needed to accurately reconstitute the object.
	Use different "factory" functions to format the output in a way, that you would 
	like to have it (see writeSequenceFasta(...) for an example)
     */
    virtual void Write( std::ostream & output ) const = 0;

    /** read member data that has been output with the Write subroutine.
     */
    virtual void Read( std::istream & input ) = 0;

};

%name(makeEmptyProfile) %new Alignandum * makeProfile( const Regularizor * regularizor = NULL,
						       const LogOddor * logoddor = NULL );

%name(makeEmptyProfileLength) %new Alignandum * makeProfile( ALIGNANDUM_POSITION length,
							     const Regularizor * regularizor = NULL,
							     const LogOddor * logoddor = NULL );

/** create default profile from a NULL-terminated char-array. 
	@param sequence null terminated string to the concatenated sequences
	@param nsequences number of sequences
*/
%new Alignandum * makeProfile( const char * sequences, 
			       int nsequences,
			       const Weightor * weightor = NULL, 
			       const Regularizor * regularizor = NULL,
			       const LogOddor * logoddor = NULL );

    /** create default profile from a multiple alignment.
	@param mali multiple alignment
    */

%name(makeProfileFromMali) %new Alignandum * makeProfile( const MultipleAlignment * mali, 
				  			 const Weightor * weightor = NULL, 
							 const Regularizor * regularizor = NULL,
							 const LogOddor * logoddor = NULL );




/** create a sequence from a NULL-terminated string */
%new Alignandum * makeSequence( const char * sequence );

/** mutate a sequence according to a substitution matrix */
%new Alignandum * makeMutatedSequence( Alignandum * src, 
				       const MutationMatrix * matrix );

/** substitutes columns in profile dest by columns in profile row using the mapping provided, where dest is in col and source is in row
 */
Alignandum * substituteProfileWithProfile( Alignandum * dest, const Alignandum * source, 
					   const Alignata * map_source2dest );

/** add counts of profile source to profile dest, using the mapping provided, where dest is in col and
    source is in row 
*/
Alignandum * addProfile2Profile( Alignandum * dest, 
				 const Alignandum * source, 
				 const Alignata *);
/** add counts of profile source to profile dest, using the mapping provided, where dest is in col and
    source is in row 
*/
Alignandum * addSequence2Profile( Alignandum * dest, 
				  const Alignandum * source, 
				  const Alignata *);

/* for some reason, the shadow class contains an error, instead of saying 
   def addProfile2Profile(arg0,arg1,arg2):
       val = alignlibc.addProfile2Profile(arg0.this,arg1.this,arg2.this)
       val = AlignandumPtr(val)
       return val

   it says
   def addProfile2Profile(arg0,arg1,arg2):
       val = alignlibc.addProfile2Profile(arg0.this,arg1.this,arg2)
       val = AlignandumPtr(val)
       return val

   and therefore there is a type error during parameter checking. You have to
   correct this manually in Alignlib.py
*/

/** read counts in binary format from stream and return a profile */
%new Alignandum *  extractProfileBinaryCounts( std::istream & bin_input, 
					       const ALIGNANDUM_POSITION length = MAX_SEQLEN,
					       const Regularizor * regularizor = NULL,
					       const LogOddor * logoddor = NULL );

%new Alignandum * extractProfileBinaryCountsAsInt( std::istream & bin_input, 
						   const ALIGNANDUM_POSITION max_length = MAX_SEQLEN,
						   int bytes = 2, 
						   float scale_factor = 1,
						   const Regularizor * regularizor = NULL,
						   const LogOddor * logoddor = NULL );

/** write Profile counts to stream in compressed format */
void writeProfileBinaryCounts( std::ostream & bin_output, const Alignandum * src);

/** write Profile counts to stream in compressed format. Counts are stored columnwise as integers. */
void writeProfileBinaryCountsAsInt( std::ostream & bin_output, const Alignandum * src, 
				    int bytes = 2, float scale_factor = 1);

/** rescale counts from a profile by multiplying each entry by the scale_factor */
Alignandum * rescaleProfileCounts( Alignandum * dest, double scale_factor );

/** normalize counts from a profile so that all sum to total_weight per column*/
Alignandum * normalizeProfileCounts( Alignandum * dest, TYPE_COUNTS total_weight );

/** set random seed */
void SetRandomSeed( const long seed );

/** export frequencies of a profile. Note, that here the data
    is copied to a python array and so does not go out of scope
    if the profile is destroyed.
*/
const ProfileFrequencies * exportProfileFrequencies( Alignandum * dest );




