/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

%module Alignata
%{

#include "../../src/alignlib.h"

using namespace AlignLib;

#include "../../src/Alignata.h"
#include "../../src/HelpersAlignata.h"
#include "../../src/AlignataIterator.h"

std::string EMPTY_STRING = "";

%}

/** protocoll class for aligned objects. 
	
    An alignment is a mapping.
    
    Each alignment should provide iterators. However,
    since the implementation differs between the different
    container types, I went for passive iterators, unlike
    the STL iterators, which are active. A passive iterator
    relies on the collection to manage its advancement,
    while 

    This class is a protocol class and as such defines only the general interface
    
    @author Andreas Heger
    @version $Id: Alignata.i,v 1.1.1.1 2002/07/01 17:15:20 heger Exp $
    @short protocol class for aligned objects
*/
struct RESIDUEPAIR {
    ALIGNANDUM_POSITION mRow;
    ALIGNANDUM_POSITION mCol;
    CALCTYPE mScore;
};

class Alignata {
 public:

    //------------------> constructors / destructors <---------------------------------------------------------

    /** destructor */
    virtual ~Alignata();

    //------------------------------------------------------------------------------------------------------------
    virtual Alignata * getNew() const = 0;
    
    /** return an identical copy */
    virtual Alignata * getClone() const = 0;

    //------------------------------------------------------------------------------------------------------------
    class Iterator_const {
    public:
        Iterator_const() {};
 
        Iterator_const( const Iterator_const & src ) {};
 
        virtual ~Iterator_const() {};
	
	virtual Iterator_const * getClone() const = 0;

        /** dereference operator */
        virtual const RESIDUEPAIR & getReference() const = 0;

        /** for indirection */
        virtual const RESIDUEPAIR * getPointer() const = 0;
 
        /** advance one position */
        virtual void Next() = 0;
	    
        /** step back one position */
	virtual void Previous() = 0;
    };

    //------------------------------------------------------------------------------------------------------------
    class Iterator {
    public:
        Iterator() {};
 
        Iterator( const Iterator & src ) {};
 
        virtual ~Iterator() {};
	
	virtual Iterator * getClone() const = 0;

        /** dereference operator */
        virtual RESIDUEPAIR & getReference() const = 0;

        /** for indirection */
        virtual RESIDUEPAIR * getPointer() const = 0;
 
        /** advance one position */
        virtual void Next() = 0;
	    
        /** step back one position */
	virtual void Previous() = 0;
    };

    /** return const iterator */
    /* The const iterators do not work */
    /* virtual AlignataIterator_const begin() const = 0; */
    
    /** return const iterator */
    /* virtual AlignataIterator_const end() const = 0;  */

    /** return iterator */
    %new virtual AlignataIterator begin() = 0; 
    
    /** return iterator */
    %new virtual AlignataIterator end() = 0; 

    //----------------> accessors <------------------------------------------------------------------------------

    /** get the score of an alignment */
    virtual CALCTYPE getScore() const = 0;

    /** get the length of an alignment */
    virtual ALIGNANDUM_POSITION	getLength() const = 0 ;

    /** get the number of gaps in the alignment */
    virtual ALIGNANDUM_POSITION getNumGaps() const = 0;

    /** set the score of an alignment */
    virtual void setScore( CALCTYPE score ) = 0;

    /** returns true, if alignment is empty */
    virtual bool isEmpty() const = 0;

    /** returns the first residue aligned in row */
    virtual ALIGNANDUM_POSITION getRowFrom() const = 0;

    /** returns the last residue aligned in row */
    virtual ALIGNANDUM_POSITION	getRowTo() const = 0;

    /** returns the first residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColFrom() const = 0;

    /** returns the last residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColTo()   const = 0;

    /** returns the first aligned pair. Have to create a copy not a reference, because not all alignments will have
	a list of pairs ready */
    virtual RESIDUEPAIR getFirstPair() const = 0;
    
    /** returns the last aligned pair */
    virtual RESIDUEPAIR getLastPair() const = 0;

    /** adds a pair of residues to the alignment */
    virtual void addPair( RESIDUEPAIR * new_pair ) = 0; 

    /** adds a pair of residues to the alignment */
    %name(addPairExplicit) virtual void addPair( ALIGNANDUM_POSITION row,		
						 ALIGNANDUM_POSITION col, 
						 CALCTYPE score = 0) = 0; 

    /** removes a pair of residues from the alignment */
    virtual void removePair( const RESIDUEPAIR & old_pair ) = 0;

    /** move alignment */
    virtual void moveAlignment( ALIGNANDUM_POSITION row_offset, ALIGNANDUM_POSITION col_offset) = 0;

    /** maps a residue from row to column. returns 0, if not found */
    virtual ALIGNANDUM_POSITION mapRowToCol( ALIGNANDUM_POSITION pos ) const = 0;

    /** maps a residue from row to column. returns 0, if not found */
    virtual ALIGNANDUM_POSITION mapColToRow( ALIGNANDUM_POSITION pos ) const = 0;

    /** removes a row-region in an alignment */
    virtual void removeRowRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to ) = 0;

    /** removes a column-region in an alignment */
    virtual void removeColRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) = 0;

    /** switch row and column in the alignment */
    virtual void switchRowCol() = 0;

    /** clear the current alignemnt */
    virtual void Clear() = 0;

    /*-----------------> I/O <------------------------------------------------------------------------------ */
      /** Write data members to stream */
    virtual void Write(std::ostream & output ) const = 0;	       

    /** Read data members from stream */
    virtual void Read( std::istream & input) = 0;
    
};

/** factory functions */ 

/** factory functions */ 

/** return a pointer to an Alignata-object */
%new Alignata * makeAlignataSet(); 

/** return a pointer to an Alignata-object */
%new Alignata * makeAlignataSetCol(); 

/** return a pointer to an Alignata-object */
%new Alignata * makeAlignataHash(); 

/** return a pointer to an Alignata-object */
%new Alignata * makeAlignataHashDiagonal(); 

/** return a pointer to an Alignata-object */
%new Alignata * makeAlignataVector(); 

/** return a pointer to a new Alignata-object */
%new Alignata * makeAlignataMatrixRow();

/** return a pointer to a new Alignata-object */
%new Alignata * makeAlignataMatrixDiagonal();

/** return a pointer to a new Alignata-object */
%new Alignata * makeAlignataMatrixUnsorted();

//----------------------------------------------------------------------------------------
// typemaps for dealing with functions like writeAlignataCompressed
%typemap(python,ignore) (std::string & out_row, std::string & out_col) 
     (std::string temp_string_row, std::string temp_string_col) {
  /* Assign the pointer to a local variable */
  $1 = &temp_string_row;       
  $2 = &temp_string_col;
}

%typemap (python,argout) (std::string & out_row, std::string & out_col) { 

  PyObject * my_result_row = Py_BuildValue("s", (char *) $1->c_str());
  PyObject * my_result_col = Py_BuildValue("s", (char *) $2->c_str());
  
  if (!PyList_Check($result))
    $result = PyList_New(0); 
  
  PyList_Append($result, my_result_row); 
  PyList_Append($result, my_result_col); 

  // decrease reference counts for objects
  Py_XDECREF(my_result_row);
  Py_XDECREF(my_result_col);

}

//----------------------------------------------------------------------------------------
// typemaps for dealing with functions like writeAlignataCompressedDiagnonal
// = single string as output, needs a local variable for storage
%typemap(python,ignore) std::string & out_string(std::string temp_string) { 
  $1 = &temp_string;       
}

%typemap (python,argout) std::string & out_string {
  $result = Py_BuildValue("s", (char *) $1->c_str());
}

/** write compressed alignment into streams. The alignment can be restricted to
    a region specifying the columns. If you want to restrict it to row-residues,
    you can get those boundaries by calling mapRowToCol
    
    @param src	Alignment
    @param out_row	string in which to store the row in compressed format
    @param out_col	string in which to store the col in compressed format
    @param col_from	beginning of segment to use
    @param col_to	end of segment to use
*/
void writeAlignataCompressed( const Alignata * src, 
			      std::string & out_row, 
			      std::string & out_col,
			      ALIGNANDUM_POSITION col_from = 0,
			      ALIGNANDUM_POSITION col_to = 0
			      );


void writeAlignataCompressedDiagonal( const Alignata * src, 
				      std::string & out_string, 
				      bool reverse = false,
				      ALIGNANDUM_POSITION row_from = 0,
				      ALIGNANDUM_POSITION row_to   = 0,
				      ALIGNANDUM_POSITION col_from = 0,
				      ALIGNANDUM_POSITION col_to   = 0,
				      TYPE_DIAGONAL diagonal_from  = -MAX_DIAGONAL,
				      TYPE_DIAGONAL diagonal_to    = MAX_DIAGONAL
			      );

//----------------------------------------------------------------------------------------
  /** write an alignment in rsdb-format */
void writeAlignataRSDB( std::ostream & output, const Alignata * src );

/** return a pointer to an Alignata-object, that has been created by combining two others */
typedef enum { RR, RC, CR, CC } COMBINATION_MODE;

Alignata * combineAlignata( Alignata * dest, const Alignata *src1, const Alignata * src2, const COMBINATION_MODE mode);

/** copy one alignment into another. I can not do this using a copy constructor, since virtual functions are not
    resolved by then.

    @param row_from	beginning of segment to use
    @param row_to	end of segment to use
    @param col_from	beginning of segment to use
    @param col_to	end of segment to use
    @param diagonal_form beginning of tube to use
    @param diagonal_to end of tube to use


*/ 
Alignata * copyAlignata( Alignata * dest, 
			 const Alignata * src, 
			 ALIGNANDUM_POSITION row_from = 0,
			 ALIGNANDUM_POSITION row_to   = 0,
			 ALIGNANDUM_POSITION col_from = 0,
			 ALIGNANDUM_POSITION col_to   = 0,
			 TYPE_DIAGONAL diagonal_from  = -MAX_DIAGONAL,
			 TYPE_DIAGONAL diagonal_to    = MAX_DIAGONAL
			 );


/** copy one alignment into another. In contrast to copyAlignata, this function keeps the dots outside
     of region.
     @param row_from	beginning of segment to use
     @param row_to	end of segment to use
     @param col_from	beginning of segment to use
     @param col_to	end of segment to use
     @param diagonal_form beginning of tube to use
     @param diagonal_to end of tube to use
*/
Alignata * copyAlignataRemoveRegion( Alignata * dest, 
				     const Alignata * src, 
				     ALIGNANDUM_POSITION row_from = 0,
				     ALIGNANDUM_POSITION row_to   = 0,
				     ALIGNANDUM_POSITION col_from = 0,
				     ALIGNANDUM_POSITION col_to   = 0,
				     TYPE_DIAGONAL diagonal_from  = 1,
				     TYPE_DIAGONAL diagonal_to    = 0
				     );


/** copy one alignment into another using filter 
*/
%name(copyAlignataFilterWithAlignata) Alignata * copyAlignata( Alignata * dest, 
							       const Alignata * src, 
							       const Alignata * filter, 
							       const COMBINATION_MODE mode);

/** add one alignment to another
*/
Alignata * addAlignata2Alignata( Alignata * dest, const Alignata * src );

/** add one alignment to another. Map src using map_src2new.
*/	     
Alignata * addMappedAlignata2Alignata( Alignata * dest, 
	 			       const Alignata * src, 
				       const Alignata * map_src2new,
				       const COMBINATION_MODE mode );

/** add one alignment to another. Map both row and column.
 */
Alignata * addMappedAlignatas2Alignata( Alignata * dest, 
					const Alignata * src, 
					const Alignata * map_src_row2dest_row, 
					const Alignata * map_src_col2dest_col );
					 

/** Print aligment in table format */
void writeAlignataTable( std::ostream & output, 
			 const Alignata * src,
			 unsigned int ncols = 8,
			 bool with_scores = true);

void writePairAlignment( std::ostream & output, const Alignandum * row, const Alignandum * col, const Alignata * ali );
 
/** helper functions */
/** fill a multiple alignment given compressed strings */
Alignata * fillAlignataCompressed( Alignata *, 
				   ALIGNANDUM_POSITION row_from, 
				   const std::string & row_ali,
				   ALIGNANDUM_POSITION col_from, 
				   const std::string & col_ali
				   );
 
 
/** fill a multiple alignment given an explicit alignment */
Alignata * fillAlignataExplicit( Alignata * dest,
				 ALIGNANDUM_POSITION row_from, 
				 const std::string & row_ali,
				 ALIGNANDUM_POSITION col_from, 
				 const std::string & col_ali
				 );

Alignata * filterAlignataRemovePairs( Alignata * dest, 
                                      const Alignata * filter, 
	                              const COMBINATION_MODE mode );

/** create an identity alignment between residues from and to in row using an offset for col
 */
Alignata * fillAlignataIdentity( Alignata * dest, 
				 ALIGNANDUM_POSITION row_from, 
				 ALIGNANDUM_POSITION row_to, 
				 ALIGNANDUM_POSITION col_offset);

/** fill a multiple alignment given a compressed string in diagonal format */
Alignata * fillAlignataCompressedDiagonal( Alignata * dest, 
					   const std::string & ali,
					   bool reverse = false);
					  


Alignata * rescoreAlignment( Alignata * dest,
			     const Alignandum * row,
			     const Alignandum * col,
			     const SubstitutionMatrix * matrix = NULL);


Alignata * rescoreAlignmentPrivate( Alignata * dest,
				    const Alignandum * row,
				    const Alignandum * col,
				    const SubstitutionMatrix * matrix = NULL);

%name(rescoreAlignment_flat) Alignata * rescoreAlignment( Alignata * dest,
							  const CALCTYPE score = 0);

 /** calculate Alignment score given gap-penalties for row and column */
Alignata * calculateAffineScore( Alignata * dest, 
				 CALCTYPE gop, 
				 CALCTYPE gep );

/** fill an alignment with a repeat unit from a wrap-around alignment */
Alignata * fillAlignataRepeatUnit( Alignata * dest, 
				   const Alignata * source,
				   ALIGNANDUM_POSITION first_row_residue = 0,
				   bool skip_negative_ends = false);

/** 
    return the maps of row/col of an alignment to the summation of the alignment. This is useful for 
    building multiple alignemnts.
    
    For example: With two sequences of length 10 and alignment src between them:
    
    src = 3 4 | 4 5 | 5 7 | 9 9 

    the result would be:

    A: insert_gaps_row = false, insert_gaps_col = false, use_end_row = true, use_end_col = true
    dest1 = 1 1 | 2 2 |     |     |     | 3 6 | 4 7 |     | 5 9 | 6 10 | 7 11 | 8 12 |      | 9 14 | 10 15 |       |
    dest2 =     |     | 1 3 | 2 4 | 3 5 | 4 6 | 5 7 | 6 8 | 7 9 |      |      |      | 8 13 | 9 14 |       | 10 16 | 
    
    B: insert_gaps_row = true, insert_gaps_col = false, use_end_row = true, use_end_col = false
    dest1 = 1 1 | 2 2 | 3 3 | 4 4 | 5 5 | 6  6 | 7  7 | 8  8 | 9  9 | 10 10 |
    dest2 =     |     | 4 3 | 5 4 | 7 5 |      |      |      | 9  9 |       

    C: insert_gaps_row = true, insert_gaps_col = false, use_end_row = false, use_end_col = false
    dest1 = 3 1 | 4 2 | 5 3 | 6  4 | 7  5 | 8  6 | 9  7 | 
    dest2 = 4 1 | 5 2 | 7 3 |      |      |      | 9  7 |       

    D: insert_gaps_row = true, insert_gaps_col = true, use_end_row = false, use_end_col = false
    dest1 = 3 1 | 4 2 | 5 3 | 9  4 | 
    dest2 = 4 1 | 5 2 | 7 3 | 9  4 |       

    A can be used for building complete multiple alignments, where no part of the
    sequences are missing.
    B can be used for building multiple alignments for representatives that include
    all the neighbours.
    C can be used for showing pairwise alignments.

    Note:
 
    If there is a gap in both sequences, the residues in the first alignment dest1 are 
    aligned to the right of the gap, while the residues in the second alignment dest2
    are aligned to the right.

    @param	map_row2combined     
    @param	map_col2combined
    @param      source
    @param	insert_gaps_row
    @param	insert_gaps_col
    @param	use_end_row
    @param	use_end_col
    @param	row_length
    @param	col_length
    
 */

void fillAlignataSummation( Alignata * dest1, 
			    Alignata * dest2, 
			    const Alignata * src,
			    bool insert_gaps_row = true,
			    bool insert_gaps_col = true, 
			    bool use_end_row = false,
			    bool use_end_col = false, 
			    ALIGNANDUM_POSITION row_length = 0,
			    ALIGNANDUM_POSITION col_length = 0);




/** complement a pairwise alignment. If there is a gap of the same length in both row and
    col, the corresponding residues are added to the alignment.
*/
Alignata * complementAlignata( Alignata * dest, 
			       const ALIGNANDUM_POSITION max_length );

 /** remove all those residues from an alignmnent, which are not
     in sequence. This ensures, that col_i < col_i+1 and row < row_i+1
     Only use with AlignataVector

 */
Alignata * flattenAlignata( Alignata * dest );


/** split an alignment, if there are gaps larger than a certain threshold either in row or
     col or both.
*/
FragmentVector * splitAlignata( Alignata * src, 
				int max_gap_width,
				bool split_row = true, 
				bool split_col = true ); 


/** split an alignment at points of intersection with another alignment.
 */ 
%name(splitAlignataWithAlignata) FragmentVector * splitAlignata( const Alignata * src1, 
								 const Alignata * src2, 
								 const COMBINATION_MODE mode );
                                          
/** calculate percent similarity of alignment */
double calculatePercentSimilarity( const Alignata * src);  
    
/** calculate percent identity of alignment. Since this depends on the objects mapped on the alignment, 
    you have to supply them. */
double calculatePercentIdentity (const Alignata * src, 
				 const Alignandum * row, 
				 const Alignandum * col); 


/** remove small fragments from alignment.
    This method removes fragments from an alignment. A fragment
    is a part of an alignment, that is short (max_fragment_length)
    and surrounded by large gaps (min_gap_length).
 */
void removeFragments( Alignata * dest,
		      unsigned int window_size,
		      unsigned int min_gap_length,
		      ALIGNANDUM_POSITION row_length = 0);


/** starting from the ends of an alignment, remove 
    residues which do not contribute to a positive score.
*/
void pruneAlignata( Alignata * src,
		    const CALCTYPE gop,
		    const CALCTYPE gep);







