/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

%module Alignata
%{
#include "../../src/alignlib.h"

using namespace AlignLib;

#include "../../src/Alignata.h"
#include "../../src/HelpersAlignata.h"
#include "../../src/AlignataIterator.h"

 std::string EMPTY_STRING = "";
%}

/** protocoll class for aligned objects. 
	
    An alignment is a mapping.
    
    Each alignment should provide iterators. However,
    since the implementation differs between the different
    container types, I went for passive iterators, unlike
    the STL iterators, which are active. A passive iterator
    relies on the collection to manage its advancement,
    while 

    This class is a protocol class and as such defines only the general interface
    
    @author Andreas Heger
    @version $Id: AlignataIterator.i,v 1.1.1.1 2002/07/01 17:15:20 heger Exp $
    @short protocol class for aligned objects
*/
struct RESIDUEPAIR {
    ALIGNANDUM_POSITION mRow;
    ALIGNANDUM_POSITION mCol;
    CALCTYPE mScore;
};

class Alignata {
 public:

    //------------------> constructors / destructors <---------------------------------------------------------

    /** destructor */
    virtual ~Alignata();

    //------------------------------------------------------------------------------------------------------------
    virtual Alignata * getNew() const = 0;
    
    /** return an identical copy */
    virtual Alignata * getClone() const = 0;

    //------------------------------------------------------------------------------------------------------------
    class Iterator_const {
    public:
        Iterator_const() {};
 
        Iterator_const( const Iterator_const & src ) {};
 
        virtual ~Iterator_const() {};
	
	virtual Iterator_const * getClone() const = 0;

        /** dereference operator */
        virtual const RESIDUEPAIR & getReference() const = 0;

        /** for indirection */
        virtual const RESIDUEPAIR * getPointer() const = 0;
 
        /** advance one position */
        virtual void Next() = 0;
	    
        /** step back one position */
	virtual void Previous() = 0;
    };

    //------------------------------------------------------------------------------------------------------------
    class Iterator {
    public:
        Iterator() {};
 
        Iterator( const Iterator & src ) {};
 
        virtual ~Iterator() {};
	
	virtual Iterator * getClone() const = 0;

        /** dereference operator */
        virtual RESIDUEPAIR & getReference() const = 0;

        /** for indirection */
        virtual RESIDUEPAIR * getPointer() const = 0;
 
        /** advance one position */
        virtual void Next() = 0;
	    
        /** step back one position */
	virtual void Previous() = 0;
    };

    /** return const iterator */
    /* The const iterators do not work */
    /* virtual AlignataIterator_const begin() const = 0; */
    
    /** return const iterator */
    /* virtual AlignataIterator_const end() const = 0;  */

    /** return iterator */
    %new virtual AlignataIterator begin() = 0; 
    
    /** return iterator */
    %new virtual AlignataIterator end() = 0; 

    //----------------> accessors <------------------------------------------------------------------------------

    /** get the score of an alignment */
    virtual CALCTYPE getScore() const = 0;

    /** get the length of an alignment */
    virtual ALIGNANDUM_POSITION	getLength() const = 0 ;

    /** get the number of gaps in the alignment */
    virtual ALIGNANDUM_POSITION getNumGaps() const = 0;

    /** set the score of an alignment */
    virtual void setScore( CALCTYPE score ) = 0;

    /** get percent similarity of alignment */
    virtual double getPercentSimilarity() const = 0;  
    
    /** get percent identity of alignment. Since this depends on the objects mapped on the alignment, you
	have to supply them. */
    virtual double getPercentIdentity (const Alignandum &, const Alignandum &) const = 0; 

    /** returns true, if alignment is empty */
    virtual bool isEmpty() const = 0;

    /** returns the first residue aligned in row */
    virtual ALIGNANDUM_POSITION getRowFrom() const = 0;

    /** returns the last residue aligned in row */
    virtual ALIGNANDUM_POSITION	getRowTo() const = 0;

    /** returns the first residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColFrom() const = 0;

    /** returns the last residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColTo()   const = 0;

    /** returns the first aligned pair. Have to create a copy not a reference, because not all alignments will have
	a list of pairs ready */
    virtual RESIDUEPAIR getFirstPair() const = 0;
    
    /** returns the last aligned pair */
    virtual RESIDUEPAIR getLastPair() const = 0;

    /** adds a pair of residues to the alignment */
    virtual void addPair( RESIDUEPAIR & new_pair ) = 0; 

    /** removes a pair of residues from the alignment */
    virtual void removePair( const RESIDUEPAIR & old_pair ) = 0;

    /** move alignment */
    virtual void moveAlignment( ALIGNANDUM_POSITION row_offset, ALIGNANDUM_POSITION col_offset) = 0;

    /** maps a residue from row to column. returns 0, if not found */
    virtual ALIGNANDUM_POSITION mapRowToCol( ALIGNANDUM_POSITION pos ) const = 0;

    /** maps a residue from row to column. returns 0, if not found */
    virtual ALIGNANDUM_POSITION mapColToRow( ALIGNANDUM_POSITION pos ) const = 0;

    /** switch row and column in the alignment */
    virtual void switchRowCol() = 0;

    /** clear the current alignemnt */
    virtual void Clear() = 0;

    /*-----------------> I/O <------------------------------------------------------------------------------ */
      /** Write data members to stream */
    virtual void Write(std::ostream & output ) const = 0;	       

    /** Read data members from stream */
    virtual void Read( std::istream & input) = 0;
    
};

/** factory functions */ 

/** factory functions */ 

/** return a pointer to an Alignata-object */
%new Alignata * makeAlignataSet(); 

/** return a pointer to an Alignata-object */
%new Alignata * makeAlignataVector(); 

/** return a pointer to a new Alignata-object */
%new Alignata * makeAlignataMatrixRow();

/** return a pointer to a new Alignata-object */
%new Alignata * makeAlignataMatrixDiagonal();

/** return a pointer to a new Alignata-object */
%new Alignata * makeAlignataMatrixUnsorted();

//----------------------------------------------------------------------------------------
// typemaps for dealing with functions like writeAlignataCompressed
%typemap(python,ignore) std::string * out_row(std::string temp_string) {
      $target = &temp_string;       /* Assign the pointer to a local variable */
}
%typemap(python,ignore) std::string * out_col(std::string temp_string) {
      $target = &temp_string;       /* Assign the pointer to a local variable */
}

//----------------------------------------------------------------------------------------
// output-arguments is std::string * out_row, out_col
%typemap (python,argout) std::string * out_row {

      PyObject *o;

      o = Py_BuildValue("s", (char *) $source->c_str());
      
      if ((!$target) || ($target == Py_None)) {
	$target = o;
      } else {
	// append to return list
	if (!PyList_Check($target)) {
		PyObject *o2 = $target;
		$target = PyList_New(0);
		PyList_Append($target,o2);
		Py_XDECREF(o2);
	}
	PyList_Append($target,o);
	Py_XDECREF(o);
      }
}
%typemap (python,argout) std::string * out_col {

      PyObject *o;

      o = Py_BuildValue("s", (char *) $source->c_str());
      
      if ((!$target) || ($target == Py_None)) {
	$target = o;
      } else {
	// append to return list
	if (!PyList_Check($target)) {
		PyObject *o2 = $target;
		$target = PyList_New(0);
		PyList_Append($target,o2);
		Py_XDECREF(o2);
	}
	PyList_Append($target,o);
	Py_XDECREF(o);
      }
}

%typemap(python,freearg) std::string * out_row {
}

%typemap(python,freearg) std::string * out_col {
}


/** write compressed alignment into streams. The alignment can be restricted to
    a region specifying the columns. If you want to restrict it to row-residues,
    you can get those boundaries by calling mapRowToCol
    
    @param src	Alignment
    @param out_row	string in which to store the row in compressed format
    @param out_col	string in which to store the col in compressed format
    @param col_from	beginning of segment to use
    @param col_to	end of segment to use
*/
void writeAlignataCompressed( const Alignata * src, 
			      std::string & out_row , 
			      std::string & out_col,
			      ALIGNANDUM_POSITION col_from = 0,
			      ALIGNANDUM_POSITION col_to = 0
			      );

//----------------------------------------------------------------------------------------
  /** write an alignment in rsdb-format */
void writeAlignataRSDB( std::ostream & output, const Alignata * src );

/** return a pointer to an Alignata-object, that has been created by combining two others */
typedef enum { RR, RC, CR, CC } COMBINATION_MODE;

Alignata * combineAlignata( Alignata * dest, const Alignata *src1, const Alignata * src2, const COMBINATION_MODE mode);

/** copy one alignment into another. I can not do this using a copy constructor, since virtual functions are not
    resolved by then.
*/
Alignata * copyAlignata( Alignata * dest, const Alignata * src);

/** add one alignment to another
*/
Alignata * addAlignata2Alignata( Alignata * dest, const Alignata * src);

void writePairAlignment( std::ostream & output, const Alignandum * row, const Alignandum * col, const Alignata * ali );
 
/** helper functions */
/** fill a multiple alignment given compressed strings */
Alignata * fillAlignataCompressed( Alignata *, 
				   ALIGNANDUM_POSITION row_from, 
				   const std::string & row_ali,
				   ALIGNANDUM_POSITION col_from, 
				   const std::string & col_ali
				   );
 
 
/** fill a multiple alignment given an explicit alignment */
Alignata * fillAlignataExplicit( Alignata * dest,
				 ALIGNANDUM_POSITION row_from, 
				 const std::string & row_ali,
				 ALIGNANDUM_POSITION col_from, 
				 const std::string & col_ali
				 );

Alignata * filterAlignataRemovePairs( Alignata * dest, const Alignata * filter, const COMBINATION_MODE mode );

/** create an identity alignment between residues from and to in row using an offset for col
 */
Alignata * fillAlignataIdentity( Alignata * dest, 
				 ALIGNANDUM_POSITION row_from, 
				 ALIGNANDUM_POSITION row_to, 
				 ALIGNANDUM_POSITION col_offset);

/** fill a multiple alignment given a compressed string in diagonal format */
Alignata * fillAlignataCompressedDiagonal( Alignata * dest, 
					   const std::string & ali
					   );


Alignata * rescoreAlignment( Alignata * dest,
			     const Alignandum * row,
			     const Alignandum * col );

%name(rescoreAlignment_flat) Alignata * rescoreAlignment( Alignata * dest,
							  const CALCTYPE score = 0);
