/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

%module Alignator
%{
#include "../../src/alignlib.h"

using namespace AlignLib;

#include "../../src/Alignator.h"
#include "../../src/HelpersAlignator.h"
#include "../../src/Statistics.h"

%}

/**
   Base class for aligning two objects. The default call to use (with base class) is
   
   Alignator a;
   Alignata r;

   a( row, col, &r );

   The two objects to be aligned to each other are called row and col. This comes from
   the dynamic programming matrix where you have one sequence along the rows and the other 
   along the columns.

   This class is a protocol class and as such defines only the general interface
   
   @author Andreas Heger
   @version $Id: Alignator.i,v 1.1.1.1 2002/07/01 17:15:20 heger Exp $
   @short protocol class for aligned objects
*/ 

class Alignator {
  /* class member functions-------------------------------------------------------------- */
 public:
    /* constructors and desctructors------------------------------------------------------- */
    /** destructor */
    virtual ~Alignator ();

    /** method for aligning two arbitrary objects */
    virtual Alignata * Align(const Alignandum *, const Alignandum *, Alignata *) = 0;		

    /** method for aligning two arbitrary objects */
    virtual CALCTYPE getAlignmentScore(const Alignandum *, const Alignandum *) = 0;		

    /** accessors */
    /** set gap opening penalty for row */
    virtual void setRowGop( CALCTYPE gop ) = 0;

    /** set gap extension penalty for row */
    virtual void setRowGep( CALCTYPE gep ) = 0;

    /** set gap opening penalty for col */
    virtual void setColGop( CALCTYPE gop ) = 0;

    /** set gap extension penalty for col */
    virtual void setColGep( CALCTYPE gep ) = 0;

    /** set substitution matrix to use */
    void setSubstitutionMatrix( const SubstitutionMatrix * matrix );
    
};

/** factory functions */
%new Alignator * makeFullDP( CALCTYPE gop, CALCTYPE gep); // , const SubstitutionMatrix * matrix = NULL);

/** factory functions */
%new Alignator * makeAlignatorFullDPWrap( CALCTYPE gop, CALCTYPE gep); // , const SubstitutionMatrix * matrix = NULL);

/** Perform full dynamic programming - Needleman-Wunsch */ 
%new Alignator * makeAlignatorFullDPGlobal( CALCTYPE gop, CALCTYPE gep, 
					    bool penalize_left = true, 
					    bool penalize_right = true ); // const SubstitutionMatrix * matrix = NULL);

/** make an alignator object, which aligns identical residues */
%new Alignator * makeAlignatorIdentity();

/** make an alignator object, which aligns identical residues */
%new Alignator * makeAlignatorSimilarity(const SubstitutionMatrix * matrix = NULL);

/** make an alignator object, which aligns similar tuples */
%new Alignator * makeAlignatorTuples(int ktuple = 3, const SubstitutionMatrix * matrix = NULL);

/** make an alignator object, which returns a dummy alignments */
%new Alignator * makeAlignatorDummy( const Alignata * ali );


/** make an alignator object, which does a dot-alignment. */
%new Alignator * makeAlignatorDotsSquared(CALCTYPE gop, 
				          CALCTYPE gep, 
				          Alignator * alignator, 
				          const SubstitutionMatrix * matrix = NULL);

/** make an alignator object, which does a dot-alignment with wrapping around. */
%new Alignator * makeAlignatorDotsSquaredDiagonal(CALCTYPE gop, 
						  CALCTYPE gep, 
						  Alignator * alignator, 
						  CALCTYPE diagnal_gop = 0,
						  CALCTYPE diagonal_gep = 0,
						  const SubstitutionMatrix * matrix = NULL);

%new Alignator * makeAlignatorFragmentsSquared(CALCTYPE gop, 
					       CALCTYPE gep, 
					       Fragmentor * fragmentor, 
					       const SubstitutionMatrix * matrix  = NULL);

/** iteratively align src1 and src2 and combine alignments as long as the score is larger than min_score */
Alignata * performIterativeAlignment( Alignata * dest, 
				      const Alignandum * src_1, 
				      const Alignandum * src_2, 
				      Alignator * a, 
				      CALCTYPE min_score); 

/** iteratively align src1 and src2 and combine alignments as long as the score is larger than min_score */
%name( performIterativeAlignmentNonConst) Alignata * performIterativeAlignment( Alignata * dest, 
				          Alignandum * src_1, 
				          Alignandum * src_2, 
				          Alignator * a, 
				          CALCTYPE min_score); 

struct NormalDistributionParameters {
    double mMean;
    double mStandardDeviation;
    double getMean();
    double getStandardDeviation();
};

/** iteratively align src1 and src2 and combine alignments as long as the score is larger than min_score */
void calculateZScoreParameters( NormalDistributionParameters * result,
				const Alignandum * mutable, 
				const Alignandum * immutable, 
				Alignator * alignator,
				unsigned int n_iterations,
				unsigned int n_iterations_shuffle = 1,
				ALIGNANDUM_POSITION window_size = 0);

%new NormalDistributionParameters * makeNormalDistributionParameters() { return new NormalDistributionParameters; };
