/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

%module Alignatum
%{
#include "../../src/alignlib.h"

using namespace AlignLib;

#include "../../src/Renderer.h"
#include "../../src/Alignata.h"

#include "../../src/Alignatum.h"
#include "../../src/HelpersAlignatum.h"

%}

/** Objects of this type represent an aligned object. This class is responsible for
    producing a nice, real-world representation of the aligned object. To this end, 
    this class' provides several methods for modifying an aligned string.

    There is no method for inserting residues here. Do this, by modifying alignments
    before creating an Alignatum-object.
    
    This class is a protocol class and as such defines only the general interface
	
    @author Andreas Heger
    @version $Id: Alignatum.i,v 1.2 2003/08/16 09:06:44 aheger Exp $
    @short protocol class for aligned objects
*/

class Alignatum {

 public:
    // constructors and desctructors

    /* destructor */
    virtual ~Alignatum ();

    /*-----> accessors <----------------------------------------------------- */

    /** write object into stream nicely formatted */
    virtual void writeRow( std::ostream & output, 
			   ALIGNANDUM_POSITION segment_start = 0,
			   ALIGNANDUM_POSITION segment_end = 0,
			   const Renderer * renderer = NULL) const = 0;  

    /** return a copy of the string representation */
    virtual std::string getString() const = 0;

    /** return the string representation */
    virtual const std::string & getStringReference() const = 0;
    
    /** return the number of the first residue */
    virtual ALIGNANDUM_POSITION getFrom() const = 0;

    /** return the number of the last residue */
    virtual ALIGNANDUM_POSITION getTo() const = 0;

    /** return the length of the line */
    virtual ALIGNANDUM_POSITION getAlignedLength() const = 0;

    /** return the length of the line without gaps */
    virtual ALIGNANDUM_POSITION getTrueLength() const = 0;
    
    /** add the specified number of gaps in the front and in the back */
    virtual void addGaps(int before, int after)	= 0;		
    
    /** add one or more gaps in the middle 
	@param position where gap(s) should be inserted 
	@param count the number of gaps to be inserted
    */
    virtual void insertGaps( int position, ALIGNANDUM_POSITION count = 1) = 0;
    
    /** remove leading/or trailing gaps */
    virtual void removeEndGaps() = 0;

    /** remove one or more positions from the aligned object */
    virtual void removeColumns( int position, ALIGNANDUM_POSITION count = 1) = 0; 

    /** remap the current alignatum object using ali */
    virtual void mapOnAlignment(const Alignata * map_old2new,
				const ALIGNANDUM_POSITION new_length = 0,
				const bool unaligned_chars = false ) = 0;

    /** return a copy of the object, mapped if so desired 
	@param ali alignment to be used for the mapping. If empty, return identical copy
	@param skip_gaps true, if gaps in other sequence shall be skipped 
	@param is_in_row true, if alignment is in row
    */
    virtual Alignatum * getClone() const = 0;


    /** return an empty copy of this object */
    virtual Alignatum * getNew() const = 0;

    /** write into stream */
    virtual void  Write( std::ostream & output ) const = 0;

    /** read from stream */
    virtual void  Read( std::istream & input ) = 0;
    
};


/* -------------------------------------------------------------------------------------------------------------------- */
/* 1. factory functions */
/* note: This things are created with the new pragma, which means they are deleted, if the python handle
   goes out of scope. Since they are usually created for adding them to a multiple alignment, make sure,
   that you set thisown = 0.

   i.e.


   alignatum = makeAlignatumFromCharString("AAAAA")
   alignatum.thisown = 0
*/

%new Alignatum * makeAlignatumFromString( const std::string & src, 
					  ALIGNANDUM_POSITION from = 1, 
					  ALIGNANDUM_POSITION to = 0);
    
%name( makeAlignatumFromCharString) %new Alignatum * makeAlignatumFromString( const char * src, 
									      ALIGNANDUM_POSITION from = 1, 
									      ALIGNANDUM_POSITION to = 0);

%name( makeAlignatumFromAlignandum) %new Alignatum * makeAlignatum(const Alignandum * src, 
								   const Alignata * ali = NULL,
								   const ALIGNANDUM_POSITION max_length = 0);
    

/** return an empty Alignatum object, that can be filled for example by the function Read */
%new Alignatum * makeAlignatum();

/** return an empty Alignatum object, that 
    can store a description */
%new  Alignatum * makeAlignatumFasta();

%name( makeAlignatumFastaFromString) %new Alignatum * makeAlignatumFasta(const std::string & description,
									 const std::string & src,
									 ALIGNANDUM_POSITION from = 1, 
									 ALIGNANDUM_POSITION to = 0);











