/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


%module MultipleAlignment
%{
#include "../../src/alignlib.h"

using namespace AlignLib;

#include "../../src/Renderer.h"
#include "../../src/Alignatum.h"
#include "../../src/MultipleAlignment.h"
#include "../../src/HelpersMultipleAlignment.h"
%}

/** 
    Multiple alignments are collection of aligned sequences (more specifically: objects of type
    @ref Alignata). A multiple alignment
    receives submitted aligned objects, or unaligned objects together with a multiple
    alignment, etc. Ownership belongs to the multiple alignment.
    
    Iterators for multiple alignments: not implemented yet, will come. So far, use
    operator[].
    
    The functions for obtaining an iterator over the multiple alignments is modelled
    after containers in the STL. I did not want to restrict the implementation to 
    vectors, (or lists, etc.), because some implementations might like to have it
    in a hash. But nevertheless, it should be possible to iterate over a multiple
    alignment and therefore we need a container.
    
    (in re maps: actually, it might be nice to then to templatize operator[](T&),
    so that access via id can be possible then as well. Before doing so, check for
    code bloat. On the other hand, when interacting via iterators, the whole class
    could be a template, i.e. MultipleAlignment< vector >, or MultipleAlignment< list >
    )
    
    The multiple alignment is a sort of container. Output-formatting, etc. is done
    by the @ref Alignata-objects.
    
    The interface for this class is quite fat, because multiple alignments are used
    in a variety of contexts.

    This class is a protocol class and as such defines only the general interface

    About the parameter skip_gaps: When skip_gaps is set to true, then the current
    multiple alignment is regarded as a master multiple alignment and no gaps are 
    inserted into this alignment. When it is not set, gaps are added in the middle
    and the ends.

    @author Andreas Heger
    @version $Id: MultipleAlignment.i,v 1.3 2003/08/16 09:06:44 aheger Exp $
    @short protocol class for multiple alignments
*/

class MultipleAlignment {


    // class member functions
 public:

    /** destructor */
    virtual ~MultipleAlignment ();

    //---------------------------------------------------------------------------------------
    /*------- accessors --------------------------------------------------------------------*/
    /** returns the length of the multiple alignment. All objects in a multiple alignment have
	the same length */
    virtual ALIGNANDUM_POSITION getLength() const = 0;

    /** returns the width of the multiple alignment, i.e., the number of objects in this multiple
	alignment */
    virtual int getWidth() const = 0;

    /** sets the length of the multiple alignment. Raises an exception, if the mali is not empty
	*/
    virtual void setLength( ALIGNANDUM_POSITION length) = 0;
    
    /** returns a const reference to the object at row in the multiple alignment. This allows treating
	the multiple alignment as a two-dimensional matrix. Since string does define operator[] as well, you
	can access the symbol in column c and row r by calling 
	symbol =  multiple_alignment[r][c]
    */
    // virtual const std::string & operator[]( int row ) const = 0;		

    /** returns a pointer to the object at row from the multiple alignment. The pointer is not const,
	so you are free to do all sort of ugly stuff.*/
    virtual Alignatum * getRow( int row ) const = 0;

    /** erases an entry form the multiple alignment */
    virtual void eraseRow( int row ) = 0;
    
    /* ------------------ mutators ----------------------------------------------------------- */

    /*------------------- functions for adding new members to the multiple alignment---------*/

    /** add an aligned object to the multiple alignment. Ownership of this object is transferred to
	the multiple alignment. 
	@param src	       pointer to the aligned object to be added.
	@param alignment pointer to the alignment used for combining these two objects. If it is
		       not supplied, then it is assumed, that it is the identity alignment. In
		       that case src has to have the same length the multiple alignment. Note, the
		       multiple alignment is in col, the src is in row of the multiple alignment, so
		       when calling the member-function Alignata::Map() with a residue from
		       src, you get the correct position in the multiple alignment.
	@param skip_gaps false, if gaps in existing multiple alignment shall be introduced
    */
    %name( addAlignatum) virtual void Add( Alignatum * src,
	      const Alignata * alignment = NULL,
	      bool mali_is_in_row = true,
	      bool insert_gaps_mali = true,
	      bool insert_gaps_alignatum= true,
	      bool use_end_mali = false,
	      bool use_end_alignatum = false) = 0;

    /** add the contents of a multiple alignment to the multiple alignment by mapping it through an alignment
	@param src	       pointer to the unaligned object to be added.
	@param alignment pointer to the alignment used for combining these two objects. If it is
		       not supplied, then it is assumed, that it is the identity alignment. In
		       that case src has to have the same length the multiple alignment. Note, the
		       multiple alignment is in col, the src is in row of the multiple alignment, so
		       when calling the member-function Alignata::Map() with a residue from
		       src, you get the correct position in the multiple alignment.
          @param skip_gaps false, if gaps in existing multiple alignment shall be introduced
			   
     */
    %name( addMultipleAlignment) virtual void Add( const MultipleAlignment * src,
	      const Alignata * alignment = NULL,
	      bool mali_is_in_row = true,
	      bool insert_gaps_mali = true,
	      bool insert_gaps_alignatum= true,
	      bool use_end_mali = false,
	      bool use_end_alignatum = false) = 0;
    
    /** returns true, if there are no aligned objects in this alignment */
    virtual bool isEmpty() const = 0;

    /** clears the multiple alignment */
    virtual void Clear();

    /** register a new renderer */
    virtual void registerRenderer( const Renderer * renderer = NULL) = 0;
    
    /** returns the consensus string for the multiple alignment */
    virtual std::string getConsensusString () const = 0;			

    /** returns a clone of this object */
    virtual MultipleAlignment * getClone() const = 0;
    
    /** returns an empty version of this object */
    virtual MultipleAlignment * getNew() const = 0;
    
    /** Write the multiple alignment to a stream
     */
    virtual void Write( std::ostream & output, 
			ALIGNANDUM_POSITION segment_from = 0, 
			ALIGNANDUM_POSITION segment_to = 0) const = 0;

    /** Read a multiple alignment form a stream. I am not quite sure
	how this will work, because there might be different @ref Alignata
	objects in there.
     */
    virtual void Read( std::istream & ) =0;
};

/** factory functions */

/** return empty alignment */
%new MultipleAlignment * makeMultipleAlignment(); 

/** return empty alignment */
%new MultipleAlignment * makeMultipleAlignmentDots( bool compress_unaligned_colums = true); 

/** fill multiple alignment ali from contents of a file */
MultipleAlignment * fillMultipleAlignment( MultipleAlignment * ali, const char * filename);

/** fill multiple alignment ali from memory, sequences contains the concatenated rows without separator, however there has to be a \0 at the end of sequences
*/
%name(fillMultipleAlignmentFromString) MultipleAlignment * fillMultipleAlignment( MultipleAlignment * ali, 
										  const char * sequences, 
										  int nsequences);

/** fill multiple alignment ali from contents of a file, use an Alignatum-object for parsing */
%name(fillMultipleAlignmentFromFile) MultipleAlignment * fillMultipleAlignment( MultipleAlignment * ali, 
										const char * filename, 
										const Alignatum * alignatum_template);

/** extract a multiple Alignment object from a stream in FASTA format */
MultipleAlignment * extractMultipleAlignmentFasta( MultipleAlignment * ali, 
						   std::istream & input );

/** get Convservation-string for multiple alignment. This returns a string, where 
    each residue is marked, which is conserved at least > cutoff % */

std::string calculateConservation( const MultipleAlignment * mali, TYPE_FREQUENCIES min_frequency);



/** calculate counts in mali categorised. The first row in Matrix is empty, so that
    the numbering is consistent with the numbering of columns in the multiple alignment 
*/
%new CountsMatrix * makeCountsByCategory( const MultipleAlignment * mali, 
					  const unsigned int * map_residue2category = NULL);

/** make a map from residues to categories. The following order has been suggested by Hannes for
    surface area calculations:
    'K': 1, 'R': 1,
    'D': 2, 'E': 2,
    'H': 3, 'F': 3, 'W':3, 'Y': 3, 'C': 3,
    'N': 4, 'Q': 4, 'S':4, 'T': 4,
    'A': 5, 'I': 5, 'L': 5,'M': 5, 'V': 5,
    'G': 6, 'P': 6
*/
const unsigned int * getMapResidue2CategorySurface();

/** mapping, where each residue is mapped to its own 
    category 
*/
const unsigned int * getMapResidue2CategoryAll();

/** return a vector of entropies calculated for a CountsMatrix
 */
%new VectorDouble * makeEntropyVector( const CountsMatrix * src);

/** copy a multiple alignment
 */
MultipleAlignment * copyMultipleAlignment( MultipleAlignment * dest, 
					   const MultipleAlignment * src,
					   unsigned int first_row = 0,
					   unsigned int last_row = 0 );



