/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// note: SWIG does not support overloaded functions. Therefore I have problems
// with the factory functions like makeProfile and some methods like Mask. 
// The solution: rename these functions

// note: streams are converted to strings. If there is an ostream in the parameter list,
// a strstream is created and the result is returned as a string. For example, the output-routines
// like alignandum->Write( std::ostream & ) are called result = alignandum.Write(), print result

// for this conversion to work, there are some naming conventions:
// output streams are named output for string-streams and bin_output for binary streams.

// binary streams are more difficult, since \0 is interpreted as end of string. binary streams are
// therefore converted to/from file objects. 

// #defines have to be remapped into global variables, e.g. MAX_SEQLEN. Unfortunately, they can not
// be declared const.

// factory functions should be prefaced with the %new directive, so the ownership-variable
// self.thisown is set properly

/* It seems, that you have to name parameters, i.e. 
   func( int * = NULL) does not work, but func (int * ptr = NULL) does. 
 */

/* I had some problems with default values like f(pos = 0). I used a work-around, where I allocated
   global variables that where supplied as default values:
   int pos_null = 0;
   f( pos = pos_null);

   However, what SWIG does is it then takes the default value correctly, but since it is
   pointer based, it tries to free that memory from the global variable after the function
   is executed. Therefore test in the freearg part of the typemaps, if one of the globals
   has been passed (see ALIGNANDUM_POSITION as an example).

   -> return values
   Using the return-values does not work, to. Therefore, do not use constructs like:

   alignment = fillAlignataIdentity( makeAlignataSet(), 1, 1, 0)
   but instead:

   alignment = makeAlignataSet()
   fillAlignataIdentity( alignment, 1, 1, 0)

 */

/* I had a difficult to detect memory leak, because I used the wrong order of %name and %new.
   The correct syntax is:

   %name(makeProfileFromMali) %new Alignandum * makeProfile( const MultipleAlignment * mali, ...)
  
   ( and not:
   %new %name(makeProfileFromMali) Alignandum * makeProfile( const MultipleAlignment * mali, ...)
   )

   In the shadow-class val.thisown was not set
*/


%module alignlib
%{
#include <strstream>
#include <fstream>
    
#include "../../src/alignlib.h"
#include "../../src/Matrix.h"
  // for FragmentVector, clean this up later
#include "../../src/HelpersAlignata.h" 

using namespace AlignLib;

%}

typedef unsigned char	TYPESTRING;
typedef double CALCTYPE;
typedef long TYPE_DIAGONAL;	
typedef long TYPE_DOT;	
typedef long TYPE_FRAGMENT;	
typedef int ALIGNANDUM_POSITION;	
typedef int ALIGNATUM_POSITION;	
typedef double TYPE_SEQUENCE_WEIGHTS;
typedef	unsigned char RESIDUE;
typedef double TYPE_PROFILE; 
typedef double TYPE_COUNTS;
typedef double TYPE_FREQUENCIES;

//-------------->std::string<-------------------------------------------------------
%typemap (python,in) std::string & (std::string temp_string) {
  if (!PyString_Check($input)) {
	PyErr_SetString( PyExc_TypeError, "Not a valid string" );
	return NULL;
  }
  temp_string = PyString_AsString( $input );
  $1 = &temp_string;
}

%typemap (python,out) std::string {
  $result = Py_BuildValue( "s", (char *) $1.c_str());
}

//-------------->std::ostream<------------------------------------------------------
// std::ostream is omitted as parameter for functions, but is converted as an
// output string.
//----------------------------------------------------------------------------------
%typemap (python,ignore) std::ostream & output (std::ostrstream temp_stream) {
  // Assign the pointer to a local variable (so that the memory is there)
  $1 = &temp_stream;       
}

%typemap (python,argout) std::ostream & output {

      PyObject *o;

      std::ostrstream * s = (std::ostrstream*)$1;
      *s << std::ends;
	
      o = Py_BuildValue("s", (char *)s->str());
      s->freeze(false);

      if ((!$result) || ($result == Py_None)) {
	$result = o;
      } else {
	// append to return list
	if (!PyList_Check($result)) {
		PyObject *o2 = $result;
		$result = PyList_New(0);
		PyList_Append($result,o2);
		Py_XDECREF(o2);
	}
	PyList_Append($result,o);
	Py_XDECREF(o);
      }
}
//-------------->std::istream<------------------------------------------------------
// std::istream is omitted as parameter for functions, but is converted as an
// output string.
//----------------------------------------------------------------------------------
%typemap (python,in) std::istream & input {

    if (!PyString_Check($input)) {
	PyErr_SetString( PyExc_TypeError, "Not a valid string" );
	return NULL;
    }
    
    int length = PyString_Size( $input );

    $1 = new std::istrstream( PyString_AsString( $input ), length);
}

%typemap(python,freearg) std::istream & input {
  delete $1;
}

//-------------->std::istream<------------------------------------------------------
// binary input and output streams are converted from strings into files.
%typemap (python,in) std::istream & bin_input {

    if (!PyString_Check($input)) {
	PyErr_SetString( PyExc_TypeError, "Not a valid string" );
	return NULL;
    }
    
    $1 = new std::ifstream( PyString_AsString( $input ));
}

%typemap(python,freearg) std::istream * bin_input{
  delete $1;
}

%typemap (python,in) std::ostream & bin_output {

  if (!PyString_Check($input)) {
    PyErr_SetString( PyExc_TypeError, "Not a valid string" );
    return NULL;
  }
    
  $1 = new std::ofstream( PyString_AsString($input) );

}

%typemap(python,freearg) std::ostream & bin_output {
    delete $1;
}

//-----------------> TYPE_PROFILE_COLUMN * <--------------------------------------------
%typemap(python,out) ProfileFrequencies * {
  unsigned int num_rows, num_cols, col, row;
  num_rows  = PROFILEWIDTH;
  num_cols  = $1->size();
  $result = PyTuple_New(num_cols);
  for (col = 0; col <= num_cols; col++) {
    PyObject * c = PyTuple_New(num_rows);
    for (row = 0; row < num_rows; row++) {
      PyTuple_SetItem(c, row, Py_BuildValue( "d", (double)(*$1)[col][row]));
    }
    PyTuple_SetItem($result,col, c);
  }	
  
  delete $1;
}  

//-----------------> CountsMatrix * <--------------------------------------------
%typemap(python,out) CountsMatrix * {
  unsigned int num_rows, num_cols, i, x;
  num_rows = $1->getNumRows();
  num_cols = $1->getNumCols();
  
  $result = PyList_New(num_rows);
  for (i = 0; i < num_rows; i++) {
    
    PyObject * c = PyList_New(num_cols);
    for (x = 0; x < num_cols; x++) {
      PyList_SetItem(c, x, Py_BuildValue( "i", (unsigned int)(*$1)[i][x]));
    }
    PyList_SetItem($result,i, c);
  }	
  
  delete $1;
}

%typemap(python,in) CountsMatrix * {
  /* Check if is a list */
  if (PyList_Check($input)) {
    unsigned int num_rows = PyList_Size($input);
    
    // retrieve first line
    PyObject *o = PyList_GetItem($input,0);
    unsigned int num_cols = PyList_Size(o);
    
    $1 = new CountsMatrix( num_rows, num_cols );
    
    int i = 0; int x= 0;
    for (i = 0; i < num_rows; i++) {
      PyObject *o = PyList_GetItem($input,i);
      for (x =0; x < num_cols; x++) {
	PyObject * value = PyList_GetItem(o,x);
	(*$1)[i][x] = PyInt_AsLong(value);
      }
    }
  } else {
    PyErr_SetString(PyExc_TypeError,"not a list");
    return NULL;
  }
}


// This cleans up the char ** array we malloc'd before the function call
%typemap(python,freearg) CountsMatrix * {
    delete $1;
}

//-----------------> MutationMatrix * <--------------------------------------------
%typemap(python,out) MutationMatrix * {
  unsigned int num_rows, num_cols, i, x;
  num_rows  = $1->getNumRows();
  num_cols  = $1->getNumCols();
  
  $result = PyList_New(num_rows);
  for (i = 0; i < num_rows; i++) {
    
    PyObject * c = PyList_New(num_cols);
    for (x = 0; x < num_cols; x++) {
      PyList_SetItem(c, x, Py_BuildValue( "d", (double)(*$1)[i][x]));
    }
    PyList_SetItem($result,i, c);
  }	
  
  delete $1;
}

%typemap(python,in) MutationMatrix * {
  /* Check if is a list */
  if (PyList_Check($input)) {
    unsigned int num_rows = PyList_Size($input);
    
    // retrieve first line
    PyObject *o = PyList_GetItem($input,0);
    unsigned int num_cols = PyList_Size(o);
    if (num_cols != num_rows) {
      PyErr_SetString(PyExc_TypeError,"has to be a symmetric matrix");
      return NULL;
    }
    
    $1 = new MutationMatrix( num_rows, num_rows );
    
    int i = 0; int x= 0;
    for (i = 0; i < num_rows; i++) {
      PyObject *o = PyList_GetItem($input,i);
      for (x =0; x < num_cols; x++) {
	PyObject * value = PyList_GetItem(o,x);
	(*$1)[i][x] = PyFloat_AsDouble(value);
      }
    }
  } else {
    PyErr_SetString(PyExc_TypeError,"not a list");
	return NULL;
  }
}

// This cleans up the char ** array we malloc'd before the function call
%typemap(python,freearg) MutationMatrix * {
    delete $1;
}

//----------------> FragmentVector * <-----------------------------------
/* Convert a list of fragments to a list of tuples.
   note: you have to convert the mangled strings into
   shadow class objects. I don't know how to do this yet
   except explicitely in python:

   val = fragmentor.Fragment( a, b, sample )
   fragments = map( lambda x: AlignataPtr(x), val)
   for fragment in fragments:
	fragment.thisown = 1
*/ 
%typemap(python,out) FragmentVector * {
  
  unsigned int len = $1->size();
  unsigned int i;

  $result = PyList_New(len);

  for (i = 0; i < len; i++) {
    PyObject * my_resultobj = SWIG_NewPointerObj( (void *)(*$1)[i], SWIGTYPE_p_Alignata, 1);
    PyList_SetItem($result,i, my_resultobj);
    // reduce reference count by one, as it is now contained in the list.
    // Py_XDECREF( my_resultobj );
  }

  /* delete the container (but not the content, which is now in a python list) */
  delete $1;
}

      
//----------------> std::vector<double> * <-----------------------------------
%typemap(python,out) VectorDouble * {
    unsigned int length = $1->size();
    $result = PyList_New(length);
    for (unsigned int i = 0; i < length; i++) 
	PyList_SetItem($result,i,Py_BuildValue( "d", (double)(*$1)[i]));
	
    delete $1;
}


%include Alignata.i
%include Alignator.i 
%include Alignandum.i 
%include Alignatum.i
%include Regularizor.i
%include Weightor.i 
%include LogOddor.i
%include SubstitutionMatrix.i
%include MultipleAlignment.i 
%include Renderer.i
%include Translator.i
%include Fragmentor.i 

//%include CountsMatrix.i



