/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef ALIGNATA_H
#define ALIGNATA_H 1

#include <iosfwd>
#include <string>
#include "alignlib.h"

namespace AlignLib {

    /** protocoll class for aligned objects. 
	
	An alignment is a mapping.

	Each alignment should provide iterators. However,
	since the implementation differs between the different
	container types, I went for passive iterators, unlike
	the STL iterators, which are active. A passive iterator
	relies on the collection to manage its advancement and 
	therefore has to cooperate closely with its container.
	
	This class is a protocol class and as such defines only the general interface

	@author Andreas Heger
	@version $Id: Alignata.h,v 1.1.1.1 2002/07/01 17:15:20 heger Exp $
	@short protocol class for aligned objects
    */

class Alignandum;
class AlignataIterator;
class AlignataIterator_const;
    
struct RESIDUEPAIR {
    friend std::ostream & operator<< (std::ostream &, const RESIDUEPAIR &);

  RESIDUEPAIR() : mRow(0), mCol(0), mScore(0) {}
  RESIDUEPAIR( ALIGNANDUM_POSITION a, ALIGNANDUM_POSITION b, CALCTYPE c) : mRow(a), mCol(b), mScore(c) {}
  RESIDUEPAIR& operator=( const RESIDUEPAIR & src) { mRow = src.mRow; mCol = src.mCol; mScore = src.mScore; return *this;}
  RESIDUEPAIR( const RESIDUEPAIR & src) : mRow(src.mRow), mCol(src.mCol), mScore(src.mScore) {}  
  ALIGNANDUM_POSITION mRow;
  ALIGNANDUM_POSITION mCol;
  CALCTYPE mScore;
};

bool operator==( const RESIDUEPAIR & x, const RESIDUEPAIR & y); 
bool operator!=( const RESIDUEPAIR & x, const RESIDUEPAIR & y); 

class Alignata {
    friend std::ostream & operator<<(std::ostream &output, const Alignata &);
    friend std::istream & operator>>(std::istream &input, Alignata &);

 public:

    //------------------> constructors / destructors <---------------------------------------------------------
    /** empty constructor */
    Alignata();

    /** copy constructor */
    Alignata( const Alignata &src );

    /** destructor */
    virtual ~Alignata();

    //------------------------------------------------------------------------------------------------------------
    virtual Alignata * getNew() const = 0;
    
    /** return an identical copy */
    virtual Alignata * getClone() const = 0;

    //------------------------------------------------------------------------------------------------------------
    class Iterator_const {
    public:
        Iterator_const() {};
 
        Iterator_const( const Iterator_const & src ) {};
 
        virtual ~Iterator_const() {};
	
	virtual Iterator_const * getClone() const = 0;

        /** dereference operator */
        virtual const RESIDUEPAIR & getReference() const = 0;

        /** for indirection */
        virtual const RESIDUEPAIR * getPointer() const = 0;
 
        /** advance one position */
        virtual void Next() = 0;
	    
        /** step back one position */
	virtual void Previous() = 0;
    };

    //------------------------------------------------------------------------------------------------------------
    class Iterator {
    public:
        Iterator() {};
 
        Iterator( const Iterator & src ) {};
 
        virtual ~Iterator() {};
	
	virtual Iterator * getClone() const = 0;

        /** dereference operator */
        virtual RESIDUEPAIR & getReference() const = 0;

        /** for indirection */
        virtual RESIDUEPAIR * getPointer() const = 0;
 
        /** advance one position */
        virtual void Next() = 0;
	    
        /** step back one position */
	virtual void Previous() = 0;
    };

    /** return const iterator */
    virtual AlignataIterator_const begin() const = 0; 
    
    /** return const iterator */
    virtual AlignataIterator_const end() const = 0; 

    /** return iterator */
    virtual AlignataIterator begin() = 0; 
    
    /** return iterator */
    virtual AlignataIterator end() = 0; 

    //----------------> accessors <------------------------------------------------------------------------------

    /** get the score of an alignment */
    virtual CALCTYPE getScore() const = 0;

    /** get the length of an alignment */
    virtual ALIGNANDUM_POSITION	getLength() const = 0 ;

    /** get the number of gaps in the alignment */
    virtual ALIGNANDUM_POSITION getNumGaps() const = 0;

    /** set the score of an alignment */
    virtual void setScore( CALCTYPE score ) = 0;

    /** returns true, if alignment is empty */
    virtual bool isEmpty() const = 0;

    /** returns the first residue aligned in row */
    virtual ALIGNANDUM_POSITION getRowFrom() const = 0;

    /** returns the last residue aligned in row */
    virtual ALIGNANDUM_POSITION	getRowTo() const = 0;

    /** returns the first residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColFrom() const = 0;

    /** returns the last residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColTo()   const = 0;

    /** returns the first aligned pair. Have to create a copy not a reference, because not all alignments will have
	a list of pairs ready */
    virtual RESIDUEPAIR getFirstPair() const = 0;
    
    /** returns the last aligned pair */
    virtual RESIDUEPAIR getLastPair() const = 0;

    /** adds a pair of residues to the alignment */
    virtual void addPair( RESIDUEPAIR * new_pair ) = 0; 

    /** adds a pair of residues to the alignment */
    virtual void addPair( ALIGNANDUM_POSITION row, ALIGNANDUM_POSITION col, CALCTYPE score = 0) = 0; 

    /** removes a pair of residues from the alignment */
    virtual void removePair( const RESIDUEPAIR & old_pair ) = 0;

    /** retrieves a pair of residues from the alignment */
    virtual RESIDUEPAIR getPair( ALIGNANDUM_POSITION row ) const = 0;

    /** move alignment */
    virtual void moveAlignment( ALIGNANDUM_POSITION row_offset, ALIGNANDUM_POSITION col_offset) = 0;

    /** maps a residue from row to column. returns 0, if not found */
    virtual ALIGNANDUM_POSITION mapRowToCol( ALIGNANDUM_POSITION pos ) const = 0;

    /** maps a residue from row to column. returns 0, if not found */
    virtual ALIGNANDUM_POSITION mapColToRow( ALIGNANDUM_POSITION pos ) const = 0;

    /** removes a row-region in an alignment */
    virtual void removeRowRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to ) = 0;

    /** removes a column-region in an alignment */
    virtual void removeColRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) = 0;

    /** switch row and column in the alignment */
    virtual void switchRowCol() = 0;

    /** clear the current alignemnt */
    virtual void Clear() = 0;

    /*-----------------> I/O <------------------------------------------------------------------------------ */
      /** Write data members to stream */
    virtual void Write(std::ostream & output ) const = 0;	       

    /** Read data members from stream */
    virtual void Read( std::istream & input) = 0;
    
};

}
#endif /* _ALIGNATA_H */

