/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef ALIGNATA_ITERATOR_H
#define ALIGNATA_ITERATOR_H 1

#include <iosfwd>
#include "alignlib.h"
#include "Alignata.h"

namespace AlignLib {

    /** This class takes any of the iterators
	of aligned pairs of alignata-objects
	and converts it into an iterator, that
	is STL-compatible in its syntax (i.e. 
	no overloading, etc.). The penalty for
	this universality in syntax is paid in
	performance. There are two additional
	indirections.

	@author Andreas Heger
	@version $Id: AlignataIterator.h,v 1.1.1.1 2002/07/01 17:15:20 heger Exp $
	@short protocol class for aligned objects
    */

class AlignataIterator_const {
 public:

    AlignataIterator_const() : mIterator(NULL) {};

    AlignataIterator_const( Alignata::Iterator_const * it) : mIterator(it) {} ;
    
    AlignataIterator_const( const AlignataIterator_const & src ) { mIterator = src.mIterator->getClone();};
    
    ~AlignataIterator_const() { delete mIterator; };

    /** assignment operator */
    inline AlignataIterator_const & operator=( const AlignataIterator_const & other ) {
      if (*this != other) {
	/* mIterator = other.mIterator; does not work, I guess, because of shallow copying */
	delete mIterator;
	mIterator = other.mIterator->getClone();
      }
      
      return *this;
    }

    /** comparison operator */
    inline bool operator==( const AlignataIterator_const & other) {   
      return ( (*mIterator).getPointer() == (*other.mIterator).getPointer() );
    }
    
    /** comparison operator */
    inline bool operator!=( const AlignataIterator_const & other) {
      return ( (*mIterator).getPointer() != (*other.mIterator).getPointer() );
    }
      
    /** indirection operator */
    inline const RESIDUEPAIR * operator->() const { return &(mIterator->getReference()); }

    /** dereference operator */
    inline const RESIDUEPAIR & operator*() const { return mIterator->getReference(); }
 
    /** prefix ++ */
    inline AlignataIterator_const & operator++() { mIterator->Next(); return *this; }
 
    /** postfix ++ */
    inline AlignataIterator_const operator++(int) { AlignataIterator_const tmp = *this; mIterator->Next(); return tmp; }
 
    /** prefix -- */
    inline AlignataIterator_const & operator--() { mIterator->Previous(); return *this; }
 
    /** postfix -- */
    inline AlignataIterator_const  operator--(int) { 
      AlignataIterator_const tmp = *this; mIterator->Previous(); return tmp; }
    
 private:
    Alignata::Iterator_const * mIterator;
};

class AlignataIterator {
 public:
    AlignataIterator() : mIterator(NULL) {};

    AlignataIterator( Alignata::Iterator * it) : mIterator(it) {} ;
    
    AlignataIterator( const AlignataIterator & src ) { mIterator = src.mIterator->getClone();};
    
    ~AlignataIterator() { delete mIterator; };

    /** assignment operator */
    inline AlignataIterator & operator=( const AlignataIterator & other ) {
      if (*this != other) {
	/* mIterator = other.mIterator; does not work, I guess, because of shallow copying */
	delete mIterator;
	mIterator = other.mIterator->getClone();
      }
      
      return *this;
    }

    /** comparison operator */
    inline bool operator==( const AlignataIterator & other) {   
      return ( (*mIterator).getPointer() == (*other.mIterator).getPointer() );
    }
    
    /** comparison operator */
    inline bool operator!=( const AlignataIterator & other) {
      return ( (*mIterator).getPointer() != (*other.mIterator).getPointer() );
    }
      
    /** indirection operator */
    inline RESIDUEPAIR * operator->() const { return &(mIterator->getReference()); }

    /** dereference operator */
    inline RESIDUEPAIR & operator*() const { return mIterator->getReference(); }
 
    /** prefix ++ */
    inline AlignataIterator & operator++() { mIterator->Next(); return *this; }
 
    /** postfix ++ */
    inline AlignataIterator operator++(int) { AlignataIterator tmp = *this; mIterator->Next(); return tmp; }
 
    /** prefix -- */
    inline AlignataIterator & operator--() { mIterator->Previous(); return *this; }
 
    /** postfix -- */
    inline AlignataIterator  operator--(int) { 
      AlignataIterator tmp = *this; mIterator->Previous(); return tmp; }
    
 private:
    Alignata::Iterator * mIterator;
};



}

#endif /* _ALIGNATA_H */

