/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef ALIGNATUM_H
#define ALIGNATUM_H 1

#include <iosfwd>
#include <string>
#include "alignlib.h"

namespace AlignLib {

/** Objects of this type represent an aligned object. This class is responsible for
    producing a nice, real-world representation of the aligned object. To this end, 
    this class' provides several methods for modifying an aligned string.

    There is no method for inserting residues here. Do this, by modifying alignments
    before creating an Alignatum-object.
    
    This class is a protocol class and as such defines only the general interface

    @author Andreas Heger
    @version $Id: Alignatum.h,v 1.3 2003/08/16 09:09:25 aheger Exp $
    @short protocol class for aligned objects
*/

class Alignandum;
class Alignata;
class Renderer;

class Alignatum {
    // class member functions
    friend std::ostream & operator<<( std::ostream &, const Alignatum &);
    friend std::istream & operator>>( std::istream &, Alignatum &);
 public:
    // constructors and desctructors
    /** when building a new Alignatum sequence by cloning and an alignment, this specifies,
	if the new sequence is in row or column
    */

    /** constructor */
    Alignatum  ();

    /** copy constructor */
    Alignatum  (const Alignatum &);

    /* destructor */
    virtual ~Alignatum ();

    /*-----> accessors <----------------------------------------------------- */

    /** write object into stream nicely formatted */
    virtual void writeRow( std::ostream & output, 
			   ALIGNANDUM_POSITION segment_start = 0, 
			   ALIGNANDUM_POSITION segment_end = 0,
			   const Renderer * = NULL)  const = 0;

    /** return a copy of the string representation */
    virtual std::string getString() const = 0;

    /** return the string representation */
    virtual const std::string & getStringReference() const = 0;
    
    /** return the number of the first residue */
    virtual ALIGNANDUM_POSITION getFrom() const = 0;

    /** return the number of the last residue */
    virtual ALIGNANDUM_POSITION getTo() const = 0;

    /** return the length of the line */
    virtual ALIGNANDUM_POSITION getAlignedLength() const = 0;

    /** return the length of the line without gaps */
    virtual ALIGNANDUM_POSITION getTrueLength() const = 0;

    /** add the specified number of gaps in the front and in the back */
    virtual void addGaps(int before, int after)	= 0;		
    
    /** add one or more gaps in the middle 
	@param position where gap(s) should be inserted 
	@param count the number of gaps to be inserted
    */
    virtual void insertGaps( int position, ALIGNANDUM_POSITION count = 1) = 0;
    
    /** remove leading/or trailing gaps */
    virtual void removeEndGaps() = 0;

    /** remove one or more positions from the aligned object */
    virtual void removeColumns( int position, ALIGNANDUM_POSITION count = 1) = 0; 

    /** remap the current alignatum object using ali 
	If unaligned chars is true, lower case unaligned characters will be
	put before the next aligned character (as much as fit) 
    */
    virtual void mapOnAlignment(const Alignata * map_old2new,
				const ALIGNANDUM_POSITION new_length = 0,
				const bool unaligned_chars = false ) = 0;
				


    /**
	@param skip_gaps true, if gaps in other sequence shall be skipped 
	@param is_in_row true, if alignment is in row
    */
    virtual Alignatum * getClone() const = 0;

    /** return an empty copy of this object */
    virtual Alignatum * getNew() const = 0;

    /** write into stream */
    virtual void  Write( std::ostream & output ) const = 0;

    /** read from stream */
    virtual void  Read( std::istream & input ) = 0;
    
};


}

#endif /* ALIGNATUM_H */

