/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <iostream>
#include <string>
#include <stdio.h>
#include <strstream>
#include <iomanip>
#include "alignlib.h"
#include "Alignandum.h"
#include "AlignataIterator.h"
#include "Alignata.h"
#include "AlignException.h"

#include "HelpersAlignata.h"
#include "Translator.h"
#include "HelpersTranslator.h"

#include "Alignatum.h"
#include "HelpersAlignatum.h"

#include "SubstitutionMatrix.h"
#include "HelpersSubstitutionMatrix.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

//---------------------------------------------------------------------------------------------------
inline TYPE_DIAGONAL calculateDiagonal( const RESIDUEPAIR & p) { return (p.mCol - p.mRow); }

//--------------------------------------------------------------------------------------------------
/** write an alignment in compressed form into two streams*/
void writeAlignataCompressed( const Alignata * src, 
			      std::string & out_row, 
			      std::string & out_col,
			      ALIGNANDUM_POSITION col_from,
			      ALIGNANDUM_POSITION col_to) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void writeAlignataCompressed( const Alignata *, std::string&, std::string &, ALIGNANDUM_POSITION, ALIGNANDUM_POSITION)" << endl;
#endif

  out_row = "";
  out_col = "";
  
  // sanity checks
  if (src->isEmpty()) 
      return;

  if (col_from < src->getColFrom() || 0 == col_from)
      col_from = src->getColFrom();
  if (col_to > src->getColTo() || 0 == col_to)
      col_to = src->getColTo();

  AlignataIterator_const it(src->begin());
  AlignataIterator_const it_end(src->end());

  std::ostrstream os_row;
  std::ostrstream os_col;

  ALIGNANDUM_POSITION last_col = it->mCol; ALIGNANDUM_POSITION len_col = 0;
  while (last_col < col_from) {
    ++it;
    last_col = it->mCol;
  }
  ALIGNANDUM_POSITION last_row = it->mRow; ALIGNANDUM_POSITION len_row = 0;

  ALIGNANDUM_POSITION d_row, d_col;
  ALIGNANDUM_POSITION current_row = 0, current_col = 0;
    
  // start iteration at col_from + 1
  ++it; len_row++; len_col++;
  for (; it != it_end; ++it) {

    current_row = it->mRow;
    current_col = it->mCol;

#ifdef DEBUG
    if (verbose > LL4) cout << "current_row: " << current_row << " " << "current_col: " << current_col << endl;
#endif

    if (current_col > col_to)
      break;
    
    if ((d_row = current_row - last_row - 1) > 0) {
      os_col << "+" << len_col << "-" << d_row;
      
      len_col = 0;
      len_row += d_row;
    }
    
    if ((d_col = current_col - last_col - 1) > 0) {
      os_row << "+" << len_row << "-" << d_col;
      len_row = 0;
      len_col += d_col;
    }
    
    last_row = current_row;
    last_col = current_col;
    len_row++; len_col++;
  }


  os_col << "+" << len_col << ends;
  os_row << "+" << len_row << ends;

  
  out_row = os_row.str(); 
  out_col = os_col.str(); 

  // unfreeze streams, so the memory gets freed
  os_row.freeze(false);
  os_col.freeze(false);
}       

//---------------------------------------------------------------------------------------------------------
/** write an alignment in rsdb-format */
void writeAlignataRSDB( std::ostream & output, const Alignata * src ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void writeAlignataRSDB( std::ostream &, const Alignata *)" << endl;
#endif

#define MY_SEPARATOR "\t"
  if ( src->getLength() < 1) {
    output << 0;                            // no alignment performed
  } else {
    output << src->getLength()  << MY_SEPARATOR << src->getScore() << MY_SEPARATOR;
    output << src->getRowFrom() << MY_SEPARATOR << src->getRowTo() << MY_SEPARATOR;
    output << src->getColFrom() << MY_SEPARATOR << src->getColTo() << MY_SEPARATOR;
    
    std::string s_row;
    std::string s_col;
    
    writeAlignataCompressed( src, s_row, s_col );
    
    output << s_row << MY_SEPARATOR;
    output << s_col;
  }
#undef MY_SEPARATOR
}

//---------------------------------------------------------------------------------------------------------
/** write an alignment in rsdb-format */
Alignata * readAlignataPairs( Alignata * src, std::istream & input, bool reverse ) {
#ifdef DEBUG 
  if (verbose > LL3) cout << "void readAlignataPairs( Alignata *, std::istream &)" << endl;
#endif

  src->Clear();

  while (input) {
    ALIGNANDUM_POSITION row, col;
    CALCTYPE score;
    
    input >> row >> col >> score;
    if (reverse)
      src->addPair( new RESIDUEPAIR(col, row, score) );
    else
      src->addPair( new RESIDUEPAIR(row, col, score) );
  }

  return src;
}


//-----------------------------------------------------------------------------------------------
/** convert one type of Alignata-object into another using functions from the public interface */
Alignata * copyAlignata( Alignata * dest, 
			 const Alignata * src,
			 ALIGNANDUM_POSITION row_from,
			 ALIGNANDUM_POSITION row_to,
			 ALIGNANDUM_POSITION col_from,
			 ALIGNANDUM_POSITION col_to,
			 TYPE_DIAGONAL diagonal_from,
			 TYPE_DIAGONAL diagonal_to ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "Alignata * copyAlignata( Alignata *, const Alignata *) " << endl;
#endif

  // check parameters for filters and set them to sensible values
  if (col_from < src->getColFrom() || col_from == 0)
      col_from = src->getColFrom();
  if (col_to > src->getColTo() || col_to == 0)
      col_to = src->getColTo();
  if (row_from < src->getRowFrom() || row_from == 0)
      row_from = src->getRowFrom();
  if (row_to > src->getRowTo() || row_to == 0)
      row_to = src->getRowTo();

  if (diagonal_from > diagonal_to) {
      diagonal_from = -MAX_DIAGONAL;
      diagonal_to   =  MAX_DIAGONAL;
  }

  dest->Clear();
  
  AlignataIterator_const it(src->begin());
  AlignataIterator_const it_end(src->end());

  for (; it != it_end; ++it) {
    const RESIDUEPAIR & p = *it;
    
    // apply filter
    TYPE_DIAGONAL	this_diagonal = calculateDiagonal(p);
    ALIGNANDUM_POSITION this_row      = p.mRow;
    ALIGNANDUM_POSITION this_col      = p.mCol;
    
    if (this_col < col_from || this_col > col_to) 
	continue;
    if (this_row < row_from || this_row > row_to) 
	continue;
    if (this_diagonal < diagonal_from || this_diagonal > diagonal_to)
	continue;

    dest->addPair( new RESIDUEPAIR(p) );

  }
  
  return dest;
}
//-----------------------------------------------------------------------------------------------
/** convert one type of Alignata-object into another using functions from the public interface */
Alignata * copyAlignataRemoveRegion( Alignata * dest, 
				     const Alignata * src,
				     ALIGNANDUM_POSITION row_from,
				     ALIGNANDUM_POSITION row_to,
				     ALIGNANDUM_POSITION col_from,
				     ALIGNANDUM_POSITION col_to,
				     TYPE_DIAGONAL diagonal_from,
				     TYPE_DIAGONAL diagonal_to ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "Alignata * copyAlignata( Alignata *, const Alignata *) " << endl;
#endif

  // check parameters for filters and set them to sensible values
  if (col_from < src->getColFrom() || col_from == 0)
      col_from = src->getColFrom();
  if (col_to > src->getColTo() || col_to == 0)
      col_to = src->getColTo();
  if (row_from < src->getRowFrom() || row_from == 0)
      row_from = src->getRowFrom();
  if (row_to > src->getRowTo() || row_to == 0)
      row_to = src->getRowTo();

  dest->Clear();
  
  AlignataIterator_const it(src->begin());
  AlignataIterator_const it_end(src->end());

  for (; it != it_end; ++it) {
    const RESIDUEPAIR & p = *it;
    
    // apply filter
    TYPE_DIAGONAL	this_diagonal = calculateDiagonal(p);
    ALIGNANDUM_POSITION this_row      = p.mRow;
    ALIGNANDUM_POSITION this_col      = p.mCol;
    
    if (this_col >= col_from && this_col <= col_to) 
	continue;
    if (this_row >= row_from && this_row <= row_to) 
	continue;
    if (this_diagonal >= diagonal_from && this_diagonal <= diagonal_to)
	continue;

    dest->addPair( new RESIDUEPAIR(p) );

  }

  return dest;
}
//-----------------------------------------------------------------------------------------------
/** convert one type of Alignata-object into another using functions from the public interface. 
 Residues are skipped, which are part of filter */
Alignata * copyAlignata( Alignata * dest, 
			 const Alignata * src,
			 const Alignata * filter, 
			 const COMBINATION_MODE mode ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "Alignata * copyAlignata( Alignata*, const Alignata*, const Alignata*) " << endl;
#endif

  dest->Clear();
  
  AlignataIterator_const it(src->begin());
  AlignataIterator_const it_end(src->end());

  for (; it != it_end; ++it) {
    const RESIDUEPAIR & p = *it;
    bool keep = true;

    // apply filter
    if (filter) 
      switch (mode) {
      case RR:
	if (filter->mapRowToCol( p.mRow )) keep = false; break;
      case CR:
	if (filter->mapRowToCol( p.mCol )) keep = false; break;
      case RC:
	if (filter->mapColToRow( p.mRow )) keep = false; break;
      case CC:
	if (filter->mapColToRow( p.mCol )) keep = false; break;
      }

    if (keep)
      dest->addPair( new RESIDUEPAIR(p) );
  }
  
  return dest;
}

//----------------------------------------------------------------------------------
/** add one alignment to another
 */
Alignata * addAlignata2Alignata( Alignata * dest, 
				 const Alignata * src ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "Alignata * addAlignata2Alignata( Alignata *, const Alignata *)" << endl;
#endif

  AlignataIterator_const it(src->begin());
  AlignataIterator_const it_end(src->end());

  for (; it != it_end; ++it) 
    dest->addPair( new RESIDUEPAIR(*it) );

  dest->setScore( dest->getScore() + src->getScore());
  
  return dest;
}
//----------------------------------------------------------------------------------
 /** add one alignment to another. Map src using map_src2new.
  */
    Alignata * addMappedAlignata2Alignata( Alignata * dest, 
					   const Alignata * src, 
					   const Alignata * map_src2new,
					   const COMBINATION_MODE mode ) {
  
#ifdef DEBUG
    if (verbose > LL3) cout << "Alignata * addMappedAlignata2Alignata( Alignata *, const Alignata *)" << endl;
#endif

  AlignataIterator_const it(src->begin());
  AlignataIterator_const it_end(src->end());

  for (; it != it_end; ++it) {
    ALIGNANDUM_POSITION row = (*it).mRow;
    ALIGNANDUM_POSITION col = (*it).mCol;
    CALCTYPE score = (*it).mScore;
    
    switch (mode) {
    case RR: row = map_src2new->mapRowToCol(row); break;
    case RC: row = map_src2new->mapColToRow(row); break;
    case CR: col = map_src2new->mapRowToCol(col); break;
    case CC: col = map_src2new->mapColToRow(col); break;
    }
    
    if (row != 0 && col != 0)
      dest->addPair( new RESIDUEPAIR( row, col, score) );
  }

  dest->setScore( dest->getScore() + src->getScore());
  
  return dest;
}

//----------------------------------------------------------------------------------
 /** add one alignment to another. Map both row and column.
  */

 Alignata * addMappedAlignatas2Alignata( Alignata * dest, 
					 const Alignata * src, 
					 const Alignata * map_src_row2dest_row, 
					 const Alignata * map_src_col2dest_col ) {
   
#ifdef DEBUG
    if (verbose > LL3) cout << "Alignata * addMappedAlignatas2Alignata( Alignata *, const Alignata *)" << endl;
#endif
    
    AlignataIterator_const it(src->begin());
    AlignataIterator_const it_end(src->end());

    for (; it != it_end; ++it) {
      ALIGNANDUM_POSITION row = map_src_row2dest_row->mapRowToCol((*it).mRow);
      ALIGNANDUM_POSITION col = map_src_col2dest_col->mapRowToCol((*it).mCol);
      CALCTYPE score = (*it).mScore;
    
      if (row != 0 && col != 0)
	dest->addPair( new RESIDUEPAIR( row, col, score) );
    }

    dest->setScore( dest->getScore() + src->getScore());
  
    return dest;
}

//-----------------------------------------------------------------------------------------
/** create a new Alignatum-object by conversion from others. This method requires the addPair 
    method, that's why I can not put it into a constructor, the virtual mechanism is not 
    guaranteed to work in constructors 
*/
Alignata * combineAlignata( Alignata * dest, 
			    const Alignata * src1, 
			    const Alignata * src2, 
			    const COMBINATION_MODE mode ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignata::combineAlignata( const ImplAlignata &, const ImplAlignata &, const MODE )" << endl;
#endif
  
    dest->Clear();
  
    AlignataIterator_const it1(src1->begin());
    AlignataIterator_const it1_end(src1->end());
    AlignataIterator_const it2(src2->begin());
    AlignataIterator_const it2_end(src2->end());

    while ( it1 != it1_end && it2 != it2_end ) {
    
      const RESIDUEPAIR & x_pair = *it1;
      const RESIDUEPAIR & y_pair = *it2;
      
      ALIGNANDUM_POSITION map1 = 0;
      ALIGNANDUM_POSITION value1 = 0;
      
      ALIGNANDUM_POSITION map2 = 0;
      ALIGNANDUM_POSITION value2 = 0;
      
      switch (mode) {
	
      case RR:
	map1 = x_pair.mRow; value1 = x_pair.mCol;
	map2 = y_pair.mRow; value2 = y_pair.mCol;
	break;
	
      case CR:
	map1 = x_pair.mCol; value1 = x_pair.mRow;
	map2 = y_pair.mRow; value2 = y_pair.mCol;
	break;
	
      case RC:
	map1 = x_pair.mRow; value1 = x_pair.mCol;
	map2 = y_pair.mCol; value2 = y_pair.mRow;
	break;
	
      case CC:
	map1 = x_pair.mCol; value1 = x_pair.mRow;
	map2 = y_pair.mCol; value2 = y_pair.mRow;
	break;
      }
      
      // cout << "map1:" << map1 << " value1:" << value1 << " map2:" << map2 << " value2:" << value2 << endl;
      
      if (map1 == map2) {
	dest->addPair( new RESIDUEPAIR(value1, value2, 0)); 
	++it1;
	++it2;
      } else { 
	if (map1 < map2) 
	  ++it1;
	else 
	  ++it2;
      }
      
    }
    
    return dest;
}

//------------------------------------------------------------------------------------------------------------
/** Print pretty pairwise aligment */
void writeAlignataTable( std::ostream & output, 
			 const Alignata * src,
			 unsigned int ncols,
			 bool with_scores) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void writeAlignmentTable( std::ostream &, const Alignata *)" << endl;
#endif

  if (src->isEmpty()) 
      return;

  output << "length=" << src->getLength() 
	 << " score=" << src->getScore() 
	 << " gaps=" << src->getNumGaps() 
	 << endl;

  AlignataIterator_const it = src->begin();
  AlignataIterator_const it_end = src->end();

  unsigned int col = 0;

  for (; it!=it_end; ++it) {
    output << std::setw(6) << it->mRow << std::setw(6) << it->mCol;
    if (with_scores) 
      output << setw(6) << setprecision(2) << it->mScore;
    if (++col == ncols) {
      output << endl;
      col = 0;
    } else {
      output << '|';
    }
  }

  return;
}
 
 
//------------------------------------------------------------------------------------------------------------
/** Print pretty pairwise aligment */
void writePairAlignment( std::ostream & output, const Alignandum * row, const Alignandum * col, const Alignata * src ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void writePairAlignment( std::ostream &, const Alignandum *, const Alignandum *, const Alignata *)" << endl;
#endif
  
  if (src->isEmpty()) 
    return;

  Alignata * map_row2new = makeAlignataVector();
  Alignata * map_col2new = makeAlignataVector();
  
  fillAlignataSummation( map_row2new, 
			 map_col2new, 
			 src, 
			 true, true);
  
  Alignatum * row_alignatum = makeAlignatum( row, map_row2new);
  Alignatum * col_alignatum = makeAlignatum( col, map_col2new);
  
  output << *row_alignatum << endl;
  output << *col_alignatum << endl;

  delete row_alignatum;
  delete col_alignatum;
  delete map_row2new;
  delete map_col2new;

  return;
}

/** convenience function */
//--------------------------------------------------------------------------------------------------------------
/** Print pretty wraparound pairwise aligment. Iteration is over rows. Pairs have to be sorted first by row and then by column? */
void writeWraparoundAlignment( std::ostream & output, const Alignandum * row, const Alignandum * col, const Alignata * ali ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void writeWraparoundAlignment( std::ostream &, const Alignandum *, const Alignandum *, const Alignata *)" << endl;
#endif

  int col_len = col->getLength();
 
  int *inserts    = new int[col_len + 1];
  int *col_counts = new int[col_len + 1];
  int *position   = new int[col_len + 1];
  int i;

    
  for (i = 0; i <= col_len; i++) {
    col_counts[i] = 0;
    inserts[i]    = 0;
    position[i]   = -1;
  }
 
  AlignataIterator_const it = ali->begin();
  AlignataIterator_const it_end = ali->end();
  
  int nrepeats = 1;
 
  // go through alignemnt by rows and maximum insertions before every row-position
  int last_col = it->mCol - 1;
  int last_row = it->mRow - 1;
 
  for (; it != it_end; ++it) {
    int current_col = it->mCol;
    int current_row = it->mRow;
    
    int col_ins = current_col - last_col - 1;                 // negative, if wrapping around
    int row_ins = current_row - last_row - 1;
    
    if (col_ins < 0) nrepeats ++;
    
    if (row_ins > MAX_INSERTLENGTH) row_ins = MAX_INSERTLENGTH;
    
    // if ( (col_ins > 1) && (inserts[current_col] < col_ins)) inserts[current_col] = col_ins;
    if ( (row_ins > 1) && (inserts[current_col] < row_ins)) 
      inserts[current_col] = row_ins;

    col_counts[current_col]++;
    last_col = current_col;
    last_row = current_row;
  }
 
  // count total number of insertions for calculating the total length of the alignment
  int total_inserts = 0; for (i = 1; i <= col_len; i++) total_inserts += inserts[i];
  
  // calculate length of alignment and starting and ending position, ali_len is chosen, so that positions in strings later start at 1
  int first_pos = 1;       while (first_pos <= col_len && col_counts[first_pos] < 1) first_pos++;
  int last_pos  = col_len; while (last_pos  >= 1       && col_counts[last_pos]  < 1) last_pos --;
  int ali_len   = last_pos - first_pos + 1 + total_inserts + 1;
  
  // create map for starting positions
  for (i = first_pos; i <= last_pos; i++) position[i] = 1 + position[i-1] + inserts[i];
  
  // allocate memory, row 0 is for row, the remaining for the columns
  char gap_char = getDefaultTranslator()->getGapChar();
  char * buffer = new char[ali_len * (nrepeats + 1)];
  for (i = 0; i < (ali_len * (nrepeats + 1)); i++) buffer[i] = gap_char;
  
  // build columns
  it = ali->begin();
  last_col = it->mCol - 1;
  last_row = it->mRow - 1;
  int repeat_no = 0;
  for (; it != it_end; ++it) {                         
    int current_col = it->mCol;
    int current_row = it->mRow;
    
    int col_ins        = current_col - last_col - 1;
    int row_ins        = current_row - last_row - 1;
    
    if (col_ins < 0) repeat_no++;                   // check if new repeat appeared
    
    if (col_ins > MAX_INSERTLENGTH) col_ins = MAX_INSERTLENGTH;
    if (row_ins > MAX_INSERTLENGTH) row_ins = MAX_INSERTLENGTH;
    
    // some remarks on the following:
    
    // add aligned residue and inserts for row, row is mapped onto col
    int pos = position[current_col] + ali_len * repeat_no;
    int x   = current_row;
    buffer[pos] = row->asChar( x-- );
    for (i = pos - 1; i >= pos - row_ins; i--) buffer[i] = row->asChar( x-- ) + 32;
    
    // add aligned residues and inserts for col
    pos = position[current_col] + ali_len * (nrepeats);
    x   = current_col;
    buffer[pos] = col->asChar( x-- );
    
    for (i = pos - 1; i >= pos - col_ins; i--) {
      int c = col->asChar( x );
      if (col_counts[x] < 1)
	c += 32;
      buffer[i] = c;
      x--;
    }
 
    last_row = current_row;
    last_col = current_col;
    
  }
  
  // write result to stream
  
  int pos = 0;
  for (i = 0; i <= nrepeats; i++) {
    output << std::string( &buffer[pos], ali_len - 1) << endl;
    pos += ali_len;
  }
  
  delete [] buffer;
  delete [] position;
  delete [] inserts;
  delete [] col_counts;
  
  return;
}

//----------------------------------------------------------------------------------------------------------------------
/** writes a compressed alignment in diagonal format to string out.
    The diagonal format is:
    diagonal:-n1+n2-n3+n4;diagonal;...

    Although other alignment types can be written in this format, it makes most sense
    to use it for ImplAlignataMatrixDiagonal, because then most compression is achieved.
    In ImplAlignataMatrixDiagonal the aligned pairs are sorted first by diagonal and then
    by row.
    
    @param src		alignment that is to be written
    @param	 diagonal_from	only look at diagonals larger than diagonal_from
    @param	 diagonal_to	only look at diagonals smaller than diagonal_to
    
*/
 void writeAlignataCompressedDiagonal( const Alignata * src, 
				       std::string & out_string, 
				       bool reverse,
				       ALIGNANDUM_POSITION row_from,
				       ALIGNANDUM_POSITION row_to,
				       ALIGNANDUM_POSITION col_from,
				       ALIGNANDUM_POSITION col_to,
				       TYPE_DIAGONAL diagonal_from,
				       TYPE_DIAGONAL diagonal_to
				       ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void writeAlignataCompressedDiagonals( const Alignata *, std::string &, TYPE_DIAGONAL, TYPE_DIAGONAL)" << endl;
#endif

  out_string = "";

  // sanity check
  if (src->isEmpty()) 
      return;

  // check parameters for filters and set them to sensible values
  if (col_from < src->getColFrom() || col_from == 0)
      col_from = src->getColFrom();
  if (col_to > src->getColTo() || col_to == 0)
    col_to = src->getColTo();
  if (row_from < src->getRowFrom() || row_from == 0)
      row_from = src->getRowFrom();
  if (row_to > src->getRowTo() || row_to == 0)
    row_to = src->getRowTo();

  if (diagonal_from > diagonal_to) {
      diagonal_from = -MAX_DIAGONAL;
      diagonal_to   =  MAX_DIAGONAL;
  }

  // declare variables you need for iteration of the pairs
  AlignataIterator_const it(src->begin());
  AlignataIterator_const it_end(src->end());

  std::ostrstream out;
  
  TYPE_DIAGONAL last_diagonal = calculateDiagonal( *it );
  TYPE_DIAGONAL this_diagonal = 0;

  ALIGNANDUM_POSITION this_row	 = it->mRow;
  ALIGNANDUM_POSITION this_col	 = it->mCol;
  ALIGNANDUM_POSITION last_row	 = this_row -1;
  ALIGNANDUM_POSITION emissions = 0;
  ALIGNANDUM_POSITION initial_gaps = 0;

  bool first = true;

  // now iterate over all pairs in the alignment
  for (;it!= it_end; ++it) {
    
    this_diagonal = calculateDiagonal(*it);
    this_row      = it->mRow;
    this_col      = it->mCol;

#ifdef DEBUG
    if (verbose > LL5) {
	cout << "Pair:" << *it << endl;
	cout << "last_diagonal=" << last_diagonal << " this_diagonal=" << this_diagonal 
	     << " emissions=" << emissions << "last_row=" << last_row << endl;
    }
#endif

    // apply filters
    if (this_col < col_from || this_col > col_to) 
      continue;
    if (this_row < row_from || this_row > row_to) 
      continue;
    if (this_diagonal < diagonal_from || this_diagonal > diagonal_to)
      continue;
    
    if (last_diagonal != this_diagonal || last_row >= this_row || first) {

	if (!first) 
	    out  << "+" << emissions << ";";

	// write last emission and switch to new diagonal
	if (this_diagonal < 0) 
	    initial_gaps = this_col - 1; 
	else
	    initial_gaps = this_row - 1;

	if (reverse)
	    out << -this_diagonal << ":-" << initial_gaps;
	else
	    out << this_diagonal << ":-" << initial_gaps;

	first = false;
      
	last_diagonal = this_diagonal;
	last_row = this_row;
	emissions = 1;
	continue;
    }
    
    // insert a gap
    if (last_row < this_row - 1) {
	out << "+" << emissions << "-" << (this_row - last_row - 1);
	emissions = 0;
    }
    
    last_row = this_row;
    emissions ++;
	   
  }
  
  out << "+" << emissions << ends;
  out_string = out.str(); 

  // unfreeze streams, so the memory gets freed
  out.freeze(false);
}       

//----------------------------------------------------------------------------------------------------------------------
/** writes a compressed alignment in diagonal format to string out.
    The diagonal format is:
    diagonal:-n1+n2-n3+n4;diagonal;...

    Although other alignment types can be written in this format, it makes most sense
    to use it for ImplAlignataMatrixDiagonal, because then most compression is achieved.
    In ImplAlignataMatrixDiagonal the aligned pairs are sorted first by diagonal and then
    by row.
    
    @param src		alignment that is to be written
    @param	 diagonal_from	only look at diagonals larger than diagonal_from
    @param	 diagonal_to	only look at diagonals smaller than diagonal_to
    
*/
Alignata * fillAlignataCompressedDiagonal( Alignata * dest, 
					   const std::string & ali,
					   bool reverse) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void fillAlignataCompressedDiagonals( const Alignata * dest, const std::string &)" << endl;
#endif

  std::istrstream is_ali( ali.c_str() );   

  // set these/use these parameters to shift alignment
  ALIGNANDUM_POSITION row_from = 1;
  ALIGNANDUM_POSITION col_from = 1;

  // row and col are the index of the next dot to be written.
  ALIGNANDUM_POSITION row = row_from;   
  ALIGNANDUM_POSITION col = col_from;   
 
  dest->Clear();

  // read diagonal wise
  while (!is_ali.eof()) {
    
    // read the diagonal
    TYPE_DIAGONAL diagonal;
    is_ali >> diagonal;
    is_ali.ignore();		// skip colon


    // for a new diagonal, position yourself at the first residue 
    // on the diagonal
    if (diagonal < 0) {
      row = -diagonal + row_from;
      col = 1;
    } else {
      row = 1;
      col = diagonal + col_from;
    }
    
    
    while (is_ali.peek() != ';' && !is_ali.eof()) {

      ALIGNANDUM_POSITION d = 0;   
      is_ali >> d;
      // write a gap
      if (d < 0) {
	row -= d;
	col -= d;
      } else {
	// emit dots
	while (d > 0) {
	    if (reverse)
		dest->addPair( col++, row++, 0 );       
	    else 
		dest->addPair( row++, col++, 0 );       
	  d--;
	}
      }
    } 
    is_ali.ignore();		// skip semicolon
  }   
    
  return dest;
}

//----------------------------------------------------------------------------------------------------
//** the same as above, but use filter. duplication for speed resaons */
Alignata * fillAlignataCompressedDiagonal( Alignata * dest, 
					   const std::string & ali,
					   ALIGNANDUM_POSITION filter_row_from,
					   ALIGNANDUM_POSITION filter_row_to,
					   ALIGNANDUM_POSITION filter_col_from,
					   ALIGNANDUM_POSITION filter_col_to,
					   TYPE_DIAGONAL filter_diagonal_from,
					   TYPE_DIAGONAL filter_diagonal_to
					   ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void fillAlignataCompressedDiagonals( const Alignata * dest, const std::string &)" << endl;
#endif

  if (filter_row_to == 0)
    filter_row_to = UPPER_BOUNDS_POSITION;

  if (filter_col_to == 0)
    filter_row_to = UPPER_BOUNDS_POSITION;

  std::istrstream is_ali( ali.c_str() );   

  // set these/use these parameters to shift alignment
  ALIGNANDUM_POSITION row_from = 1;
  ALIGNANDUM_POSITION col_from = 1;

  // row and col are the index of the next dot to be written.
  ALIGNANDUM_POSITION row = row_from;   
  ALIGNANDUM_POSITION col = col_from;   
 
  dest->Clear();

  // read diagonal wise
  while (!is_ali.eof()) {
    
    // read the diagonal
    TYPE_DIAGONAL diagonal;
    is_ali >> diagonal;
    is_ali.ignore();		// skip colon


    // for a new diagonal, position yourself at the first residue 
    // on the diagonal
    if (diagonal < 0) {
      row = -diagonal + row_from;
      col = 1;
    } else {
      row = 1;
      col = diagonal + col_from;
    }
    
    
    while (is_ali.peek() != ';' && !is_ali.eof()) {

      ALIGNANDUM_POSITION d = 0;   
      is_ali >> d;
      // write a gap
      if (d < 0) {
	row -= d;
	col -= d;
      } else {
	// emit dots
	while (d > 0) {
	  
	  if ( col >= filter_col_from && 
	       col <= filter_col_to &&
	       row >= filter_row_from && 
	       row <= filter_row_to &&
	       diagonal >= filter_diagonal_from &&
	       diagonal <= filter_diagonal_to)
	    dest->addPair( row, col, 0 );       

	  row++; col++;
	  d--;
	}
      }
    } 
    is_ali.ignore();		// skip semicolon
  }   
    
  return dest;
}


//---------------------------------------------------------------------------------------
/** remove residues from an alignment, that are part of another alignment
    @param mode: specifies, which residues are looked up. If mode = RR, then every pairs is eliminated from dest,
	where the row is also present as a row-residue in filter.
 
*/
Alignata * filterAlignataRemovePairs( Alignata * dest, const Alignata * filter, const COMBINATION_MODE mode ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignata::combineAlignata( const ImplAlignata &, const ImplAlignata &, const MODE )" << endl;
#endif
    
    // delete pairs in-situ.
    // Note: The iterators should not be invalidated, if you remove a residue. This should work with
    // the Alignata-objects that are based on <set>
    // Another alternative would be to iterate over the filter. However, then the problem is,
    // that multiple residues would have to be deleted.

    AlignataIterator it1(dest->begin());
    AlignataIterator it1_end(dest->end());

    while ( it1 != dest->end()) {
	
	const RESIDUEPAIR & x_pair = *it1;
	switch (mode) {
	    
	case RR:
	    if (filter->mapRowToCol( x_pair.mRow ))
		dest->removePair( x_pair );
	    break;
	case CR:
	    if (filter->mapRowToCol( x_pair.mCol ))
		dest->removePair( x_pair );
	    break;
	case RC:
	    if (filter->mapColToRow( x_pair.mRow ))
		dest->removePair( x_pair );
	    break;
	case CC:
	    if (filter->mapColToRow( x_pair.mCol ))
		dest->removePair( x_pair );
	    break;
	}
	it1++;
    }
    
    return dest;
}

//-------------------------------------------------------------------------------------------------
// this works only for pairswise alignments, where both alignments are sorted by row.
Alignata * filterPairAlignataRemovePairsPairWiseSorted( Alignata * dest, 
							const Alignata * filter, 
							const COMBINATION_MODE mode ) {

#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignata::combineAlignata( const ImplAlignata &, const ImplAlignata &, const MODE )" << endl;
#endif
  
    // work with a temporary copy, because I am not entirely sure, if my iterators will fail,
    // if I delete residues while they are active: i.e. spend some time on improving the iterators :=)

    const Alignata * copy = dest->getClone();
  
    AlignataIterator_const it1(copy->begin());
    AlignataIterator_const it1_end(copy->end());
    AlignataIterator_const it2(filter->begin());
    AlignataIterator_const it2_end(filter->end());
    
    while ( it1 != it1_end && it2 != it2_end ) {
    
      const RESIDUEPAIR & x_pair = *it1;
      const RESIDUEPAIR & y_pair = *it2;
      
      ALIGNANDUM_POSITION map1 = 0;
      ALIGNANDUM_POSITION map2 = 0;
      
      switch (mode) {
	
      case RR:
	map1 = x_pair.mRow;
	map2 = y_pair.mRow; 
	break;
	
      case CR:
	map1 = x_pair.mCol; 
	map2 = y_pair.mRow;
	break;
	
      case RC:
	map1 = x_pair.mRow; 
	map2 = y_pair.mCol;
	break;
	
      case CC:
	map1 = x_pair.mCol;
	map2 = y_pair.mCol; 
	break;
      }
      
      // cout << "map1:" << map1 << " map2:" << map2 << endl;

      if (map1 == map2) {
	  dest->removePair( x_pair );
	  it1++;
	  it2++;
      } else { 
	  if (map1 < map2) 
	      it1++;
	  else 
	      it2++;
      }
      
    }

    delete copy;
    
    return dest;
}

/*--------------------------------------------------------------------------------------------------------------
  create an identity alignment between residues from and to in row using an offset for col
*/
Alignata * fillAlignataIdentity( Alignata * dest, 
				 ALIGNANDUM_POSITION row_from, 
				 ALIGNANDUM_POSITION row_to, 
				 ALIGNANDUM_POSITION col_offset) {
#ifdef DEBUG
  if (verbose > LL3) cout << "Alignata * fillAlignataIdentity( Alignata *, ALIGNANDUM_POSITION, ALIGNANDUM_POSITION, ALIGNANDUM_POSITION)" << endl;
#endif
  ALIGNANDUM_POSITION i;
  for (i = row_from; i <= row_to; i++) 
    dest->addPair( new RESIDUEPAIR( i, i + col_offset, 0));
    
    return dest;
}

//-------------------------------------------------------------------------------------------------
// calculate percent identity, normalised on alignment mLength
double calculatePercentIdentity( const Alignata * src,
				 const Alignandum * row, 
				 const Alignandum * col) {
  
    if (src->getLength() == 0) return 0;

    AlignataIterator_const it = src->begin();
    AlignataIterator_const it_end = src->end();
    
    int nidentities = 0;

    for (; it != it_end; ++it) {
	RESIDUEPAIR p = *it;

	if ( row->asResidue(p.mRow) == col->asResidue(p.mCol) ) 
	  nidentities++;
    }
    
    return (double)nidentities / (double)src->getLength();
}


//----------------------------------------------------------------------------------------------------
double calculatePercentSimilarity( const Alignata * src) {
  
    if (src->getLength() == 0) return 0;

    AlignataIterator_const it = src->begin();
    AlignataIterator_const it_end = src->end();

    int nsimilarities = 0;

    for (; it != it_end; ++it)
      if ((*it).mScore > 0) 
	nsimilarities++;
    
    return (double)nsimilarities / (double)src->getLength();
}

/*--------------------------------------------------------------------------------------------------------------
  The following cases are dealt with:
	gaps in the beginning of an alignment
	gaps in the middle of an alignment
*/
Alignata * fillAlignataCompressed( Alignata * dest, 
				   ALIGNANDUM_POSITION row_from, 
				   const std::string & row_ali,
				   ALIGNANDUM_POSITION col_from, 
				   const std::string & col_ali
				   ) {
#ifdef DEBUG   
  if (verbose > LL3) cout << "fillAlignataCompressed( int, int, const std::string &, int, int, const std::string) " << endl; 
#endif  
 
  std::istrstream is_row( row_ali.c_str() );   
  std::istrstream is_col( col_ali.c_str() );  

  ALIGNANDUM_POSITION row = row_from;   
  ALIGNANDUM_POSITION col = col_from;   
  ALIGNANDUM_POSITION d_row = 0;   
  ALIGNANDUM_POSITION d_col = 0;  
 
  dest->Clear();
  
#ifdef DEBUG
  cout << row_ali << " " << col_ali << endl;
#endif

  if (!(is_row >> d_row)) return dest;
  if (!(is_col >> d_col)) return dest;

  while (true) {
      
      // cout << "0:" << d_row << " " << d_col << " " << row << " " << col << endl;
      // entry: d_row and d_col > 0
      // end: d_col or d_row or both are 0
      // emit pairs for aligned regions, i.e. d_row and d_col are positive
      while (d_row > 0 && d_col > 0) {         
	dest->addPair( new RESIDUEPAIR (row, col, 0) );         
	d_row--; 
	d_col--;         
	row++; 
	col++;       
      }  

      // cout << "1:" << d_row << " " << d_col << " " << row << " " << col << endl;
      
      // entry: d_row and d_col < 0
      // end: d_col or d_row or both are 0
      // simply skip. This region is a gap in both sequences
      if (d_row < 0 && d_col < 0) {         
	  if (d_row < d_col) {
	      d_row -= d_col;
	      d_col = 0;
	  } else {
	      d_col -= d_row;
	      d_row = 0;
	  }
      }
      
      // cout << "2:" << d_row << " " << d_col << " " << row << " " << col << endl;

      // entry: d_row > 0, d_col < 0:
      // exit: d_row or d_col or both are 0
      // emitting characters only from row
      if (d_row > 0 && d_col < 0) {
	  if (d_row > -d_col ) {
	      d_row += d_col;
	      row   -= d_col;		// emit
	      d_col  = 0;
	  } else {
	      d_col += d_row;
	      row   += d_row;		// emit
	      d_row  = 0;
	  }
      }

      // cout << "3:" << d_row << " " << d_col << " " << row << " " << col << endl;

      // entry: d_row < 0, d_col > 0:
      // exit: d_row or d_col or both are 0
      // emitting characters only from col
      if (d_col > 0 && d_row < 0) {
	  if (d_col > -d_row ) {
	      d_col += d_row;
	      col   -= d_row;		// emit 
	      d_row  = 0;
	  } else {
	      d_row += d_col;
	      col   += d_col;		// emit
	      d_col  = 0;
	  }
      }

      // cout << "4:" << d_row << " " << d_col << " " << row << " " << col << endl;
      
      if (d_row == 0) 
	  if (!(is_row >> d_row)) break;	// read new d_row
      
      if (d_col == 0) 
	  if (!(is_col >> d_col)) break;	// read new d_col
      
  }	  

  return dest;
}    

//--------------------------------------------------------------------------------------------------------------
Alignata * fillAlignataExplicit( Alignata * dest, 
				 ALIGNANDUM_POSITION row_from, 
				 const std::string & row_ali,
				 ALIGNANDUM_POSITION col_from, 
				 const std::string & col_ali
				 ) {
#ifdef DEBUG   
  if (verbose > LL3) cout << "fillAlignataExplicit( int, const std::string &, int, const std::string) " << endl; 
#endif  

  char gap_char = getDefaultTranslator()->getGapChar();
  
  ALIGNANDUM_POSITION row = row_from;   
  ALIGNANDUM_POSITION col = col_from;   

  for (unsigned int i = 0; i < row_ali.size(); i++) {

    if (row_ali[i] != gap_char && col_ali[i] != gap_char) 
      dest->addPair( new RESIDUEPAIR (row, col, 0) );         

    if (row_ali[i] != gap_char) 
      row++;

    if (col_ali[i] != gap_char) 
      col++;
  }

  return dest;
}    

//----------------------------------------------------------------------------------------------------
Alignata * rescoreAlignment( Alignata * dest,
			     const Alignandum * row,
			     const Alignandum * col, 
			     const SubstitutionMatrix * matrix) {
#ifdef DEBUG
  if (verbose > LL3) cout << "rescoreAlignment(Alignata, const Alignandum *, const Alignandum *) " << endl; 
#endif
  
  AlignataIterator it(dest->begin());
  AlignataIterator it_end(dest->end());
  
  if (!matrix) 
    matrix = getDefaultSubstitutionMatrix();
  
  for (; it != it_end; ++it)
      it->mScore = matrix->getScore( row->asResidue( it->mRow), col->asResidue( it->mCol) );

  return dest;
}

//----------------------------------------------------------------------------------------------------
Alignata * rescoreAlignment( Alignata * dest,
			     const CALCTYPE score ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "rescoreAlignment(Alignata, const Alignandum *, const Alignandum *) " << endl; 
#endif
  
  AlignataIterator it(dest->begin());
  AlignataIterator it_end(dest->end());
  
  for (; it != it_end; ++it)
      it->mScore = score;

  return dest;
}


//----------------------------------------------------------------------------------------------------
Alignata * calculateAffineScore( Alignata * dest,
				 const CALCTYPE gop, 
				 const CALCTYPE gep) {
#ifdef DEBUG
  if (verbose > LL3) cout << "rescoreAlignment(Alignata, const Alignandum *, const Alignandum *) " << endl; 
#endif
  
  CALCTYPE score = 0;

  AlignataIterator it(dest->begin());
  AlignataIterator it_end(dest->end());
  
  ALIGNANDUM_POSITION last_row = 1000000;
  ALIGNANDUM_POSITION last_col = 1000000;

  for (; it != it_end; ++it) {
      ALIGNANDUM_POSITION d;
      ALIGNANDUM_POSITION row = it->mRow;
      ALIGNANDUM_POSITION col = it->mCol;

      if ( (d = (row - last_row - 1)) > 0)
	  score += gop + d * gep;
      
      if ( (d = (col - last_col - 1)) > 0)
	  score += gop + d * gep;

      score += it->mScore;
      last_row = row;
      last_col = col;
  }


  dest->setScore(score);
  return dest;
}

//----------------------------------------------------------------------------------------------------
/** fill an alignment with a repeat unit from a wrap-around alignment */
/* code still broken with the skip_negative_ends */
Alignata * fillAlignataRepeatUnit( Alignata * dest, 
				   const Alignata * src,
				   ALIGNANDUM_POSITION first_row_residue,
				   bool skip_negative_ends) {

  if (first_row_residue == 0)
    first_row_residue = src->getRowFrom();
  
  AlignataIterator_const it(src->begin());
  AlignataIterator_const it_end(src->end());

  dest->Clear();

  while ( (it != it_end) && (it->mRow < first_row_residue)) ++it;
  if (it == it_end) return dest;

  ALIGNANDUM_POSITION last_col = it->mCol - 1;
  
  skip_negative_ends = false;
  if (skip_negative_ends) 
    while ( (it != it_end) && (it->mCol > last_col) ) {
      if (it->mScore > 0) break;
      last_col = it->mCol;
    }

  ALIGNANDUM_POSITION last_positive = it->mRow;
  while ( (it != it_end) && (it->mCol > last_col) ) {
    last_col = it->mCol;
    dest->addPair( new RESIDUEPAIR(it->mRow, last_col, it->mScore) );         
    if (it->mScore > 0) last_positive = it->mRow;
    ++it;
  }

  if (skip_negative_ends)
    dest->removeRowRegion( last_positive + 1, dest->getRowTo());
  
  return dest;
}

//------------------------------------------------------------------------------------------------------
  inline ALIGNANDUM_POSITION insertResidues( Alignata * dest, 
					     ALIGNANDUM_POSITION last_res, 
					     ALIGNANDUM_POSITION current_res, 
					     ALIGNANDUM_POSITION current_combined) {
      
    for( ALIGNANDUM_POSITION i = last_res + 1; i < current_res; i++) 
      dest->addPair( new RESIDUEPAIR(i, current_combined++, 0.0));
    return current_combined;
  }


/** 
    return the maps of row/col of an alignment to the summation of the alignment. This is useful for 
    building multiple alignemnts.

    For example: With two sequences of length 10 and alignment src between them:
    
    src = 3 4 | 4 5 | 5 7 | 9 9 

    the result would be:

    A: insert_gaps_row = true, insert_gaps_col = true, use_end_row = true, use_end_col = true
    dest1 = 1 1 | 2 2 |     |     |     | 3 6 | 4 7 |     | 5 9 | 6 10 | 7 11 | 8 12 |      | 9 14 | 10 15 |       |
    dest2 =     |     | 1 3 | 2 4 | 3 5 | 4 6 | 5 7 | 6 8 | 7 9 |      |      |      | 8 13 | 9 14 |       | 10 16 | 
    
    B: insert_gaps_row = false, insert_gaps_col = true, use_end_row = true, use_end_col = false
    dest1 = 1 1 | 2 2 | 3 3 | 4 4 | 5 5 | 6  6 | 7  7 | 8  8 | 9  9 | 10 10 |
    dest2 =     |     | 4 3 | 5 4 | 7 5 |      |      |      | 9  9 |       

    C: insert_gaps_row = false, insert_gaps_col = true, use_end_row = false, use_end_col = false
    dest1 = 3 1 | 4 2 | 5 3 | 6  4 | 7  5 | 8  6 | 9  7 | 
    dest2 = 4 1 | 5 2 | 7 3 |      |      |      | 9  7 |       

    D: insert_gaps_row = false, insert_gaps_col = false, use_end_row = false, use_end_col = false
    dest1 = 3 1 | 4 2 | 5 3 | 9  4 | 
    dest2 = 4 1 | 5 2 | 7 3 | 9  4 |       

    E: insert_gaps_row = true, insert_gaps_col = true, use_end_row = false, use_end_col = false
    dest1 = 3 1 | 4 2 |     | 5 4 | 6  5 | 7  6 | 8  7 |     | 9 9 |
    dest2 = 4 1 | 5 2 | 6 3 | 7 4 |      |      |      | 8 8 | 9 9 |

    A can be used for building complete multiple alignments, where no part of the
    sequences are missing.
    B can be used for building multiple alignments for representatives that include
    all the neighbours.
    E can be used for showing pairwise alignments.

    Note:
 
    If there is a gap in both sequences, the residues in the first alignment dest1 are 
    aligned to the right of the gap, while the residues in the second alignment dest2
    are aligned to the right.

    @param	map_row2combined     
    @param	map_col2combined
    @param      source
    @param	insert_gaps_row
    @param	insert_gaps_col
    @param	use_end_row
    @param	use_end_col
    @param	row_length
    @param	col_length
    
*/    
void fillAlignataSummation( Alignata * dest1, 
			    Alignata * dest2, 
			    const Alignata * src,
			    bool insert_gaps_row,
			    bool insert_gaps_col, 
			    bool use_end_row,
			    bool use_end_col, 
			    ALIGNANDUM_POSITION row_length,
			    ALIGNANDUM_POSITION col_length) {
#ifdef DEBUG
    if (verbose > LL3) 
      cout << "ImplAlignata::fillAlignataSummation()" << endl;
#endif

    dest1->Clear();
    dest2->Clear();
  
    AlignataIterator_const it(src->begin());
    AlignataIterator_const it_end(src->end());
    
    // current position in combined alignment
    ALIGNANDUM_POSITION current_combined = 1;
    ALIGNANDUM_POSITION current_row = it->mRow;
    ALIGNANDUM_POSITION current_col = it->mCol;

    // add positions before first aligned position
    if (use_end_row)
      current_combined = insertResidues( dest1, 0, current_row, current_combined);
    if (use_end_col)
      current_combined = insertResidues( dest2, 0, current_col, current_combined);

    ALIGNANDUM_POSITION last_row = current_row;
    ALIGNANDUM_POSITION last_col = current_col;

    // iteration over alignmnet
    for ( ; it != it_end; ++it ) {
      CALCTYPE current_score = it->mScore;
      current_row = it->mRow;
      current_col = it->mCol;

      // add gaps before match:
      // 1. row
      if (insert_gaps_col)
	current_combined = insertResidues( dest1, last_row, current_row, current_combined);	

      // 2. col
      if (insert_gaps_row) 
	current_combined = insertResidues( dest2, last_col, current_col, current_combined);

      // add match
      dest1->addPair(new RESIDUEPAIR(current_row, current_combined, current_score));
      dest2->addPair(new RESIDUEPAIR(current_col, current_combined, current_score));

      current_combined++;

      last_row = current_row;
      last_col = current_col;
      
    }

    // add positions after last aligned position
    if (use_end_row)
      current_combined = insertResidues( dest1, current_row, row_length + 1, current_combined);
    if (use_end_col)
      current_combined = insertResidues( dest2, current_col, col_length + 1, current_combined);

}
//-----------------------------------------------------------------------------------------
 /** remove all those residues from an alignmnent, which are not
     in sequence. This ensures, that col_i < col_i+1 and row < row_i+1
     Only use with AlignataVector

 */
Alignata * flattenAlignata( Alignata * dest ) {
  
  AlignataIterator it(dest->begin());
  AlignataIterator it_end(dest->end());

  ALIGNANDUM_POSITION last_row = dest->getRowFrom() - 1;
  ALIGNANDUM_POSITION last_col = dest->getColFrom() - 1;

  ALIGNANDUM_POSITION max_row = dest->getRowTo();
  
  for (; it != it_end; ++it) {
    
    const RESIDUEPAIR & p = *it;
    
    ALIGNANDUM_POSITION current_row = p.mRow;
    ALIGNANDUM_POSITION current_col = p.mCol;

    // if there is an intersection, save current alignment
    if (current_row <= last_row || current_col <= last_col) {
      dest->removePair( p );
    } else {
      last_row = current_row;
      last_col = current_col;
    }
    
    // patch, as iterator becomes invalidated
    if (current_row == max_row) 
      break;

  }

  return dest;
}


//-----------------------------------------------------------------------------------------
/** split an alignment, if there are gaps larger than a certain threshold either in row or
    col or both.
*/
FragmentVector * splitAlignata( const Alignata * src, 
				const int max_gap_width,
				bool split_row, bool split_col) {
    
  AlignataIterator_const it(src->begin());
  AlignataIterator_const it_end(src->end());
      
  FragmentVector * result = new FragmentVector();

  ALIGNANDUM_POSITION last_col = src->getColFrom() -1;
  ALIGNANDUM_POSITION last_row = src->getRowFrom() -1;

  Alignata * current_ali = src->getNew();

  for (; it != it_end; ++it) {
      
    const RESIDUEPAIR & p = *it;
    ALIGNANDUM_POSITION current_row = p.mRow;
    ALIGNANDUM_POSITION current_col = p.mCol;
      
    if ( (split_col && ((current_col - last_col) > max_gap_width)) ||
	 (split_row && ((current_row - last_row) > max_gap_width))) {
	
      result->push_back( current_ali );
      
      current_ali = src->getNew();
    }
	   
    current_ali->addPair( new RESIDUEPAIR(p) );
    
    last_row = current_row;
    last_col = current_col;

  }
  
  result->push_back( current_ali );    
  
  return result;
}

//-----------------------------------------------------------------------------------------
/** split an alignment at points of intersection with another alignment.
 */ 
FragmentVector * splitAlignata( const Alignata * src1, 
				const Alignata * src2, 
				const COMBINATION_MODE mode ) {

  AlignataIterator_const it1(src1->begin());
  AlignataIterator_const it1_end(src1->end());

  AlignataIterator_const it2(src2->begin());
  AlignataIterator_const it2_end(src2->end());
      
  FragmentVector * result = new FragmentVector();



  bool in_row1 = true;
  bool in_row2 = true;

  switch (mode) {
  case RR: break;
  case CR:
    in_row1 = false; break;
  case RC:
    in_row2 = false; break;
  case CC:
    in_row1 = false; in_row2 = false; break;
  }

  ALIGNANDUM_POSITION other_pos = ((in_row2) ? src2->getRowFrom() : src2->getColFrom());
  
  Alignata * current_ali = src1->getNew();

  for (; it1 != it1_end; ++it1) {
    
    const RESIDUEPAIR & p = *it1;

    ALIGNANDUM_POSITION current_pos = (in_row1) ? p.mRow : p.mCol;

    // if there is an intersection, save current alignment
    if (current_pos > other_pos) {
      
      /* test if alignment has residues in it 
	 (might not be the case in first iteration). */
      if (current_ali->getLength() > 0) {
	result->push_back( current_ali );      
	current_ali = src1->getNew();
      }

      // advance to next pos in other alignment
      while ( other_pos < current_pos && ++it2 != it2_end) 
	other_pos = (in_row2) ? it2->mRow : it2->mCol;
      
      if (it2 == it2_end)
	other_pos = UPPER_BOUNDS_POSITION;
      
    }
	   
    current_ali->addPair( new RESIDUEPAIR(p) );
  }
  
  result->push_back( current_ali );    
  
  return result;

}



//-----------------------------------------------------------------------------------------
/** complement a pairwise alignment. If there is a gap of the same length in both row and
    col, the corresponding residues are added to the alignment.
*/
Alignata * complementAlignata( Alignata * dest, 
			       const ALIGNANDUM_POSITION max_length ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignata::combineAlignata( const ImplAlignata &, const ImplAlignata &, const MODE )" << endl;
#endif
  
    AlignataIterator it(dest->begin());
    AlignataIterator it_end(dest->end());

    ALIGNANDUM_POSITION last_row = it->mRow;
    ALIGNANDUM_POSITION last_col = it->mCol;

    for (; it != it_end; ++it ) {

	ALIGNANDUM_POSITION this_row = it->mRow;
	ALIGNANDUM_POSITION this_col = it->mCol;

	ALIGNANDUM_POSITION gap_row = this_row - last_row - 1;
      
	if ( gap_row > 0 && 
	     gap_row <= max_length &&
	     gap_row == (this_col - last_col - 1) ) {
	    while (++last_row < this_row) {
		++last_col;
		dest->addPair( new RESIDUEPAIR(last_row, last_col, 0));
	    }
	}
	
	last_row = this_row;
	last_col = this_col;
    }
    
    return dest;
}

/** remove small fragments from alignment.
    This method removes fragments from an alignment. A fragment
    is a part of an alignment, that is short (max_fragment_length)
    and surrounded by large gaps (min_gap_length). 
    Fragments are only removed in col.
 */

void  removeFragments( Alignata * dest,
		       unsigned int window_size,
		       unsigned int min_gap_length,
		       ALIGNANDUM_POSITION row_length ) {

  
  /* use a sliding window of size window_size, and count 
     the following indicators for the central residue:
     
     num_left_gaps: number of gaps on left hand side of first
     residue in window
     num_right_gaps: number of gaps on right hand side of last
     residue in window
     
     left_pos: position of leftmost residue in window
     right_pos: position of rightmost residue in window

     delete a residue:
     if	num_left_gaps  larger than min_gap_length
     and num_right_gaps larger than min_gap_length
     and num_residues
     
  */
  
  if (row_length == 0) 
    row_length = dest->getRowTo();
  
  for (ALIGNANDUM_POSITION this_pos = dest->getRowFrom(); this_pos <= dest->getColTo(); this_pos ++) {
	
    // adjust left position
    ALIGNANDUM_POSITION left_pos = this_pos - window_size;
    while (dest->mapRowToCol( left_pos ) == 0) left_pos ++;
    
    // adjust right position
    ALIGNANDUM_POSITION right_pos = this_pos + window_size;
    while (dest->mapRowToCol( right_pos ) == 0) right_pos --;
    
    //--------------------------------------------------------------------------
    // count number of gaps on left and right side. This can probably be
    // made more efficient.
    // at the ends calculate number of gaps until end of row.
    unsigned int num_left_gaps = window_size - (this_pos - left_pos);

    if (left_pos == dest->getRowFrom()) {
      num_left_gaps = left_pos - 1;
    } else {
      ALIGNANDUM_POSITION x = left_pos;
      while (x > dest->getRowFrom() && dest->mapRowToCol( --x ) == 0) num_left_gaps ++;
    }

    unsigned int num_right_gaps = window_size - (right_pos - this_pos);

    if (right_pos == dest->getRowTo()) {
      num_right_gaps = row_length - right_pos;
    } else {
      ALIGNANDUM_POSITION x = right_pos;
      while (x < dest->getRowTo() && dest->mapRowToCol( ++x ) == 0) num_right_gaps ++;    
    }

    // std::cout << "center=" << this_pos << " left=" << left_pos << " gl=" << num_left_gaps 
    // << " right=" << right_pos << " gr=" << num_right_gaps << endl;

    //------------------------------------------------------------------------
    // check if region is to be deleted
    if ( (num_left_gaps > min_gap_length) &&
	 (num_right_gaps > min_gap_length) ) {
      dest->removeRowRegion( left_pos, right_pos );
      this_pos += window_size;
    }
    
    // go to next non-gap position
    while (this_pos <= dest->getColTo() && dest->mapRowToCol(this_pos) == 0) this_pos++;
    
  }      
  
}

//-----------------------------------------------------------------------------------------------
/** starting from the end of an alignment, remove
    residues as long as the score increases when these
    residues are removed.
*/
void pruneAlignata( Alignata * src, 
		    CALCTYPE gop,
		    CALCTYPE gep ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "Alignata * pruneAlignata( Alignata *) " << endl;
#endif
  
  //-----------------------------------------------------------
  // remove starting from left
  {
    AlignataIterator it(src->begin());
    AlignataIterator it_end(src->end());

    const RESIDUEPAIR & p = *it;
  
    CALCTYPE score = -p.mScore;
    ALIGNANDUM_POSITION last_row = src->getRowFrom();
    ALIGNANDUM_POSITION last_col = src->getColFrom();

    ++it;

    for (; it != it_end, score > 0; ++it) {

      const RESIDUEPAIR & p = *it;    
      // apply filter
      ALIGNANDUM_POSITION this_row      = p.mRow;
      ALIGNANDUM_POSITION this_col      = p.mCol;
      ALIGNANDUM_POSITION d;

      if ( (d = this_row - last_row - 1) > 0) 
	score -= gop + d * gep;
      if ( (d = this_col - last_col - 1) > 0) 
	score -= gop + d * gep;

      score -= p.mScore;

      last_row = this_row;
      last_col = this_col;
    }
  
    if (--last_row >= src->getRowFrom()) 
      src->removeRowRegion( src->getRowFrom(), last_row );
    if (--last_col >= src->getColFrom()) 
      src->removeColRegion( src->getColFrom(), last_col );
  }

  //-----------------------------------------------------------
  // remove starting from right. Ideally you would use reverse
  // iterators, but as I have not implemented them yet, I use this 
  // patch.
  {

    ALIGNANDUM_POSITION last_row = src->getRowTo();
    ALIGNANDUM_POSITION last_col = src->getColTo();    
    const RESIDUEPAIR & p = src->getPair( last_row);      
    CALCTYPE score = -p.mScore;

    ALIGNANDUM_POSITION this_row = last_row - 1;
    
    for (; last_row >= src->getRowFrom(), score > 0; --this_row) {

      ALIGNANDUM_POSITION this_col = src->mapRowToCol( this_row );
      
      if (!this_col)
	continue;

      const RESIDUEPAIR & p = src->getPair( this_row);      

      // apply filter

      ALIGNANDUM_POSITION d;

      if ( (d = last_row - this_row - 1) > 0) 
	score -= gop + d * gep;
      if ( (d = last_col - this_col - 1) > 0) 
	score -= gop + d * gep;

      score -= p.mScore;

      last_row = this_row;
      last_col = this_col;
    }
  
    if (++last_row <= src->getRowTo()) 
      src->removeRowRegion( last_row, src->getRowTo() );
    if (++last_col <= src->getColTo()) 
      src->removeColRegion( last_col, src->getColTo() );
  }    

}



} // namespace AlignLib

