/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <iostream>
#include <string>
#include <stdio.h>
#include <strstream>
#include "alignlib.h"

#include "Alignandum.h"
#include "AlignataIterator.h"
#include "Alignata.h"
#include "AlignException.h"

#include "HelpersAlignata.h"
#include "Translator.h"
#include "HelpersTranslator.h"
#include "ImplTranslator.h"

#include "Alignatum.h"
#include "HelpersAlignatum.h"

#include "SubstitutionMatrix.h"
#include "HelpersSubstitutionMatrix.h"

#include "ImplSequence.h"
#include "ImplProfile.h"

#include "AlignException.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

  // a class that calculates the match-score between a row and column.

    class Match {

	typedef CALCTYPE (Match::*MATCH_FUNCTION_POINTER)( ALIGNANDUM_POSITION, ALIGNANDUM_POSITION); 
    public:
	
	Match( const Alignandum * row, const Alignandum * col, const SubstitutionMatrix * matrix = NULL) {
	    
	    if (matrix)
		mSubstitutionMatrix = matrix;
	    else 
		mSubstitutionMatrix = getDefaultSubstitutionMatrix();
	    
	    //-------------------------------------------------------------------------------------------
	    // try to typecast row and col
	    // const SequenceTCO & t1 = dynamic_cast<const SequenceTCO&>(row);
	    const ImplSequence * s1 = dynamic_cast<const ImplSequence*>(row);
	    const ImplProfile * p1 = dynamic_cast<const ImplProfile*>(row);
	    
	    // const SequenceTCO & t2 = dynamic_cast<const SequenceTCO&>(col);
	    const ImplSequence * s2 = dynamic_cast<const ImplSequence*>(col);
	    const ImplProfile * p2 = dynamic_cast<const ImplProfile*>(col);
	    
	    // setup static pointers to the data locations
	    if (s1) {
		mRowSequence = s1->getData().mSequencePointer;
	    } else if (p1) {
	        mRowProfile     = p1->getData().mProfilePointer;
		mRowFrequencies = p1->getData().mFrequenciesPointer;
	    }
	    
	    if (s2) {
		mColSequence = s2->getData().mSequencePointer;
	    } else if (p2) {
		mColProfile     = p2->getData().mProfilePointer;
		mColFrequencies = p2->getData().mFrequenciesPointer;
	    }
	    
	    //!! to do: add TCO
	    if (s1 && s2) {
		mMatchFunction = &Match::matchSequenceSequence;
	    }
	    
	    if (p1 && p2) {
		mMatchFunction = &Match::matchProfileProfile;
	    }

	    if (s1 && p2) {
	      mMatchFunction = &Match::matchSequenceProfile;
	    }

	    if (s2 && p1) {
	      mMatchFunction = &Match::matchProfileSequence;
	    }
	    
	    // throw exception if you can not handle row/col combination
	    //!! to be implemented
	    
	    if (!row->isPrepared() )
		throw AlignException( "Row has not been prepared in HelpersAlignataPrivate::Match!");

	    if (!col->isPrepared() ) 
		throw AlignException( "Col has not been prepared in HelpersAlignataPrivate::Match!");
	}

	/** call matchfunction for row and col */
	CALCTYPE operator()(ALIGNANDUM_POSITION row, ALIGNANDUM_POSITION col) {
	    return (this->*mMatchFunction)( row, col );
	}

    private:
	//------------------------------------< match functions >--------------------------------------------
	CALCTYPE matchSequenceSequence( ALIGNANDUM_POSITION row, ALIGNANDUM_POSITION col ) {
	    return (mSubstitutionMatrix->getScore(mRowSequence[row],mColSequence[col]));
	}
	
	CALCTYPE matchProfileSequence( ALIGNANDUM_POSITION row, ALIGNANDUM_POSITION col ) {
	  if (mColSequence[col] == CODE_MASK)
	    return MASK_VALUE;
	  else
	    return mRowProfile[row][mColSequence[col]];
	}
	
	CALCTYPE matchSequenceProfile( ALIGNANDUM_POSITION row, ALIGNANDUM_POSITION col ) {
	  if (mRowSequence[row] == CODE_MASK)
	    return MASK_VALUE;
	  else
	    return mColProfile[col][mRowSequence[row]];
	}
	
	CALCTYPE matchProfileProfile( ALIGNANDUM_POSITION row, ALIGNANDUM_POSITION col ) {
	    CALCTYPE score = 0;
	    for (ALIGNANDUM_POSITION i = 0; i < PROFILEWIDTH; i++ ) 
		score += mRowProfile[row][i] * mColFrequencies[col][i] + 
		         mColProfile[col][i] * mRowFrequencies[row][i];
	    return score;
	}

	/** pointer to function that calculates match-score */
	MATCH_FUNCTION_POINTER mMatchFunction;
	
	/** pointer to member data of row/col : AlignandumSequence */
	const RESIDUE * mRowSequence; 
	const RESIDUE * mColSequence;
	
	/** pointer to member data of row/col AlignandumProfile */
	const TYPE_PROFILE_COLUMN * mRowProfile;
	const TYPE_PROFILE_COLUMN * mColProfile;
	const TYPE_FREQUENCIES_COLUMN * mRowFrequencies;
	const TYPE_FREQUENCIES_COLUMN * mColFrequencies;
	
	/** pointer to substitution matrix to use */
	const SubstitutionMatrix * mSubstitutionMatrix;

    };


//----------------------------------------------------------------------------------------------------
Alignata * rescoreAlignmentPrivate( Alignata * dest,
				    const Alignandum * row,
				    const Alignandum * col, 
				    const SubstitutionMatrix * matrix) {
#ifdef DEBUG
  if (verbose > LL3) cout << "rescoreAlignment(Alignata, const Alignandum *, const Alignandum *) " << endl; 
#endif
  
  Match match( row, col, matrix);

  AlignataIterator it(dest->begin());
  AlignataIterator it_end(dest->end());

  for (; it != it_end; ++it) 
      it->mScore = match( it->mRow, it->mCol );

  return dest;
}


} // namespace AlignLib
