/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <iostream>
#include <string>
#include <stdio.h>
#include <strstream>
#include "ImplAlignata.h"
#include "Alignandum.h"
#include "Alignatum.h"
#include "alignlib.h"
#include "AlignataIterator.h"
#include "Alignata.h"
#include "AlignException.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

//--------------------------------------------------------------------------------------------------------------
// constructors and desctructors
//--------------------------------------------------------------------------------------------------------------
ImplAlignata::ImplAlignata() : mChangedLength(true), mLength(0), mScore(0), mNumGaps( 0 ) {
}

ImplAlignata::ImplAlignata( const ImplAlignata & src ) :
  mChangedLength(src.mChangedLength), 
  mLength( src.mLength), 
  mScore (src.mScore), 
  mNumGaps (src.mNumGaps) { 
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignata::ImplAlignata( const ImplAlignata& ) " << endl;
#endif
}

ImplAlignata::~ImplAlignata() {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignata::~ImplAlignata() " << endl;
#endif
}

//--------------------------------------------------------------------------------------------------------------
void ImplAlignata::setChangedLength() {
  mChangedLength = true;
}
//--------------------------------------------------------------------------------------------------------------
void ImplAlignata::setScore( CALCTYPE score ) {
    mScore = score;
}
//--------------------------------------------------------------------------------------------------------------
CALCTYPE ImplAlignata::getScore() const {
    return mScore;
}
//--------------------------------------------------------------------------------------------------------------
ALIGNANDUM_POSITION ImplAlignata::getLength() const {

  if (mChangedLength)
    calculateLength();

  return mLength;
}
//--------------------------------------------------------------------------------------------------------------
void ImplAlignata::Clear() {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplAlignata::Clear()" << endl;
#endif
    mChangedLength = false;
    mLength = 0;
    mScore = 0;
    mNumGaps = 0;
}

//--------------------------------------------------------------------------------------------------------------
ALIGNANDUM_POSITION ImplAlignata::getNumGaps() const {
  if (mChangedLength)
    calculateLength();

  return mNumGaps;
}
//--------------------------------------------------------------------------------------------------------------
bool ImplAlignata::isEmpty() const {
    return (getLength() == 0);
}

//-----------------------------------------------------------------------------------------------------------   
void ImplAlignata::Write( std::ostream& output ) const {
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignata::Write( std::ostream&) const" << endl;
#endif

    output << "Length: " << getLength() << "\tScore: " << getScore() << "\tGaps: " << getNumGaps() << endl;
    output << "Row\tColumn\tScore\t" << endl;

    AlignataIterator_const it(begin());
    AlignataIterator_const it_end(end());
    
    for (; it != it_end; ++it)
      output << *it << endl;

}

//--------------------------------------------------------------------------------------------------------------
/** Read data members from stream */
void ImplAlignata::Read( std::istream & input) {
};

//--------------------------------------------------------------------------------------------------------------
void ImplAlignata::moveAlignment( ALIGNANDUM_POSITION row_offset, ALIGNANDUM_POSITION col_offset) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignata::moveAlignment( ALIGNANDUM_POSITION, ALIGNANDUM_POSITION) const" << endl;
#endif
    
  AlignataIterator it     = begin();
  AlignataIterator it_end = end();
  
  for (;it != it_end; ++it) {
    RESIDUEPAIR & p = (*it);
    p.mRow += row_offset;
    p.mCol += col_offset;
  }
  
};
   
//--------------------------------------------------------------------------------------------------------------
void ImplAlignata::setLength( ALIGNANDUM_POSITION length ) const {
    mLength = length;
}

//--------------------------------------------------------------------------------------------------------------
void ImplAlignata::setNumGaps( ALIGNANDUM_POSITION num_gaps ) const {
    mNumGaps = num_gaps;
}

//--------------------------------------------------------------------------------------------------------------
void ImplAlignata::addPair( ALIGNANDUM_POSITION row, ALIGNANDUM_POSITION col, CALCTYPE score ) { 
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignata::addPair()" << endl;
#endif
  addPair(new RESIDUEPAIR( row,col,score) );
} 

//--------------------------------------------------------------------------------------------------------------
void ImplAlignata::calculateLength() const {
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignata::calculateLength() const" << endl;
#endif

  ALIGNANDUM_POSITION row_last = getRowFrom();
  ALIGNANDUM_POSITION col_last = getColFrom();
  ALIGNANDUM_POSITION row_cur, col_cur;

  mLength = 0;
  mNumGaps = 0;
  mChangedLength = false;

  // this is a patch, there is something wrong with empty alignments,
  // for vectors (and sets)
  
  if (row_last == 0 || col_last == 0)
      return;

  AlignataIterator_const it(begin());
  AlignataIterator_const it_end(end());

  ALIGNANDUM_POSITION d;
  for (; it != it_end; ++it) {
      RESIDUEPAIR p(*it);
  
      row_cur = p.mRow;
      col_cur = p.mCol;
    
      mLength++;
      if ( (d = row_cur - row_last - 1) > 0 ) {
	  mLength += d; mNumGaps += d;
      }

      if ( (d = col_cur - col_last - 1) > 0 ) {
	  mLength += d; mNumGaps += d;
      }
    
      row_last = row_cur;
      col_last = col_cur;
  }

}  

//-----------------------------------------------------------------------------------------------------------   
/** switch row and column in the alignment. Use more efficient implementations in derived classes. */
void ImplAlignata::switchRowCol() {


  AlignataIterator it     = begin();
  AlignataIterator it_end = end();
  
  for (;it != it_end; ++it) {
    RESIDUEPAIR & p = (*it);
    ALIGNANDUM_POSITION temp = p.mCol;
    p.mCol = p.mRow;
    p.mRow = temp;
  }

}

//----------------------------------------------------------------------------------------------------------
// this is very time inefficient
//-----------------------------------------------------------------------------------------------------------   
//--------------> mapping functions <----------------------------------------------------------------------------
  /* this is horribly time-inefficient, if you want to map the whole thing. */
ALIGNANDUM_POSITION ImplAlignata::mapRowToCol( ALIGNANDUM_POSITION pos ) const {
  
  if (getLength() == 0)
    return 0;

  AlignataIterator_const it = begin();
  AlignataIterator_const it_end = end();
  
  while (it != it_end) {
    if ((*it).mRow == pos)
      return (*it).mCol;
    ++it;
  }
  
  return 0;
}

//-----------------------------------------------------------------------------------------------------------   
/* this is horribly time-inefficient, if you want to map the whole thing. Better use array */
ALIGNANDUM_POSITION ImplAlignata::mapColToRow( ALIGNANDUM_POSITION pos ) const {

  if (getLength() == 0)
    return 0;

  AlignataIterator_const it = begin();
  AlignataIterator_const it_end = end();
  
  while (it != it_end) {
    if ((*it).mCol == pos)
      return (*it).mRow;
    ++it;
  }
  
  return 0;
}

//-----------------------------------------------------------------------------------------------------------   
/** This is a generic routine. It creates a new alignment by making a copy of the old one.
*/
void ImplAlignata::removeRowRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) {

  const Alignata * copy = getClone();  
  
  AlignataIterator_const it     = copy->begin();
  AlignataIterator_const it_end = copy->end();

  Clear();

  mScore = copy->getScore();

  for (; it != it_end; ++it) {
    if ( (*it).mRow < from || (*it).mRow > to)
      addPair( new RESIDUEPAIR(*it) );
  }
  
  delete copy;

  setChangedLength();
  return;
}

//-----------------------------------------------------------------------------------------------------------   
/** This is a generic routine. It creates a new alignment by making a copy of the old one.
*/
void ImplAlignata::removeColRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) {

  const Alignata * copy = getClone();  
  
  AlignataIterator_const it     = copy->begin();
  AlignataIterator_const it_end = copy->end();

  Clear();

  mScore = copy->getScore();
  
  for (; it != it_end; ++it) {
    const RESIDUEPAIR & p = (*it);
    if (p.mCol < from || p.mCol > to)
      addPair( new RESIDUEPAIR(p) );
  }
  
  delete copy;

  setChangedLength();
  return;
}


// //--------------------------------------------------------------------------------------------------------------



} // namespace AlignLib
