/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef IMPL_ALIGNATA_H
#define IMPL_ALIGNATA_H 1

#include <iosfwd>
#include "alignlib.h"
#include "Alignata.h"

namespace AlignLib {

    /** basic implementation for aligned objects. 
	
	An alignment is a mapping.
	
	@author Andreas Heger
	@version $Id: ImplAlignata.h,v 1.1.1.1 2002/07/01 17:15:20 heger Exp $
	@short protocol class for aligned objects
    */
    
class ImplAlignata : public Alignata{

 public:
    //------------------> constructors / destructors <---------------------------------------------------------
    /** empty constructor */
    ImplAlignata();

    /** copy constructor */
    ImplAlignata( const ImplAlignata &src );

    /** destructor */
    virtual ~ImplAlignata();
    
    //----------------> accessors <------------------------------------------------------------------------------
    /** get the score of an alignment */
    virtual CALCTYPE getScore() const;

    /** get the length of an alignment */
    virtual ALIGNANDUM_POSITION	getLength() const;

    /** get the number of gaps in the alignment */
    virtual ALIGNANDUM_POSITION getNumGaps() const;

    /** set the score of an alignment */
    virtual void setScore( CALCTYPE score );

    /** returns true, if alignment is empty */
    bool isEmpty() const;

    /** switch row and column in the alignment */
    virtual void switchRowCol();

    /** move alignment */
    virtual void moveAlignment( ALIGNANDUM_POSITION row_offset, ALIGNANDUM_POSITION col_offset);

    /** maps a residue from row to column. returns 0, if not found. This default, but working implementation is
	very time-inefficient, especially, if you want to map the whole thing. Other implementations in derived
	classes can be much faster */
    virtual ALIGNANDUM_POSITION mapRowToCol( ALIGNANDUM_POSITION pos ) const;

    /** maps a residue from column to row. returns 0, if not found. This default, but working implementation is
	very time-inefficient, especially, if you want to map the whole thing. Other implementations in derived
	classes can be much faster*/
    virtual ALIGNANDUM_POSITION mapColToRow( ALIGNANDUM_POSITION pos ) const;

    /** removes a row-region in an alignment */
    virtual void removeRowRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to );

    /** removes a column-region in an alignment */
    virtual void removeColRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to);

    /** clear the current alignemnt */
    virtual void Clear() ;

    /** adds a pair of residues to the alignment 
	(have to add this here, otherwise it won't compile!. It seems that overloaded
	functions can not be separately implemented) */
    virtual void addPair( RESIDUEPAIR * new_pair ) = 0; 

    /** adds a pair of residues to the alignment */
    virtual void addPair( ALIGNANDUM_POSITION row, ALIGNANDUM_POSITION col, CALCTYPE score = 0); 
    
    /*-----------------> I/O <------------------------------------------------------------------------------ */
    /** Write data members to stream */
    virtual void Write(std::ostream & output ) const;	       

    /** Read data members from stream */
    virtual void Read( std::istream & input);

 protected:
    /** the length has changed of the alignment */
    virtual void setChangedLength();

    /** set length of the alignment */
    virtual void setLength( ALIGNANDUM_POSITION length ) const;

    /** set length of the alignment */
    virtual void setNumGaps( ALIGNANDUM_POSITION num_gaps ) const;

    /** calculate alignment length */
    virtual void calculateLength() const;

    /** flag, whether alignment length has changed. Protected, because some childs need to query and 
     set this flag.
    */
    mutable bool mChangedLength;

 private:

    /** length of alignemnt, has to be mutable, since length evaluation is lazy */
    mutable ALIGNANDUM_POSITION mLength;
    
    /** total score of alignment */
    CALCTYPE mScore;
    
    /** number of gaps in alignment */
    mutable ALIGNANDUM_POSITION mNumGaps;

    
};
						  

}

#endif /* IMPL_ALIGNATA_H */

