/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream> 
#include <algorithm>
#include <strstream>
#include <set>
#include "alignlib.h"
#include "ImplAlignataHash.h"
#include "AlignataIterator.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

//------------------------------factory functions -----------------------------
Alignata * makeAlignataHash() {
  return new ImplAlignataHash();
}

//------------------------------------< constructors and destructors >-----
ImplAlignataHash::ImplAlignataHash() : ImplAlignata() {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataHash::ImplAlignataHash()" << endl;
#endif
}

ImplAlignataHash::ImplAlignataHash( const ImplAlignataHash& src) : ImplAlignata( src ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataHash::ImplAlignataHash( const ImplAlignataHash & )" << endl;
#endif

    mPairs.clear();

    PairIterator it(src.mPairs.begin()), it_end(src.mPairs.end());

    for (;it != it_end; ++it) 
      mPairs.insert( new RESIDUEPAIR(**it) );
}

ImplAlignataHash::~ImplAlignataHash( ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataHash::~ImplAlignataHash()" << endl;
#endif
    Clear();
}

//------------------------------------------------------------------------------------------------------------
ImplAlignataHash * ImplAlignataHash::getNew() const {
  return new ImplAlignataHash();
};
    
ImplAlignataHash * ImplAlignataHash::getClone() const {
  return new ImplAlignataHash( *this );
}

void ImplAlignataHash::Clear() { 
  ImplAlignata::Clear();

  PAIRSET::iterator it(mPairs.begin()), it_end(mPairs.end());
  for (;it != it_end; ++it) delete *it;
  mPairs.clear(); 

}

//-----------------------------------------------------------------------------------------------------------   

AlignataIterator_const ImplAlignataHash::begin() const { 
  return AlignataIterator_const( new ImplAlignataHash_Iterator_const(mPairs.begin()));
}

AlignataIterator_const ImplAlignataHash::end()   const { 
  return AlignataIterator_const( new ImplAlignataHash_Iterator_const(mPairs.end())); 
}

AlignataIterator ImplAlignataHash::begin() { 
  return AlignataIterator( new ImplAlignataHash_Iterator(mPairs.begin()));
}

AlignataIterator ImplAlignataHash::end() { 
  return AlignataIterator( new ImplAlignataHash_Iterator(mPairs.end())); 
}


//----------------> accessors <------------------------------------------------------------------------------

ALIGNANDUM_POSITION ImplAlignataHash::getRowFrom() const { return (mPairs.size() > 0) ? (*mPairs.begin())->mRow : 0; }
ALIGNANDUM_POSITION ImplAlignataHash::getColFrom() const { return (mPairs.size() > 0) ? (*mPairs.begin())->mCol : 0; }
ALIGNANDUM_POSITION ImplAlignataHash::getRowTo()   const { return (mPairs.size() > 0) ? (*mPairs.rbegin())->mRow : 0; }
ALIGNANDUM_POSITION ImplAlignataHash::getColTo()   const { return (mPairs.size() > 0) ? (*mPairs.rbegin())->mCol : 0; }
  
RESIDUEPAIR ImplAlignataHash::getFirstPair() const { return **(mPairs.begin()); }
RESIDUEPAIR ImplAlignataHash::getLastPair()  const { return **(mPairs.rbegin()); }

void ImplAlignataHash::addPair( RESIDUEPAIR * new_pair ) { 
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignataHash::addPair( RESIDUEPAIR &)" << endl;
#endif

  if (mPairs.find( new_pair) != mPairs.end()) {
    delete new_pair;
  } else {
    setChangedLength(); 
    mPairs.insert( new_pair ); 
  } 
} 

void ImplAlignataHash::removePair( const RESIDUEPAIR & old_pair ) { 
#ifdef DEBUG
  if (verbose > LL5) cout << "void ImplAlignataHash::removePair( RESIDUEPAIR &)" << endl;
#endif

  setChangedLength(); 
  RESIDUEPAIR p(old_pair);
  mPairs.erase(&p); 
} 

//----------------------------------------------------------------------------------------------------------
/** retrieves a pair of residues from the alignment */
RESIDUEPAIR ImplAlignataHash::getPair( ALIGNANDUM_POSITION row ) const {

  RESIDUEPAIR p(row, 0, 0);
  PairIterator it(mPairs.find( &p ));
  return **it;
} 


} // namespace AlignLib
