/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef IMPL_ALIGNATA_HASH_H
#define IMPL_ALIGNATA_HASH_H 1

#include <iosfwd>
#include "alignlib.h"
#include "ImplAlignata.h"

namespace AlignLib {

    /** class for aligned objects. 
	
	Residues are kept in a set, but in contrast to AlignataSet, both row and
	column are taken into account, when comparing keys.

	The code is basically identical to ImplAlignataSet, the comparator has 
	changed. Maybe in the future this can be made nicer by subclassing or
	template instantiation.
	
	@author Andreas Heger
	@version $Id: ImplAlignataHash.h,v 1.1.1.1 2002/07/01 17:15:20 heger Exp $
	@short protocol class for aligned objects
    */



class Alignandum;

class ImplAlignataHash : public ImplAlignata {

  /** residues are sorted by row and then by column */
  struct Comparator {
      bool operator() ( const RESIDUEPAIR * x, const RESIDUEPAIR * y) const { 
	  if (x->mRow < y->mRow) return 1;
	  if (x->mRow > y->mRow) return 0;
	  if (x->mCol < y->mCol) 
	      return 1;
	  else
	      return 0;
      }
  };  
  typedef std::set<RESIDUEPAIR *, Comparator> PAIRSET;
  typedef PAIRSET::const_iterator PairIterator_const;
  typedef PAIRSET::iterator PairIterator;
  
 public:
    //------------------> constructors / destructors <---------------------------------------------------------
    /** constructor */
    ImplAlignataHash();

    /** copy constructor */
    ImplAlignataHash( const ImplAlignataHash &src );

    /** destructor */
    virtual ~ImplAlignataHash();

    //------------------------------------------------------------------------------------------------------------
    virtual ImplAlignataHash * getNew() const;
    
    /** return an identical copy */
    virtual ImplAlignataHash * getClone() const;

    //------------------------------------------------------------------------------------------------------------
    //------------------------------------------------------------------------------------------------------------
    class ImplAlignataHash_Iterator_const : public Alignata::Iterator_const {
    public:
        ImplAlignataHash_Iterator_const(PairIterator_const it) : mIterator( it ) {};
	
        ImplAlignataHash_Iterator_const( const ImplAlignataHash_Iterator_const & src ) : mIterator( src.mIterator) {};
	
        virtual ~ImplAlignataHash_Iterator_const() {};
	
  	virtual ImplAlignataHash_Iterator_const * getClone() const {return new ImplAlignataHash_Iterator_const( mIterator);}

        /** dereference operator */
        virtual const RESIDUEPAIR & getReference() const { return *(*mIterator);}

	/** for indirection */
        virtual const RESIDUEPAIR * getPointer() const { return (*mIterator); }

        /** advance one position */
        virtual void Next() { mIterator++;}
	
        /** step back one position */
	virtual void Previous() { mIterator--;}
    private:
	PairIterator_const mIterator;
    };

    //------------------------------------------------------------------------------------------------------------
    class ImplAlignataHash_Iterator : public Alignata::Iterator {
    public:
        ImplAlignataHash_Iterator(PairIterator it) : mIterator( it ) {};
	
        ImplAlignataHash_Iterator( const ImplAlignataHash_Iterator & src ) : mIterator( src.mIterator) {};
	
        virtual ~ImplAlignataHash_Iterator() {};
	
  	virtual ImplAlignataHash_Iterator * getClone() const {return new ImplAlignataHash_Iterator( mIterator);}

        /** dereference operator
	    The solution here compiles at least. It is ugly and inefficient. However, when returning *mIterator, you get the following error:
	    "AlignLib::RESIDUEPAIR &" to an initializer of type "const
	    std::_Rb_tree<std::set<AlignLib::RESIDUEPAIR,               
	    ...
	 */
        virtual RESIDUEPAIR & getReference() const { return *(*mIterator);}

	/** for indirection 
	    The solution here compiles at least. When returning return (mIterator.operator->()), the const-type is not correctly
	    set.
	 */
        virtual RESIDUEPAIR * getPointer() const { return *mIterator; }	

        /** advance one position */
        virtual void Next() { mIterator++; }
	    
        /** step back one position */
	virtual void Previous() { mIterator--;}
    private:
	PairIterator mIterator;
    };

    /** return const iterator */
    virtual AlignataIterator_const begin() const; 
    
    /** return const iterator */
    virtual AlignataIterator_const end() const; 

    /** return const iterator */
    virtual AlignataIterator begin(); 
    
    /** return const iterator */
    virtual AlignataIterator end(); 

    //----------------> accessors <------------------------------------------------------------------------------

    /** returns the first residue aligned in row */
    virtual ALIGNANDUM_POSITION getRowFrom() const;

    /** returns the last residue aligned in row */
    virtual ALIGNANDUM_POSITION	getRowTo() const;

    /** returns the first residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColFrom() const;

    /** returns the last residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColTo()   const;

    /** returns the first aligned pair. Have to create a copy not a reference, because not all alignments will have
	a list of pairs ready */
    virtual RESIDUEPAIR getFirstPair() const;
    
    /** returns the last aligned pair */
    virtual RESIDUEPAIR getLastPair() const;

    /** adds a pair of residue to the alignment */
    virtual void addPair( RESIDUEPAIR * new_pair ); 

    /** retrieves a pair of residues from the alignment */
    virtual RESIDUEPAIR getPair( ALIGNANDUM_POSITION row ) const;

    /** removes a pair of residues from the alignment */
    virtual void removePair( const RESIDUEPAIR & old_pair );

    /** clear the current alignemnt */
    void Clear();

 private:
     PAIRSET mPairs;
};

						  

}

#endif /* _ALIGNATA_H */

