/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream> 
#include <algorithm>
#include <strstream>
#include <set>
#include "alignlib.h"
#include "ImplAlignataHashDiagonal.h"
#include "AlignataIterator.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

//------------------------------factory functions -----------------------------
Alignata * makeAlignataHashDiagonal() {
  return new ImplAlignataHashDiagonal();
}

//------------------------------------< constructors and destructors >-----
ImplAlignataHashDiagonal::ImplAlignataHashDiagonal() : ImplAlignata() {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataHashDiagonal::ImplAlignataHashDiagonal()" << endl;
#endif
}

ImplAlignataHashDiagonal::ImplAlignataHashDiagonal( const ImplAlignataHashDiagonal& src) : ImplAlignata( src ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataHashDiagonal::ImplAlignataHashDiagonal( const ImplAlignataHashDiagonal & )" << endl;
#endif

    clearContainer();

    PairIterator it(src.mPairs.begin()), it_end(src.mPairs.end());

    for (;it != it_end; ++it) 
      mPairs.insert( new RESIDUEPAIR(**it) );
}

ImplAlignataHashDiagonal::~ImplAlignataHashDiagonal( ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataHashDiagonal::~ImplAlignataHashDiagonal()" << endl;
#endif
    Clear();
}

//------------------------------------------------------------------------------------------------------------
ImplAlignataHashDiagonal * ImplAlignataHashDiagonal::getNew() const {
    return new ImplAlignataHashDiagonal();
};
    
ImplAlignataHashDiagonal * ImplAlignataHashDiagonal::getClone() const {
    return new ImplAlignataHashDiagonal( *this );
}

//-------------------------------------------------------------------
void ImplAlignataHashDiagonal::clearContainer() { 
  PAIRSET::iterator it(mPairs.begin()), it_end(mPairs.end());
  for (;it != it_end; ++it) delete *it;
  mPairs.clear(); 
}

//-------------------------------------------------------------------
void ImplAlignataHashDiagonal::Clear() { 
  ImplAlignata::Clear();

  clearContainer();
}

//-----------------------------------------------------------------------------------------------------------   

AlignataIterator_const ImplAlignataHashDiagonal::begin() const { 
  return AlignataIterator_const( new ImplAlignataHashDiagonal_Iterator_const(mPairs.begin()));
}

AlignataIterator_const ImplAlignataHashDiagonal::end()   const { 
  return AlignataIterator_const( new ImplAlignataHashDiagonal_Iterator_const(mPairs.end())); 
}

AlignataIterator ImplAlignataHashDiagonal::begin() { 
  return AlignataIterator( new ImplAlignataHashDiagonal_Iterator(mPairs.begin()));
}

AlignataIterator ImplAlignataHashDiagonal::end() { 
  return AlignataIterator( new ImplAlignataHashDiagonal_Iterator(mPairs.end())); 
}


//----------------> accessors <------------------------------------------------------------------------------

ALIGNANDUM_POSITION ImplAlignataHashDiagonal::getRowFrom() const { if (mChangedLength) calculateLength(); return mRowFrom; }
ALIGNANDUM_POSITION ImplAlignataHashDiagonal::getColFrom() const { if (mChangedLength) calculateLength(); return mColFrom; }
ALIGNANDUM_POSITION ImplAlignataHashDiagonal::getRowTo()   const { if (mChangedLength) calculateLength(); return mRowTo; }
ALIGNANDUM_POSITION ImplAlignataHashDiagonal::getColTo()   const { if (mChangedLength) calculateLength(); return mColTo; }
  
RESIDUEPAIR ImplAlignataHashDiagonal::getFirstPair() const { return **(mPairs.begin()); }
RESIDUEPAIR ImplAlignataHashDiagonal::getLastPair()  const { return **(mPairs.rbegin()); }

void ImplAlignataHashDiagonal::addPair( RESIDUEPAIR * new_pair ) { 
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignataHashDiagonal::addPair( RESIDUEPAIR &)" << endl;
#endif

  if (mPairs.find( new_pair) != mPairs.end()) {
    delete new_pair;
  } else {
    setChangedLength(); 
    mPairs.insert( new_pair ); 
  } 
} 

//----------------------------------------------------------------------------------------------------------
/** retrieves a pair of residues from the alignment */
RESIDUEPAIR ImplAlignataHashDiagonal::getPair( ALIGNANDUM_POSITION row ) const {

  RESIDUEPAIR p(row, 0, 0);
  PairIterator it(mPairs.find( &p ));
  return **it;
} 

void ImplAlignataHashDiagonal::removePair( const RESIDUEPAIR & old_pair ) { 
#ifdef DEBUG
  if (verbose > LL5) cout << "void ImplAlignataHashDiagonal::removePair( RESIDUEPAIR &)" << endl;
#endif

  setChangedLength(); 
  RESIDUEPAIR p(old_pair);
  mPairs.erase(&p); 
} 

void ImplAlignataHashDiagonal::calculateLength() const {

  mChangedLength = false;

  mRowFrom = 9999999;
  mRowTo = 0;
  mColFrom = 9999999;
  mColTo = 0;

  PairIterator_const it(mPairs.begin()), it_end(mPairs.end());

  ALIGNANDUM_POSITION length = 0;
  
  for (; it != it_end; ++it) {
      ALIGNANDUM_POSITION row = (*it)->mRow;
      ALIGNANDUM_POSITION col = (*it)->mCol;

      // get maximum boundaries
      if (row < mRowFrom) mRowFrom = row;
      if (col < mColFrom) mColFrom = col;
      if (row > mRowTo)   mRowTo = row;
      if (col > mColTo)   mColTo = col;
      length++;
  }

  setNumGaps(0);
  setLength( length );
}

} // namespace AlignLib
