/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream> 
#include <algorithm>
#include <strstream>
#include <set>
#include "alignlib.h"
#include "ImplAlignataMatrix.h"
#include "AlignataIterator.h"
#include "AlignException.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

  /** Note: There are two arrays that have to be taken care of: mPairs and mIndex. If you
      need to copy one, you have to copy the other one as well. The same applies to deletion.

      Clear() does not free the memory for mPairs and mIndices. This class works by preallocating
      memory, otherwise you would need to realloc memory every time one or more pairs are added.
      Alternatively, you could implement this class as a stack, that dynamically allocates memory.
      
      If you need to safe memory, allocate only as much as you need by giving the correct number
      of pairs to the constructor, or alternatively, create a big class and then create a copy 
      of the object with the copy constructor.

      When modifying the dots, the allocated size of the memory stays always the same.

  */

#define NODOT -1 

//------------------------------------< constructors and destructors >-----
ImplAlignataMatrix::ImplAlignataMatrix( long ndots ) : ImplAlignata(), 
  mIndex(NULL),
  mRowFrom(0), mRowTo(0), mColFrom(0), mColTo(0), 
  mAllocatedIndexSize(0) {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplAlignataMatrix::ImplAlignataMatrix()" << endl;
#endif
}

    /*  It this important to add mPairs(NULL) to the list of initializer, because
	otherwise (at least gcc) adds an implicit mPairs(src.mPairs), which is 
	absolutely disastrous 
    */
ImplAlignataMatrix::ImplAlignataMatrix( const ImplAlignataMatrix & src) : 
  ImplAlignata( src ), 
  mPairs(),			
  mIndex(NULL),
  mRowFrom( src.mRowFrom), 
  mRowTo( src.mRowTo),
  mColFrom( src.mColFrom), 
  mColTo( src.mColTo),
  mAllocatedIndexSize( src.mAllocatedIndexSize) {

#ifdef DEBUG
  if (verbose > LL3) cout << "ImplAlignataMatrix::ImplAlignataMatrix( const ImplAlignataMatrix & )" << endl;
#endif
  
  // create a deep copy of src.mPairs
  PairIterator_const it(src.mPairs.begin()), it_end(src.mPairs.end());
  for (; it != it_end; ++it) {
    RESIDUEPAIR * p = *it;
    mPairs.push_back( new RESIDUEPAIR( *p ) );
  }

  // create a copy of the index (if existing)
  if (src.mIndex) {
    mIndex = new TYPE_DOT[mAllocatedIndexSize];
    memcpy( mIndex, src.mIndex, sizeof(TYPE_DOT) * (mAllocatedIndexSize));
  }
}

ImplAlignataMatrix::~ImplAlignataMatrix( ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataMatrix::~ImplAlignataMatrix()" << endl;
#endif
    Clear();

    // mPairs and mIndex is deleted only now

    delete [] mIndex; mIndex = NULL; 
}


//-----------------------------------------------------------------------------------------------------------   

AlignataIterator_const ImplAlignataMatrix::begin() const { 
    if (mChangedLength) calculateLength();
    return AlignataIterator_const( new ImplAlignataMatrix_Iterator_const( mPairs,0, mPairs.size() )); 
}

AlignataIterator_const ImplAlignataMatrix::end() const { 
    if (mChangedLength) calculateLength();
    return AlignataIterator_const( new ImplAlignataMatrix_Iterator_const(mPairs,mPairs.size(),mPairs.size() )); 
}
//-----------------------------------------------------------------------------------------------------------   

AlignataIterator ImplAlignataMatrix::begin() { 
  if (mChangedLength) calculateLength();
  return AlignataIterator( new ImplAlignataMatrix_Iterator( mPairs,0,mPairs.size() )); 
}

AlignataIterator ImplAlignataMatrix::end() { 
  if (mChangedLength) calculateLength();
  return AlignataIterator( new ImplAlignataMatrix_Iterator(mPairs,mPairs.size(),mPairs.size() )); 
}

//----------------> accessors <------------------------------------------------------------------------------

ALIGNANDUM_POSITION ImplAlignataMatrix::getRowFrom() const { if (mChangedLength) calculateLength(); return mRowFrom; }
ALIGNANDUM_POSITION ImplAlignataMatrix::getColFrom() const { if (mChangedLength) calculateLength(); return mColFrom; }
ALIGNANDUM_POSITION ImplAlignataMatrix::getRowTo()   const { if (mChangedLength) calculateLength(); return mRowTo; }
ALIGNANDUM_POSITION ImplAlignataMatrix::getColTo()   const { if (mChangedLength) calculateLength(); return mColTo; }

RESIDUEPAIR ImplAlignataMatrix::getFirstPair() const { 
    if (mChangedLength) calculateLength(); 
    return (mPairs.size() > 0) ? (*mPairs.front()) : RESIDUEPAIR(0,0,0); 
}

RESIDUEPAIR ImplAlignataMatrix::getLastPair()  const { 
  if (mChangedLength) calculateLength(); 
  return (mPairs.size() > 0) ? (*mPairs.back()) : RESIDUEPAIR(0,0,0); 
}

//-------------------------------------------------------------------------------------------------------------
void ImplAlignataMatrix::addPair( RESIDUEPAIR * new_pair ) { 
  
  mPairs.push_back( new_pair );
  setChangedLength();		// has to be resorted
} 
//-------------------------------------------------------------------------------------------------------------
void ImplAlignataMatrix::addPair( ALIGNANDUM_POSITION row, ALIGNANDUM_POSITION col, CALCTYPE score ) { 
  
  mPairs.push_back( new RESIDUEPAIR(row,col,score) );
  setChangedLength();		// has to be resorted
} 

//-------------------------------------------------------------------------------------------------------------
/** retrieves a pair of residues from the alignment */
RESIDUEPAIR ImplAlignataMatrix::getPair( ALIGNANDUM_POSITION row ) const {
  //!! to be implemented
  return *mPairs[row];
} 

//----------------------------------------------------------------------------------------------------------------------
void ImplAlignataMatrix::removePair( const RESIDUEPAIR & old_pair ) { 
  
  setChangedLength();
  //!! to be implemented.  move and split highly inefficient
  
} 

//--------------------------------------------------------------------------------------------------------
void ImplAlignataMatrix::Clear() { 
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplAlignataMatrix::Clear()" << endl;
#endif
  ImplAlignata::Clear();

  mRowFrom = 0;
  mRowTo = 0;
  mColFrom = 0;
  mRowTo = 0;
  delete mIndex; 
  mIndex = NULL;

  PAIRVECTOR::iterator it(mPairs.begin()), it_end(mPairs.end());
  for (;it != it_end; ++it) delete *it;
  mPairs.clear();
}

//--------------------------------------------------------------------------------------------------------------
void ImplAlignataMatrix::moveAlignment( ALIGNANDUM_POSITION row_offset, ALIGNANDUM_POSITION col_offset) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignataVector::moveAlignment( ALIGNANDUM_POSITION, ALIGNANDUM_POSITION) const" << endl;
#endif
  
  ImplAlignata::moveAlignment( row_offset, col_offset);

  // reset coordinates of alignment
  mRowFrom += row_offset;
  mRowTo   += row_offset;
  mColFrom += col_offset;
  mColTo   += col_offset;
  
}
//------------------------------------> sorting subroutines <-----------------------------------------------

//----------------------------------------------------------------------------------------------------------
void ImplAlignataMatrix::sortDotsByDiagonal(ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) const {
  ALIGNANDUM_POSITION lastsmall, c1, i;
  TYPE_DIAGONAL m; /* value of median */
  RESIDUEPAIR * t;

  if (from < to ) {
    c1 = (from + to) / 2;

    t = mPairs[from]; mPairs[from] = mPairs[c1]; mPairs[c1] = t;

    m = calculateDiagonal( *mPairs[from] );					// choose median-value to compare to
    lastsmall = from;
    for (i = from + 1; i <= to; i++) {
      if ( calculateDiagonal(*mPairs[i]) < m) {						// swap lastsmall and i
	lastsmall++;
	t = mPairs[lastsmall]; mPairs[lastsmall] = mPairs[i]; mPairs[i] = t;
      }
    }
      
    t = mPairs[from]; mPairs[from] = mPairs[lastsmall]; mPairs[lastsmall] = t;
    
    sortDotsByDiagonal( from, lastsmall);
    sortDotsByDiagonal( lastsmall+1, to);
  }
}

//----------------------------------------------------------------------------------------------------------------------
void ImplAlignataMatrix::sortDotsByRow(ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) const {
  ALIGNANDUM_POSITION lastsmall, c1, i;
  ALIGNANDUM_POSITION m; /* value of median */
  RESIDUEPAIR * t;
  
  if (from < to ) {
      c1 = (from + to) / 2;

      t = mPairs[from]; 
      mPairs[from] = mPairs[c1]; 
      mPairs[c1] = t;

      m = mPairs[from]->mRow;                                 // choose median-value to compare to
      lastsmall = from;
      for (i = from + 1; i <= to; i++) {
	  if ( mPairs[i]->mRow < m) {                         // swap lastsmall and i
	      lastsmall++;
	      t = mPairs[lastsmall]; 
	      mPairs[lastsmall] = mPairs[i]; 
	      mPairs[i] = t;
	  }
      }
      
      t = mPairs[from]; mPairs[from] = mPairs[lastsmall]; mPairs[lastsmall] = t;
    
      m = lastsmall;
      sortDotsByRow( from, m);
      sortDotsByRow( m+1, to);
  }
}

//----------------------------------------------------------------------------------------------------------------------
void ImplAlignataMatrix::sortDotsByCol(ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) const {
  ALIGNANDUM_POSITION lastsmall, c1, i;
  ALIGNANDUM_POSITION m; /* value of median */
  RESIDUEPAIR * t;

  if (from < to ) {
      c1 = (from + to) / 2;

      t = mPairs[from]; 
      mPairs[from] = mPairs[c1]; 
      mPairs[c1] = t;

      m = mPairs[from]->mCol;                                 // choose median-value to compare to
      lastsmall = from;
      for (i = from + 1; i <= to; i++) {
	  if ( mPairs[i]->mCol < m) {                         // swap lastsmall and i
	      lastsmall++;
	      t = mPairs[lastsmall]; 
	      mPairs[lastsmall] = mPairs[i]; 
	      mPairs[i] = t;
	  }
      }
      
      t = mPairs[from]; mPairs[from] = mPairs[lastsmall]; mPairs[lastsmall] = t;
    
      m = lastsmall;
      sortDotsByCol( from, m);
      sortDotsByCol( m+1, to);
  }
}


//-------------------------------------------------------------------------------------------------------------------- 
void ImplAlignataMatrix::eliminateDuplicates() const {
    mRowFrom = 999999;
    mColFrom = 999999;
    mRowTo = 0;
    mColTo = 0;

    ALIGNANDUM_POSITION last_row = 0;
    ALIGNANDUM_POSITION last_col = 0;
    
    std::vector<RESIDUEPAIR *> temp = mPairs;
    mPairs.clear();
    mPairs.reserve(temp.size());

    PairIterator_const it(temp.begin()), it_end(temp.end());
    
    for (; it != it_end; ++it) {
      
      ALIGNANDUM_POSITION row = (*it)->mRow;
      ALIGNANDUM_POSITION col = (*it)->mCol;

      /* delete pairs not needed any more */
      if (last_row == row && last_col == col) {
	delete (*it);
	continue;
      }

      // get maximum boundaries
      if (row < mRowFrom) mRowFrom = row;
      if (col < mColFrom) mColFrom = col;
      if (row > mRowTo)   mRowTo = row;
      if (col > mColTo)   mColTo = col;

      mPairs.push_back(*it);
    }
}


//--------------------------------------------------------------------------------------------------------------
void ImplAlignataMatrix::allocateIndex( unsigned long size ) const {
  if (mIndex) delete mIndex;

  mAllocatedIndexSize = size;
  mIndex = new TYPE_DOT[size];
}

//--------------------------------------------------------------------------------------------------------------
// instead of calculating the length, it resorts the dots :=)
void ImplAlignataMatrix::calculateLength() const {
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignataMatrix::calculateLength() const" << endl;
#endif

  mChangedLength = false;
  mRowFrom = mRowTo = mColFrom = mColTo = 0;
  delete mIndex;
  mAllocatedIndexSize = 0;
  mIndex = NULL;
  setNumGaps(0);
  setLength(mPairs.size());

  if (mPairs.empty()) 
    return;

  // 1. sort Dots
  sortDots();
  
  // 2. eliminate duplicates. At the same time, this sets mRowFrom, mRowTo, etc.
  eliminateDuplicates();
  
  // 3. build index for quick access (row, col, diagonal, etc.)
  buildIndex();

#ifdef DEBUG
  cout << "Finishing calculateLength()" << endl;
#endif

}  

} // namespace AlignLib
