/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef IMPL_ALIGNATA_MATRIX_H
#define IMPL_ALIGNATA_MATRIX_H 1

#include <iosfwd>
#include <vector>
#include "alignlib.h"
#include "ImplAlignata.h"

namespace AlignLib {

    /** class for aligned objects. 

	if alignment is empty, mRowFrom is larger than mRowTo. This property is needed
	at some checks (for example in ImplAlignataMatrix_iterator)

	Residues are kept in a list in memory. Note, that in this form of alignments,
	several column residues can be aligned to the same row residue and vice versa.

	Think of this Alignment as a dot-matrix. Actually, this is what it is used for
	in AlignatorDot. However, the full matrix is not stored memory, but as 
	a list of aligned residues (sorted first by row, then by column). The class provides
	and index for quick mapping of row to column (always the smallest col is returned).

	Note: some of the functions meaningful in simple pairwise alignment object have
	a consistent but slightly different sense here. For example the alignment length
	returns the number of dots might therefore be longer than the sequences. 

	mapRowToCol and mapColToRow give unpredictable results.

	@author Andreas Heger
	@version $Id: ImplAlignataMatrix.h,v 1.1.1.1 2002/07/01 17:15:20 heger Exp $
	@short protocol class for aligned objects
    */



class Alignandum;


//----------------------------------------------------------------------------------------------------------------------
inline TYPE_DIAGONAL calculateDiagonal( const RESIDUEPAIR & p) { 
  return (p.mCol - p.mRow); 
}
//----------------------------------------------------------------------------------------------------------------------
// calculates the diagonal, so that (1,1) is (row_from, col_from)
inline TYPE_DIAGONAL calculateNormalizedDiagonal( const RESIDUEPAIR & p, 
						  ALIGNANDUM_POSITION row_from, 
						  ALIGNANDUM_POSITION col_from) { 
  return ( (p.mCol - col_from +1 ) - (p.mRow - row_from + 1)); 
}


class ImplAlignataMatrix : public ImplAlignata {

    friend class Iterator_const;
    friend class ImplAlignatorDots;

 public:

    typedef std::vector<RESIDUEPAIR *> PAIRVECTOR;
    typedef PAIRVECTOR::iterator PairIterator;
    typedef PAIRVECTOR::const_iterator PairIterator_const;

    //------------------> constructors / destructors <---------------------------------------------------------
    /** constructor */
    ImplAlignataMatrix( long ndots = 0);
    
    /** copy constructor */
    ImplAlignataMatrix( const ImplAlignataMatrix &src );

    /** destructor */
    virtual ~ImplAlignataMatrix();

    //------------------------------------------------------------------------------------------------------------
    //------------------------------------------------------------------------------------------------------------
    // this iterator iterates over all dots. 
    //!! to be implemented: another iterator, that iterates over rows
    // valid values for index are 0 to max_index - 1
    class ImplAlignataMatrix_Iterator_const : public Alignata::Iterator_const{
    public:
	
      ImplAlignataMatrix_Iterator_const( const PAIRVECTOR & pairs, 
					 ALIGNANDUM_POSITION index, 
					 ALIGNANDUM_POSITION max_index ) :
	mPairs( pairs ), mCurrentIndex(index), mMaximumIndex( max_index) {
      };
	
      ImplAlignataMatrix_Iterator_const( const ImplAlignataMatrix_Iterator_const & src ) : 
	mPairs( src.mPairs ), mCurrentIndex( src.mCurrentIndex ), mMaximumIndex( src.mMaximumIndex ) {};
      
      virtual ~ImplAlignataMatrix_Iterator_const() {};
      
      virtual Iterator_const * getClone() const {return new ImplAlignataMatrix_Iterator_const( *this );}
      
      /** dereference operator: runtime error, if out of bounds? */
      virtual const RESIDUEPAIR & getReference() const { 
	return (*mPairs[mCurrentIndex]);
      }
      
      /** for indirection */
      virtual const RESIDUEPAIR * getPointer() const { 
	  if (mCurrentIndex >= 0 && mCurrentIndex < mMaximumIndex)
	    return mPairs[mCurrentIndex]; 
	  else 
	    return NULL;
      }
      
      /** advance one position, until you find an aligned pair */
      virtual void Next() { mCurrentIndex++; if (mCurrentIndex > mMaximumIndex) mCurrentIndex = mMaximumIndex; }
      
      /** step back one position, until you find an aligned pair */
      virtual void Previous() { mCurrentIndex--; if (mCurrentIndex < -1) mCurrentIndex = -1; }
	
    private:
      const PAIRVECTOR & mPairs;
      ALIGNANDUM_POSITION mCurrentIndex;
      ALIGNANDUM_POSITION mMaximumIndex;
      
    };

    class ImplAlignataMatrix_Iterator : public Alignata::Iterator {
    public:
	
      ImplAlignataMatrix_Iterator( PAIRVECTOR & pairs, 
				   ALIGNANDUM_POSITION index, 
				   ALIGNANDUM_POSITION max_index ) :
	mPairs( pairs ), mCurrentIndex(index), mMaximumIndex( max_index) {
	};
	
	ImplAlignataMatrix_Iterator( const ImplAlignataMatrix_Iterator & src ) : 
	  mPairs( src.mPairs ), mCurrentIndex( src.mCurrentIndex ), mMaximumIndex( src.mMaximumIndex ) {};
	
	 virtual ~ImplAlignataMatrix_Iterator() {};
      
	virtual Iterator * getClone() const {return new ImplAlignataMatrix_Iterator( *this );}
	
	/** dereference operator: runtime error, if out of bounds? */
	virtual RESIDUEPAIR & getReference() const { 
	  return (*mPairs[mCurrentIndex]);
	}
 
	/** for indirection */
	virtual RESIDUEPAIR * getPointer() const { 
	  if (mCurrentIndex >= 0 && mCurrentIndex < mMaximumIndex)
	    return mPairs[mCurrentIndex]; 
	  else 
	    return NULL;
	}
      
	/** advance one position, until you find an aligned pair */
	virtual void Next() { mCurrentIndex++; if (mCurrentIndex > mMaximumIndex) mCurrentIndex = mMaximumIndex; }
	
	/** step back one position, until you find an aligned pair */
	virtual void Previous() { mCurrentIndex--; if (mCurrentIndex < -1) mCurrentIndex = -1; }
	
    private:
	PAIRVECTOR & mPairs;
	ALIGNANDUM_POSITION mCurrentIndex;
	ALIGNANDUM_POSITION mMaximumIndex;
	
    };
    
    /** return const iterator */
    virtual AlignataIterator_const begin() const; 
    
    /** return const iterator */
    virtual AlignataIterator_const end() const; 

    /** return const iterator */
    virtual AlignataIterator begin(); 
    
    /** return const iterator */
    virtual AlignataIterator end(); 

    //----------------> accessors <------------------------------------------------------------------------------

    /** returns the first residue aligned in row */
    virtual ALIGNANDUM_POSITION getRowFrom() const;

    /** returns the last residue aligned in row */
    virtual ALIGNANDUM_POSITION	getRowTo() const;

    /** returns the first residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColFrom() const;

    /** returns the last residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColTo()   const;

    /** returns the first aligned pair. Have to create a copy not a reference, because not all alignments will have
	a list of pairs ready */
    virtual RESIDUEPAIR getFirstPair() const;
    
    /** returns the last aligned pair */
    virtual RESIDUEPAIR getLastPair() const;

    /** adds a pair of residue to the alignment */
    virtual void addPair( RESIDUEPAIR * new_pair ); 

    /** adds a pair of residues to the alignment */
    virtual void addPair( ALIGNANDUM_POSITION row, ALIGNANDUM_POSITION col, CALCTYPE score = 0); 

    /** removes a pair of residues from the alignment */
    virtual void removePair( const RESIDUEPAIR & old_pair );

    /** retrieves a pair of residues from the alignment */
    virtual RESIDUEPAIR getPair( ALIGNANDUM_POSITION row ) const;

    /** clear the current alignemnt */
    virtual void Clear();

    /** move alignment */
    virtual void moveAlignment( ALIGNANDUM_POSITION row_offset, ALIGNANDUM_POSITION col_offset);

 protected:

    /** build index */
    virtual void buildIndex() const = 0;

    /** sort Dots by row and col */
    virtual void sortDots() const = 0;

    /** calculate alignment length */
    virtual void calculateLength() const;

    /** allocate Memory for ndots dots */
    virtual void allocateIndex( unsigned long size ) const;
    
    /** eliminate duplicates entries in mPairs. This only works, if mPairs has
	been sorted previously.
    */
    virtual void eliminateDuplicates() const;
    
    /** sort dots in range from-to by row */
    void sortDotsByRow(ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) const;

    /** sort dots in range from-to by col */
    void sortDotsByCol(ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) const;

    /** sort dots in range from from to to by diagonal */
    void sortDotsByDiagonal(ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) const;

    /** List of residue pairs, mutatble, because they get sorted in-situ */
    mutable PAIRVECTOR mPairs;

    /** index of pairs for each row */
    mutable TYPE_DOT * mIndex;

    /** coordinates of alignment */
    mutable ALIGNANDUM_POSITION mRowFrom;
    mutable ALIGNANDUM_POSITION mRowTo;
    mutable ALIGNANDUM_POSITION mColFrom;
    mutable ALIGNANDUM_POSITION mColTo;


 private:
    /* allocated size of size */
    mutable long mAllocatedIndexSize;

};

						  

}

#endif /* _ALIGNATA_H */

