/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream> 
#include <algorithm>
#include <strstream>
#include <set>
#include "alignlib.h"
#include "ImplAlignataSet.h"
#include "AlignataIterator.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

//------------------------------factory functions -----------------------------
Alignata * makeAlignataSet() {
  return new ImplAlignataSet();
}

//------------------------------------< constructors and destructors >-----
ImplAlignataSet::ImplAlignataSet() : ImplAlignata() {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataSet::ImplAlignataSet()" << endl;
#endif
}

ImplAlignataSet::ImplAlignataSet( const ImplAlignataSet& src) : ImplAlignata( src ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataSet::ImplAlignataSet( const ImplAlignataSet & )" << endl;
#endif
    
    clearContainer();

    PairIterator it(src.mPairs.begin()), it_end(src.mPairs.end());
    for (;it != it_end; ++it) 
      mPairs.insert( new RESIDUEPAIR(**it) );
}

ImplAlignataSet::~ImplAlignataSet( ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataSet::~ImplAlignataSet()" << endl;
#endif
    Clear();
}

//------------------------------------------------------------------------------------------------------------
ImplAlignataSet * ImplAlignataSet::getNew() const {
    return new ImplAlignataSet();
};
    
ImplAlignataSet * ImplAlignataSet::getClone() const {
    return new ImplAlignataSet( *this );
}

//------------------------------------------------------------------------------------------------------------
void ImplAlignataSet::clearContainer() { 
  PAIRSET::iterator it(mPairs.begin()), it_end(mPairs.end());
  for (;it != it_end; ++it) delete *it;
  mPairs.clear(); 
}

//------------------------------------------------------------------------------------------------------------
void ImplAlignataSet::Clear() { 
  ImplAlignata::Clear();
  clearContainer();
}

//-----------------------------------------------------------------------------------------------------------   

AlignataIterator_const ImplAlignataSet::begin() const { 
  return AlignataIterator_const( new ImplAlignataSet_Iterator_const(mPairs.begin()));
}

AlignataIterator_const ImplAlignataSet::end()   const { 
  return AlignataIterator_const( new ImplAlignataSet_Iterator_const(mPairs.end())); 
}

AlignataIterator ImplAlignataSet::begin() { 
  return AlignataIterator( new ImplAlignataSet_Iterator(mPairs.begin()));
}

AlignataIterator ImplAlignataSet::end() { 
  return AlignataIterator( new ImplAlignataSet_Iterator(mPairs.end())); 
}


//----------------> accessors <------------------------------------------------------------------------------

ALIGNANDUM_POSITION ImplAlignataSet::getRowFrom() const { return (mPairs.size() > 0) ? (*mPairs.begin())->mRow : 0; }
ALIGNANDUM_POSITION ImplAlignataSet::getColFrom() const { return (mPairs.size() > 0) ? (*mPairs.begin())->mCol : 0; }
ALIGNANDUM_POSITION ImplAlignataSet::getRowTo()   const { return (mPairs.size() > 0) ? (*mPairs.rbegin())->mRow : 0; }
ALIGNANDUM_POSITION ImplAlignataSet::getColTo()   const { return (mPairs.size() > 0) ? (*mPairs.rbegin())->mCol : 0; }
  
RESIDUEPAIR ImplAlignataSet::getFirstPair() const { return **(mPairs.begin()); }
RESIDUEPAIR ImplAlignataSet::getLastPair()  const { return **(mPairs.rbegin()); }

void ImplAlignataSet::addPair( RESIDUEPAIR * new_pair ) { 
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignataSet::addPair( RESIDUEPAIR *)" << endl;
#endif
  if (mPairs.find( new_pair) != mPairs.end()) {
    delete new_pair;
  } else {
    setChangedLength(); 
    mPairs.insert( new_pair ); 
  } 
} 

//----------------------------------------------------------------------------------------------------------
/** retrieves a pair of residues from the alignment */
RESIDUEPAIR ImplAlignataSet::getPair( ALIGNANDUM_POSITION row ) const {

  RESIDUEPAIR p(row, 0, 0);
  PairIterator it(mPairs.find( &p ));
  return **it;
} 


void ImplAlignataSet::removePair( const RESIDUEPAIR & old_pair ) { 
#ifdef DEBUG
  if (verbose > LL5) cout << "void ImplAlignataSet::removePair( RESIDUEPAIR &)" << endl;
#endif
  
  RESIDUEPAIR p(old_pair);
  PairIterator it(mPairs.find( &p ));
  
  if (it != mPairs.end()) {
    setChangedLength(); 
    delete *it;
    mPairs.erase(it);
  }
} 

//----------------------------------------------------------------------------------------
/** This is non-generic routine. Since pairs are accessed by row, this is quite quick.
 */
void ImplAlignataSet::removeRowRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) {

  for (ALIGNANDUM_POSITION pos = from; pos <= to; pos++) {
      RESIDUEPAIR p(pos, 0, 0);
      PairIterator it(mPairs.find( &p ));

      if (it != mPairs.end()) {
	setChangedLength(); 
	delete *it;
	mPairs.erase(it);
      }
  }
  
}

//----------------------------------------------------------------------------------------
/** This is a generic routine. It creates a new alignment by making a copy of the old one */
void ImplAlignataSet::removeColRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) {

  PairIterator it(mPairs.begin()), it_end(mPairs.end());

  for (; it != it_end; ++it) {
    if ( (*it)->mCol >= from && (*it)->mCol <= to) {
      delete *it;
      mPairs.erase(it);
    }
  }
  
  setChangedLength();

}

} // namespace AlignLib
