/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream> 
#include <algorithm>
#include <strstream>
#include <set>
#include "alignlib.h"
#include "ImplAlignataSetCol.h"
#include "AlignataIterator.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

//------------------------------factory functions -----------------------------
Alignata * makeAlignataSetCol() {
  return new ImplAlignataSetCol();
}

//------------------------------------< constructors and destructors >-----
ImplAlignataSetCol::ImplAlignataSetCol() : ImplAlignata() {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataSetCol::ImplAlignataSetCol()" << endl;
#endif
}

ImplAlignataSetCol::ImplAlignataSetCol( const ImplAlignataSetCol& src) : ImplAlignata( src ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataSetCol::ImplAlignataSetCol( const ImplAlignataSetCol & )" << endl;
#endif

    clearContainer();

    PairIterator_const it(src.mPairs.begin()), it_end(src.mPairs.end());
    for (; it != it_end; ++it) 
      mPairs.insert( new RESIDUEPAIR(**it) );
}

ImplAlignataSetCol::~ImplAlignataSetCol( ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataSetCol::~ImplAlignataSetCol()" << endl;
#endif
    Clear();
}

//------------------------------------------------------------------------------------------------------------
ImplAlignataSetCol * ImplAlignataSetCol::getNew() const {
    return new ImplAlignataSetCol();
};
    
ImplAlignataSetCol * ImplAlignataSetCol::getClone() const {
    return new ImplAlignataSetCol( *this );
}

//-------------------------------------------------------------------
void ImplAlignataSetCol::clearContainer() { 
  PAIRSET::iterator it(mPairs.begin()), it_end(mPairs.end());
  for (;it != it_end; ++it) delete *it;
  mPairs.clear(); 
}

//------------------------------------------------------------------------------------------------------------
void ImplAlignataSetCol::Clear() { 
  ImplAlignata::Clear();
  clearContainer();
}

//-----------------------------------------------------------------------------------------------------------   

AlignataIterator_const ImplAlignataSetCol::begin() const { 
  return AlignataIterator_const( new ImplAlignataSetCol_Iterator_const(mPairs.begin()));
}

AlignataIterator_const ImplAlignataSetCol::end()   const { 
  return AlignataIterator_const( new ImplAlignataSetCol_Iterator_const(mPairs.end())); 
}

AlignataIterator ImplAlignataSetCol::begin() { 
  return AlignataIterator( new ImplAlignataSetCol_Iterator(mPairs.begin()));
}

AlignataIterator ImplAlignataSetCol::end() { 
  return AlignataIterator( new ImplAlignataSetCol_Iterator(mPairs.end())); 
}


//----------------> accessors <------------------------------------------------------------------------------

ALIGNANDUM_POSITION ImplAlignataSetCol::getRowFrom() const { return (mPairs.size() > 0) ? (*mPairs.begin())->mRow : 0; }
ALIGNANDUM_POSITION ImplAlignataSetCol::getColFrom() const { return (mPairs.size() > 0) ? (*mPairs.begin())->mCol : 0; }
ALIGNANDUM_POSITION ImplAlignataSetCol::getRowTo()   const { return (mPairs.size() > 0) ? (*mPairs.rbegin())->mRow : 0; }
ALIGNANDUM_POSITION ImplAlignataSetCol::getColTo()   const { return (mPairs.size() > 0) ? (*mPairs.rbegin())->mCol : 0; }
  
RESIDUEPAIR ImplAlignataSetCol::getFirstPair() const { return **(mPairs.begin()); }
RESIDUEPAIR ImplAlignataSetCol::getLastPair()  const { return **(mPairs.rbegin()); }

void ImplAlignataSetCol::addPair( RESIDUEPAIR * new_pair ) { 
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignataSetCol::addPair( RESIDUEPAIR &)" << endl;
#endif
  if (mPairs.find( new_pair) != mPairs.end()) {
    delete new_pair;
  } else {
    setChangedLength(); 
    mPairs.insert( new_pair ); 
  } 
} 

//----------------------------------------------------------------------------------------------------------
/** retrieves a pair of residues from the alignment */
RESIDUEPAIR ImplAlignataSetCol::getPair( ALIGNANDUM_POSITION row ) const {

  RESIDUEPAIR p(row, 0, 0);
  PairIterator it(mPairs.find( &p ));
  return **it;
} 


void ImplAlignataSetCol::removePair( const RESIDUEPAIR & old_pair ) { 
#ifdef DEBUG
  if (verbose > LL5) cout << "void ImplAlignataSetCol::removePair( RESIDUEPAIR &)" << endl;
#endif

  setChangedLength(); 
  RESIDUEPAIR p(old_pair);
  mPairs.erase(&p); 
} 

//----------------------------------------------------------------------------------------
/** This is non-generic routine. Since pairs are accessed by row, this is quite quick.
 */
void ImplAlignataSetCol::removeColRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) {

  for (ALIGNANDUM_POSITION pos = from; pos <= to; pos++) {
      RESIDUEPAIR p(0, pos, 0);
      PairIterator it(mPairs.find( &p ));

      if (it != mPairs.end()) {
	setChangedLength(); 
	delete *it;
	mPairs.erase(it);
      }
  }
  
}

//----------------------------------------------------------------------------------------
/** This is a generic routine. It creates a new alignment by making a copy of the old one */
void ImplAlignataSetCol::removeRowRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) {

  PairIterator it(mPairs.begin()), it_end(mPairs.end());

  for (; it != it_end; ++it) {
    if ( (*it)->mRow >= from && (*it)->mRow <= to) {
      delete *it;
      mPairs.erase(it);
    }
  }
  
  setChangedLength();

}


} // namespace AlignLib

































