/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream> 
#include <algorithm>
#include <strstream>
#include <set>
#include "alignlib.h"
#include "ImplAlignataVector.h"
#include "AlignataIterator.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

#define MAX_ROW 1000000

/** by how much the vector grows */
#define GROWTH_FACTOR 2

//------------------------------factory functions -----------------------------
Alignata * makeAlignataVector() {
  return new ImplAlignataVector();
}

//------------------------------------< constructors and destructors >-----
ImplAlignataVector::ImplAlignataVector() : ImplAlignata(), mRowFrom(MAX_ROW), mRowTo(0) {
}

ImplAlignataVector::ImplAlignataVector( const ImplAlignataVector& src) : 
  ImplAlignata( src ), 
  mRowFrom( src.mRowFrom), 
  mRowTo( src.mRowTo) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataVector::ImplAlignataVector( const ImplAlignataVector & )" << endl;
#endif

    // do not call Clear() function, as this will delete attributes of parent
    // class like mScore.

    clearContainer();
    mPairs.resize( mRowTo + 1, NULL);

    PAIRVECTOR::const_iterator it(src.mPairs.begin()), it_end(src.mPairs.end());
    for (;it != it_end; ++it) 
      if (*it != NULL) 
	mPairs[(*it)->mRow] = new RESIDUEPAIR( (**it) );
}

ImplAlignataVector::~ImplAlignataVector( ) {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignataVector::~ImplAlignataVector()" << endl;
#endif
    Clear();
}

//------------------------------------------------------------------------------------------------------------
ImplAlignataVector * ImplAlignataVector::getNew() const {
    return new ImplAlignataVector();
}
    
ImplAlignataVector * ImplAlignataVector::getClone() const {
    return new ImplAlignataVector( *this );
}

//-----------------------------------------------------------------------------------------------------------   

AlignataIterator_const ImplAlignataVector::begin() const { 
  // skip gaps in the beginning of the alignment
  return AlignataIterator_const( new ImplAlignataVector_Iterator_const( mPairs, mRowFrom, mRowFrom, mRowTo )); 
}

AlignataIterator_const ImplAlignataVector::end() const { 
  return AlignataIterator_const( new ImplAlignataVector_Iterator_const(mPairs, 0, mRowFrom, mRowTo)); 
}

//-----------------------------------------------------------------------------------------------------------   

AlignataIterator ImplAlignataVector::begin() { 
  // skip gaps in the beginning of the alignment
  return AlignataIterator( new ImplAlignataVector_Iterator( mPairs, mRowFrom, mRowFrom, mRowTo )); 
}

AlignataIterator ImplAlignataVector::end() { 
  return AlignataIterator( new ImplAlignataVector_Iterator(mPairs, 0, mRowFrom, mRowTo)); 
}

//----------------> accessors <------------------------------------------------------------------------------

ALIGNANDUM_POSITION ImplAlignataVector::getRowFrom() const { return mRowFrom; }
ALIGNANDUM_POSITION ImplAlignataVector::getColFrom() const { 
  if (mRowFrom != MAX_ROW) 
    return mPairs[mRowFrom]->mCol; 
  else 
    return 0;
}

ALIGNANDUM_POSITION ImplAlignataVector::getRowTo()   const { return mRowTo; }
ALIGNANDUM_POSITION ImplAlignataVector::getColTo()   const { 
  if (mRowTo != 0) 
    return mPairs[mRowTo]->mCol; 
  else
    return 0;
}

  
RESIDUEPAIR ImplAlignataVector::getFirstPair() const { return *mPairs[mRowFrom]; }
RESIDUEPAIR ImplAlignataVector::getLastPair()  const { return *mPairs[mRowTo]; }

//----------------------------------------------------------------------------------------------------------------------
void ImplAlignataVector::addPair( RESIDUEPAIR * new_pair ) { 
  
  ALIGNANDUM_POSITION new_row = new_pair->mRow;
  

  if (mRowFrom > new_row || mRowFrom == 0) 
    mRowFrom = new_row;

  if (mRowTo < new_row) 
    mRowTo = new_row;

  unsigned int needed_size = mRowTo + 1;
  if (mPairs.size() < needed_size) 
    mPairs.resize( needed_size * GROWTH_FACTOR, NULL);
  
  if (mPairs[new_row] != NULL) 
    delete mPairs[new_row];

  mPairs[new_row] = new_pair; 
  setChangedLength();
} 

//--------------------------------------------------------------------------------------------------------------
void ImplAlignataVector::moveAlignment( ALIGNANDUM_POSITION row_offset, ALIGNANDUM_POSITION col_offset) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignataVector::moveAlignment( ALIGNANDUM_POSITION, ALIGNANDUM_POSITION) const" << endl;
#endif

  // create copy of mPairs (= copy of pointers)
  PAIRVECTOR copy(mPairs);
  
  PairIterator it(copy.begin()), it_end(copy.end());
  
  // set new alignment coordinates
  mRowFrom += row_offset;  
  mRowTo   += row_offset;
  unsigned int needed_size = mRowTo + 1;

  // delete old alignment and allocate needed size
  mPairs.clear();
  mPairs.resize( needed_size, NULL);

  // copy pointers from copy into mPairs
  for (; it != it_end; ++it) {
    RESIDUEPAIR * p = *it;
    if (p != NULL) {
      p->mRow += row_offset;
      p->mCol += col_offset;
      mPairs[p->mRow] = p;
    }
  }

};


//----------------------------------------------------------------------------------------------------------
/** retrieves a pair of residues from the alignment */
RESIDUEPAIR ImplAlignataVector::getPair( ALIGNANDUM_POSITION row ) const {
  return *mPairs[row];
} 

//----------------------------------------------------------------------------------------------------------
void ImplAlignataVector::removePair( const RESIDUEPAIR & old_pair ) { 

  // no resizing is done;
  if (old_pair.mRow >= mRowFrom && old_pair.mRow <= mRowTo)
    mPairs[old_pair.mRow] = NULL;

  // if first pair has been deleted, update mRowFrom
  while (mRowFrom <= mRowTo && mPairs[mRowFrom] == NULL) 
    mRowFrom++;
  
  // if last pairs has been deleted, update mRowTo
  while (mRowTo >= mRowFrom && mPairs[mRowTo] == NULL)
    mRowTo--;

  if (mRowFrom > mRowTo)
      Clear();

  setChangedLength();
} 



//----------------------------------------------------------------------------------------------------------------
void ImplAlignataVector::clearContainer() { 

  PAIRVECTOR::iterator it(mPairs.begin()), it_end(mPairs.end());
  for (;it != it_end; ++it) delete *it;
  mPairs.clear();

}

//----------------------------------------------------------------------------------------------------------------
void ImplAlignataVector::Clear() { 
  ImplAlignata::Clear();

  clearContainer();
  
  mRowFrom = MAX_ROW; 
  mRowTo = 0; 
}

//--------------> mapping functions <----------------------------------------------------------------------------
ALIGNANDUM_POSITION ImplAlignataVector::mapRowToCol( ALIGNANDUM_POSITION pos ) const {
  if (pos >= mRowFrom && pos <= mRowTo && mPairs[pos] != NULL)
    return mPairs[pos]->mCol;
  else
    return 0;
}


//-----------------------------------------------------------------------------------------------------------   
void ImplAlignataVector::resetBoundaries() {

  ALIGNANDUM_POSITION max_size = mPairs.size();

  // find new mRowFrom and mRowTo, if they have changed
  while (mRowFrom < max_size && mPairs[mRowFrom] == NULL) mRowFrom++;
  while (mRowTo > 0          && mPairs[mRowTo] == NULL) mRowTo--;

  // test >= because max_size might be zero, if all has been 
  // deleted.
  if (mRowFrom >= max_size) 
      mRowFrom = MAX_ROW;

}
//-----------------------------------------------------------------------------------------------------------   
void ImplAlignataVector::removeRowRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) {

  ALIGNANDUM_POSITION pos;

  if (from < mRowFrom)
    from = mRowFrom;
  
  if (to > mRowTo)
    to = mRowTo;
  
  // delete aligned positions
  for ( pos = from; pos <= to; pos++) {
    if (mPairs[pos] != NULL) {
      delete mPairs[pos];
      mPairs[pos] = NULL;
    }
  }

  resetBoundaries();

  setChangedLength();

  return;
}

//-----------------------------------------------------------------------------------------------------------   
  /* It is necessary to iterate from mFrowFrom to mRowTo, since the alignment need not be linear
   */
void ImplAlignataVector::removeColRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) {

  ALIGNANDUM_POSITION pos; 

  for (pos = mRowFrom; pos <= mRowTo; pos++)
    if (mPairs[pos] != NULL && mPairs[pos]->mCol >= from && mPairs[pos]->mCol <= to) {
      delete mPairs[pos];
      mPairs[pos] = NULL;
    }

  resetBoundaries();

  setChangedLength();

  return;
}




} // namespace AlignLib

