/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef IMPL_ALIGNATA_VECTOR_H
#define IMPL_ALIGNATA_VECTOR_H 1

#include <iosfwd>
#include <vector>
#include "alignlib.h"
#include "ImplAlignata.h"

namespace AlignLib {

    /** class for aligned objects. 

	if alignment is empty, mRowFrom is larger than mRowTo. This property is needed
	at some checks (for example in ImplAlignataVector_iterator)
	
	Residues are kept in a vector. 
	
	It is not assumed that the alignment is linear. Linear means, that col[i] < col[i+1] and
	row[i] < row[i+1]. 

	The advantates:
	@li mapping is much faster

	The disadvantages:
	@li uses more memory for all the pairs, that are not part of the alignment
	@li iterating over all pairs is slower, since gaps have to be skipped explicitely.

	@author Andreas Heger
	@version $Id: ImplAlignataVector.h,v 1.1.1.1 2002/07/01 17:15:20 heger Exp $
	@short protocol class for aligned objects
    */



class Alignandum;

class ImplAlignataVector : public ImplAlignata {

    typedef std::vector<RESIDUEPAIR *> PAIRVECTOR;
    typedef PAIRVECTOR::iterator PairIterator;
    typedef PAIRVECTOR::const_iterator PairIterator_const;

    friend class Iterator_const;
    
 public:

    //------------------> constructors / destructors <---------------------------------------------------------
    /** constructor */
    ImplAlignataVector();

    /** copy constructor */
    ImplAlignataVector( const ImplAlignataVector &src );

    /** destructor */
    virtual ~ImplAlignataVector();

    //------------------------------------------------------------------------------------------------------------
    virtual ImplAlignataVector * getNew() const;
    
    /** return an identical copy */
    virtual ImplAlignataVector * getClone() const;

    //------------------------------------------------------------------------------------------------------------
    //------------------------------------------------------------------------------------------------------------
    class ImplAlignataVector_Iterator_const : public Alignata::Iterator_const{
    public:
      ImplAlignataVector_Iterator_const(const PAIRVECTOR & container, 
					ALIGNANDUM_POSITION current, 
					ALIGNANDUM_POSITION from, 
					ALIGNANDUM_POSITION to) : 
	mContainer( container ),
	mCurrentRow( current ), 
	mFirstRow( from), 
	mLastRow( to ) { 
	if (mCurrentRow > mLastRow || to == 0) 
	  mCurrentRow = 0;
      };
	
      ImplAlignataVector_Iterator_const( const ImplAlignataVector_Iterator_const & src ) : 
	mContainer( src.mContainer),
	mCurrentRow( src.mCurrentRow), 
	mFirstRow( src.mFirstRow), 
	mLastRow( src.mLastRow) {};
	
      virtual ~ImplAlignataVector_Iterator_const() {};
      
      virtual Iterator_const * getClone() const {return new ImplAlignataVector_Iterator_const( *this );}
      
      /** dereference operator: runtime error, if out of bounds? */
      virtual const RESIDUEPAIR & getReference() const { return *mContainer[mCurrentRow];}
 
      /** for indirection */
      virtual const RESIDUEPAIR * getPointer() const { 
	if (mCurrentRow != 0) 
	  return mContainer[mCurrentRow]; 
	else 
	  return NULL;
      }

      /** advance one position, until you find an aligned pair */
      virtual void Next() { 
	mCurrentRow++;
	while (mCurrentRow <= mLastRow && mContainer[mCurrentRow] == NULL) { 
	  mCurrentRow++;
	}
	if (mCurrentRow > mLastRow) mCurrentRow = 0;
      }
	
        /** step back one position, until you find an aligned pair */
      virtual void Previous() { 
	mCurrentRow--; 
	while (mCurrentRow >= mFirstRow && mContainer[mCurrentRow] == NULL) { 
	  mCurrentRow--;
	}
	if (mCurrentRow < mFirstRow) mCurrentRow = 0;
      }

    private:
      const PAIRVECTOR & mContainer;
      ALIGNANDUM_POSITION mCurrentRow;
      ALIGNANDUM_POSITION mFirstRow;
      ALIGNANDUM_POSITION mLastRow;
    };

    //------------------------------------------------------------------------------------------------------------
    class ImplAlignataVector_Iterator : public Alignata::Iterator{
    public:
      ImplAlignataVector_Iterator(PAIRVECTOR & container, ALIGNANDUM_POSITION current, ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to) : 
	mContainer( container ),
	mCurrentRow( current ), 
	mFirstRow( from), 
	mLastRow( to ) { 
	if (mCurrentRow > mLastRow || to == 0) 
	  mCurrentRow = 0;
      };
	
      ImplAlignataVector_Iterator( const ImplAlignataVector_Iterator & src ) : 
	mContainer( src.mContainer),
	mCurrentRow( src.mCurrentRow), 
	mFirstRow( src.mFirstRow), 
	mLastRow( src.mLastRow) {};
	
      virtual ~ImplAlignataVector_Iterator() {};
      
      virtual Iterator * getClone() const {return new ImplAlignataVector_Iterator( *this );}
      
      /** dereference operator: runtime error, if out of bounds? */
      virtual RESIDUEPAIR & getReference() const { return *mContainer[mCurrentRow];}
 
      /** for indirection */
      virtual RESIDUEPAIR * getPointer() const { 
	if (mCurrentRow != 0) 
	  return mContainer[mCurrentRow]; 
	else 
	  return NULL;
      }

      /** advance one position, until you find an aligned pair */
      virtual void Next() { 
	mCurrentRow++;
	while (mCurrentRow <= mLastRow && mContainer[mCurrentRow] == NULL ) { 
	  mCurrentRow++;
	}
	if (mCurrentRow > mLastRow) mCurrentRow = 0;
      }
	
        /** step back one position, until you find an aligned pair */
      virtual void Previous() { 
	mCurrentRow--; 
	while (mCurrentRow >= mFirstRow && mContainer[mCurrentRow] == NULL) { 
	  mCurrentRow--;
	}
	if (mCurrentRow < mFirstRow) mCurrentRow = 0;
      }

    private:
      PAIRVECTOR & mContainer;
      ALIGNANDUM_POSITION mCurrentRow;
      ALIGNANDUM_POSITION mFirstRow;
      ALIGNANDUM_POSITION mLastRow;
    };

    /** return const iterator */
    virtual AlignataIterator_const begin() const; 
    
    /** return const iterator */
    virtual AlignataIterator_const end() const; 

    /** return const iterator */
    virtual AlignataIterator begin(); 
    
    /** return const iterator */
    virtual AlignataIterator end(); 

    //----------------> accessors <------------------------------------------------------------------------------

    /** returns the first residue aligned in row */
    virtual ALIGNANDUM_POSITION getRowFrom() const;

    /** returns the last residue aligned in row */
    virtual ALIGNANDUM_POSITION	getRowTo() const;

    /** returns the first residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColFrom() const;

    /** returns the last residue aligned in col */
    virtual ALIGNANDUM_POSITION	getColTo()   const;

    /** returns the first aligned pair. Have to create a copy not a reference, because not all alignments will have
	a list of pairs ready */
    virtual RESIDUEPAIR getFirstPair() const;
    
    /** returns the last aligned pair */
    virtual RESIDUEPAIR getLastPair() const;

    /** adds a pair of residue to the alignment */
    virtual void addPair( RESIDUEPAIR * new_pair ); 

    /** removes a pair of residues from the alignment */
    virtual void removePair( const RESIDUEPAIR & old_pair );

    /** retrieves a pair of residues from the alignment */
    virtual RESIDUEPAIR getPair( ALIGNANDUM_POSITION row ) const;

    /** maps a residue from row to column. returns 0, if not found. This is quick, since there is 
	only one lookup in the array needed.*/
    virtual ALIGNANDUM_POSITION mapRowToCol( ALIGNANDUM_POSITION pos ) const;

    /** move alignment */
    virtual void moveAlignment( ALIGNANDUM_POSITION row_offset, ALIGNANDUM_POSITION col_offset);

    /** removes a row-region in an alignment */
    virtual void removeRowRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to );

    /** removes a column-region in an alignment */
    virtual void removeColRegion( ALIGNANDUM_POSITION from, ALIGNANDUM_POSITION to);

    /** clear the current alignemnt */
    void Clear();
 protected:
    
    /** reset boundaries mRowFrom and mRowTo after the deletion of pairs */
    void resetBoundaries();

 private:

    /** clear container */
    void clearContainer();

    /** Vector of residue pairs */
    PAIRVECTOR mPairs;
     
    /** row of the first aligned residue pair */
    ALIGNANDUM_POSITION mRowFrom;
     
    /** row of the last aligned residue pair */
    ALIGNANDUM_POSITION mRowTo;
};

						  

}

#endif /* _ALIGNATA_H */

