/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#if HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef IMPL_ALIGNATOR_H
#define IMPL_ALIGNATOR_H 1

#include "alignlib.h"
#include "Alignator.h"
#include "Alignata.h"

namespace AlignLib {

/**
   Base class for aligning two objects. The default call to use (with base class) is
   
   Alignator a;
   Alignata r;

   a( row, col, &r );

   The two objects to be aligned to each other are called row and col. This comes from
   the dynamic programming matrix where you have one sequence along the rows and the other 
   along the columns.

   This class is a protocol class and as such defines only the general interface
   
   @author Andreas Heger
   @version $Id: ImplAlignator.h,v 1.1.1.1 2002/07/01 17:15:20 heger Exp $
   @short protocol class for aligned objects
*/ 

class SubstitutionMatrix;
class Alignandum;
class Alignata;

class ImplAlignator : public Alignator {
  /* class member functions-------------------------------------------------------------- */
 public:
    /* constructors and desctructors------------------------------------------------------- */
    /** constructor */
    ImplAlignator( const SubstitutionMatrix * subst_matrix,
		   CALCTYPE row_gop, 
		   CALCTYPE row_gep, 
		   CALCTYPE col_gop = 0 , 
		   CALCTYPE col_gep = 0 );

    /** destructor */
    virtual ~ImplAlignator ();

    /** copy constructor */
    ImplAlignator( const ImplAlignator & src);
    
    /** set gap opening penalty for row */
    virtual void setRowGop( CALCTYPE gop );

    /** set gap extension penalty for row */
    virtual void setRowGep( CALCTYPE gep );

    /** set gap opening penalty for col */
    virtual void setColGop( CALCTYPE gop );

    /** set gap extension penalty for col */
    virtual void setColGep( CALCTYPE gep );

    /** set substitution matrix to use */
    virtual void setSubstitutionMatrix( const SubstitutionMatrix * matrix );

    /** align two objects and just return the score */
    virtual CALCTYPE getAlignmentScore(const Alignandum *, const Alignandum * );		

 protected:
    /** get gap opening penalty for row */
    CALCTYPE getRowGop();

    /** get gap extension penalty for row */
    CALCTYPE getRowGep();

    /** get gap opening penalty for col */
    CALCTYPE getColGop();

    /** get gap extension penalty for col */
    CALCTYPE getColGep();
    
    /** get location of substitutionmatrix */
    const SubstitutionMatrix * getSubstitutionMatrix() ;

    /** get length of row object */
    ALIGNANDUM_POSITION getRowLength();

    /** get length of col */
    ALIGNANDUM_POSITION getColLength();

    /** perform initialisation before alignment. Overload, but call this function in subclasses! */
    virtual void startUp( const Alignandum * row, const Alignandum * col, Alignata * ali);
    
    /** perform cleanup after alignment */
    virtual void cleanUp(const Alignandum * row, const Alignandum * col, Alignata * ali);                     

    /** pointer to substitution matrix to use. not private, so children can access it */
    const SubstitutionMatrix * mSubstitutionMatrix;

 protected:
    /** gap opening penalty for row-object */
    CALCTYPE mRowGop;
    /** gap elongation penalty for col-object */
    CALCTYPE mRowGep;
    /** gap opening penalty for row-object */
    CALCTYPE mColGop;
    /** gap elongation penalty for col-object */
    CALCTYPE mColGep;  
    /** length of object in row */
    ALIGNANDUM_POSITION mRowLength;
    /** length of object in col */
    ALIGNANDUM_POSITION mColLength;
    
};

}

#endif /* ALIGNATOR_H */
