/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef IMPL_ALIGNATOR_DOTS_H
#define IMPL_ALIGNATOR_DOTS_H 1

#include "alignlib.h"
#include "ImplAlignator.h"
#include "ImplAlignataMatrix.h"

namespace AlignLib {
/** 
    This objects aligns two @ref Alignandum objects using a dot-matrix.
    
    @author Andreas Heger
    @version $Id: ImplAlignatorDots.h,v 1.1.1.1 2002/07/01 17:15:20 heger Exp $
    @short represents an alignment algorithm
*/

#define STACKEMPTY     0
    
    class Alignandum;
    class Alignata;
    class SubstitutionMatrix;
    class ImplAlignataMatrix;
    
class ImplAlignatorDots : public ImplAlignator {
 public:

    /* constructors and destructors */

    /** set affine gap penalties
     @param row_gop		gap opening penalty in row
     @param row_gep		gap elongation penalty in row
     @param col_gop		gap opening penalty in column, default = row
     @param col_gep		gap elongation penalty in row, default = col
     
    */
    ImplAlignatorDots( const SubstitutionMatrix * subst_matrix,
		       CALCTYPE row_gop, 
		       CALCTYPE row_gep, 
		       CALCTYPE col_gop = 0,
		       CALCTYPE col_gep = 0,
		       Alignator * dottor = NULL );

    /** copy constructor */
    ImplAlignatorDots( const ImplAlignatorDots & );

    /** destructor */
    virtual ~ImplAlignatorDots();

    /* operators------------------------------------------------------------------------------ */
    /** method for aligning two arbitrary objects */
    virtual Alignata * Align(const Alignandum *, const Alignandum *, Alignata *);

    /* member access functions--------------------------------------------------------------- */

 protected:

    /** perform initialisation before alignment. Overload, but call this function in subclasses! */
    virtual void startUp( const Alignandum * row, const Alignandum * col, Alignata * ali);
    
    /** perform cleanup after alignment */
    virtual void cleanUp(const Alignandum * row, const Alignandum * col, Alignata * ali);                     

    /** traces back through dot-trace and put it in the alignment in Alignata-object */
    virtual void traceBack( const Alignandum * row, const Alignandum * col, Alignata * result);				
    
    /** perform the alignment */
    virtual void performAlignment(const Alignandum * row, const Alignandum *col, Alignata * result);

    /** get GAP cost for a gap */
    virtual CALCTYPE getGapCost( TYPE_DOT x1, TYPE_DOT x2 ) const = 0; 

    /** look up a pair given row and column */
    ALIGNANDUM_POSITION getPairIndex( ALIGNANDUM_POSITION row, ALIGNANDUM_POSITION col ) const;

 protected:
    
    /** The dotter object that supplies the dots */
    Alignator * mDottor;
    
    /* diverse helper variables */
    ALIGNANDUM_POSITION	mNDots;

    /** pointer to dots for fast access */
    const ImplAlignataMatrix::PAIRVECTOR * mPairs;		

    /** pointer to first dot in row for fast access */
    const TYPE_DOT * mRowIndices;		
    
    /* help variables for backtracing */
    /** last dot */
    int	mLastDot;		

    /** trace of dots that are part of the alignment */
    int	*mTrace;		

    /** the score of the alignment */
    CALCTYPE mScore;

    /** pointer to the matrix of Dots (= alignment ) */
    ImplAlignataMatrix * mMatrix;

    /** flag, whether dots were created during alignment */
    bool mIsOwnDots;
    
};


}

#endif /* _ALIGNATOR_DOTS_H */

