/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream>
#include <iomanip>
#include <stdio.h>

#include <map>
#include <vector>

#include "alignlib.h"
#include "AlignException.h"
#include "ImplAlignatorDotsSquared.h"
#include "Alignandum.h"
#include "ImplAlignataMatrixRow.h"

#include "HelpersSubstitutionMatrix.h"

#include "Alignata.h"
#include "HelpersAlignata.h"


#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

#define NODOT -1

/*---------------------factory functions ---------------------------------- */

    /** make an alignator object, which does a dot-alignment. The default version can be given an AlignataMatrix-
	object */
Alignator * makeAlignatorDotsSquared(CALCTYPE gop, CALCTYPE gep, Alignator * alignator, const SubstitutionMatrix * subst_matrix) {

  if (!subst_matrix) 
      return new ImplAlignatorDotsSquared( getDefaultSubstitutionMatrix(), 
				    gop, gep, gop, gep, alignator );
  else
      return new ImplAlignatorDotsSquared( subst_matrix, 
				    gop, gep, gop, gep, alignator );
  
}

//----------------------------------------------------------------------------------------------------------------------------------------
    /** constructors and destructors */
ImplAlignatorDotsSquared::ImplAlignatorDotsSquared( const SubstitutionMatrix * subst_matrix,
				      CALCTYPE row_gop, CALCTYPE row_gep, 
				      CALCTYPE col_gop, CALCTYPE col_gep,
				      Alignator * dots) :
    ImplAlignatorDots( subst_matrix, row_gop - row_gep, row_gep, col_gop - col_gep, col_gep, dots) {
}

//------------------------------------------------------------------------------------------
ImplAlignatorDotsSquared::ImplAlignatorDotsSquared( const ImplAlignatorDotsSquared & src ) : 
  ImplAlignatorDots( src ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplAlignatorDotsSquared::ImplAlignatorDotsSquared( const ImplAlignatorDotsSquared &) " << endl;
#endif
}

//------------------------------------------------------------------------------------------
ImplAlignatorDotsSquared::~ImplAlignatorDotsSquared() {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignatorDotsSquared::~ImplAlignatorDotsSquared()" << endl;
#endif
}
//------------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------------
CALCTYPE ImplAlignatorDotsSquared::getGapCost( TYPE_DOT x1, TYPE_DOT x2 ) const {

  ALIGNANDUM_POSITION c1 = (*mPairs)[x1]->mCol;
  ALIGNANDUM_POSITION c2 = (*mPairs)[x2]->mCol;  
  ALIGNANDUM_POSITION r1 = (*mPairs)[x1]->mRow;
  ALIGNANDUM_POSITION r2 = (*mPairs)[x2]->mRow;  

  CALCTYPE gap_cost = 0;
  ALIGNANDUM_POSITION d;

  if ((d = (r2 - r1)) > 1)
      gap_cost += mRowGop + d * mRowGep;

  if ((d = (c2 - c1)) > 1)
      gap_cost += mColGop + d * mColGep;

  return gap_cost;
}

//-------------------------------------------< Alignment subroutine >----------------------------------------------
void ImplAlignatorDotsSquared::performAlignment( const Alignandum * prow, const Alignandum * pcol, Alignata * ali) {

  /**
     Overview over the algorithm
     
     1. Dots are sorted by row and then by column

     col ->
     row
     |
     
     ------------|
     |           |
     |           |
     |           |
 c-> |           |
     ------------x

     note: you do not have to consider dots in the same row or column, as it is
     not possible to match the same row to two different columns and vice versa.
     
     

  */
  // #define DEBUG
  // #define DEBUG2

  TYPE_DOT global_best_dot = NODOT;
  CALCTYPE global_best_score = 0;

  // create data structure for search region
  // sort dots in search region by increasing column
  typedef multimap <ALIGNANDUM_POSITION, TYPE_DOT> MyDotSet;

  MyDotSet search_region;
  
  // array with scores of dots
  vector<CALCTYPE> scores(mNDots,0);

  // array with dots in current row
  vector<TYPE_DOT> dot_stack(mColLength, NODOT);

  unsigned int num_row_dots = 0;
  ALIGNANDUM_POSITION last_row = 0;
  //----------------------------------> main alignment loop <----------------------------------------------------
  for ( TYPE_DOT current_dot = 0; current_dot < mNDots; current_dot++ ) {	   

    ALIGNANDUM_POSITION current_row = (*mPairs)[current_dot]->mRow;                         
    ALIGNANDUM_POSITION current_col = (*mPairs)[current_dot]->mCol;                         

    /* if a new row is entered, enter dots from stack to search-area */
    if (current_row != last_row) {
      while (num_row_dots > 0) {
	TYPE_DOT dot = dot_stack[--num_row_dots];
	search_region.insert(pair<ALIGNANDUM_POSITION, TYPE_DOT>((*mPairs)[dot]->mCol, dot));
      }
      last_row = current_row;
    }

    /* search search-area: always lookup starting at col 1 until current_col - 1. Try to find
       a positive trace leading to current dot. If it were negative, it would not be part of
       the optimum alignment up to current_dot. */
    TYPE_DOT search_best_dot   = NODOT;
    CALCTYPE search_best_score = 0;


#ifdef DEBUG2
    cout << "SEARCH_AREA" << endl;
    for (MyDotSet::iterator it = search_region.begin(); it != search_region.end(); ++it)
      cout << "  [" << (*it).first << ", " << (*it).second << "]" << endl;
#endif
    
    MyDotSet::const_iterator it(search_region.begin()), it_end(search_region.end());
    while (it != it_end && ((*it).first < current_col)) {
      
      TYPE_DOT const search_dot = (*it).second;
      CALCTYPE search_score = scores[search_dot];
      
      if (search_score > 0) {
	search_score += getGapCost( search_dot, current_dot);
	if (search_score >= search_best_score) {
	  search_best_score = search_score;
	  search_best_dot   = search_dot;
	}
      }

      ++it;
    }

    /* no positive trace found, new trace starts at current dot */
    if (search_best_dot == NODOT)
      search_best_score = (*mPairs)[current_dot]->mScore;
    else
      search_best_score += (*mPairs)[current_dot]->mScore;

#ifdef DEBUG
    cout << "current_dot=" << current_dot << " current_row=" << current_row << " current_col=" << current_col << endl;
    cout << "search_best_dot=" << search_best_dot << " search_best_score=" << search_best_score << endl;
    if (search_best_dot != NODOT) cout << "gap_cost=" << getGapCost(search_best_dot, current_dot) << endl;
#endif
    
    /* do local alignment, traces with score <= 0 are skipped */
    if (search_best_score < 0)
      continue;
    
    scores[current_dot] = search_best_score;
    mTrace[current_dot] = search_best_dot;

    dot_stack[num_row_dots++] = current_dot;    

    /* remember end point of best trace */
    if (search_best_score > global_best_score) {
      global_best_score = search_best_score;
      global_best_dot   = current_dot;
    }
    
  } /* end of alignment loop */

  mLastDot= global_best_dot;
  mScore  = global_best_score;

#ifdef DEBUG
  cout << "global_best_dot=" << global_best_dot << " global_best_score=" << global_best_score << endl;
#endif

  //--------------> cleaning up <---------------------------------------------------------------
  //--------------------------------------------------------------------------------------------------
}

} // namespace AlignLib







