/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream>
#include <iomanip>
#include <stdio.h>

#include "alignlib.h"
#include "AlignException.h"

#include "Alignandum.h"

#include "ImplAlignatorFragments.h"

#include "Fragmentor.h"

#include "Alignata.h"
#include "HelpersAlignata.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

#define NOFRAGMENT -1

/*---------------------factory functions ---------------------------------- */

//--------------------------------------------------------------------------------------------------------
    /** constructors and destructors */
ImplAlignatorFragments::ImplAlignatorFragments( const SubstitutionMatrix * subst_matrix,
						CALCTYPE row_gop, CALCTYPE row_gep, 
						CALCTYPE col_gop, CALCTYPE col_gep,
						Fragmentor * fragmentor) :
  ImplAlignator( subst_matrix, row_gop - row_gep, row_gep, col_gop - col_gep, col_gep), 
  mFragmentor(fragmentor) {
}

//--------------------------------------------------------------------------------------------------------
ImplAlignatorFragments::ImplAlignatorFragments( const ImplAlignatorFragments & src ) : 
  ImplAlignator( src ), mFragmentor(src.mFragmentor) {
#ifdef DEBUG
  if (verbose > LL3) 
    cout << "ImplAlignatorFragments::ImplAlignatorFragments( const ImplAlignatorFragments &) " << endl;
#endif
}

//--------------------------------------------------------------------------------------------------------
ImplAlignatorFragments::~ImplAlignatorFragments() {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignatorFragments::~ImplAlignatorFragments()" << endl;
#endif
}
//--------------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------------
void ImplAlignatorFragments::startUp(const Alignandum * row, const Alignandum *col, Alignata * ali) {
    ImplAlignator::startUp(row, col, ali);  

#ifdef DEBUG
    if (verbose > LL3) cout << "void AlignatorFragments::Initialize()" << endl;
#endif
    
    // setup example alignment type
    Alignata * dummy_alignment = makeAlignataSet();
    
    // create dots
    mFragments = mFragmentor->Fragment( row, col, dummy_alignment ); 
    
    // get the number of dots, which corresponds to the length of the
    // alignment in this class. Tell the matrix to sort, etc., at the 
    // same time.
    mNFragments = mFragments->size();

    mTrace        = new int[mNFragments];
    mLastFragment = -1;
}

//--------------------------------------------------------------------------------------------------------
void ImplAlignatorFragments::cleanUp(const Alignandum * row, const Alignandum *col, Alignata * ali) {

#ifdef DEBUG
  if (verbose > LL3) cout << "void AlignatorFragments::CleanUp()" << endl;
#endif

  if (mTrace != NULL) 
    delete [] mTrace;

  FragmentVector::iterator it(mFragments->begin()), it_end(mFragments->end());
  for (; it!= it_end; ++it) 
    delete *it;

  delete mFragments;
    
  ImplAlignator::cleanUp(row, col, ali);
  
}

//------------------------------------------------------------------------------------------------------
Alignata * ImplAlignatorFragments::Align(const Alignandum * row, const Alignandum * col, Alignata * result) {
#ifdef DEBUG
  if (verbose > LL3) cout << "Alignata* ImplAlignatorFragments::operator(const Alignandum *, const Alignandum *, Alignata *)" << endl;
#endif
  
  startUp(row, col, result);
  
  performAlignment(row, col, result);
    
  traceBack(row, col, result);
  
  cleanUp(row, col, result);
  
  return result;
}

//-----------------------------------------< BackTracke >-----------------------------------------------
void ImplAlignatorFragments::traceBack( const Alignandum * row, const Alignandum * col, Alignata * result) {
#ifdef DEBUG
    cout << "ImplAlignatorFragments::traceBack() starts\n";
#endif

    /* trace back along the fragments that are part of
       the optimum trace. Add pairwise alignments to 
       result */

    TYPE_FRAGMENT ifragment = mLastFragment;

    while ( ifragment >= 0) {
      addAlignata2Alignata( result, (*mFragments)[ifragment] );

#ifdef DEBUG
      cout << "Traceback: fragment " << ifragment << " with score " << (*mFragments)[ifragment]->getScore() << endl;
      writePairAlignment( cout, row, col, (*mFragments)[ifragment] );
#endif

      ifragment = mTrace[ifragment];

    }

    result->setScore( mScore );
} 

//------------------------------------------------------------------------------------------------------



} // namespace AlignLib
