/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream>
#include <iomanip>
#include <stdio.h>

#include <map>
#include <vector>
#include <algorithm>

#include "alignlib.h"
#include "AlignException.h"
#include "ImplAlignatorFragmentsSquared.h"
#include "Alignandum.h"

#include "Alignata.h"
#include "HelpersAlignata.h"

#include "HelpersSubstitutionMatrix.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

#define NOFRAGMENT -1
  
struct BorderPosition {
  BorderPosition() : mRow(0), mTop(false), mFragment(NOFRAGMENT) {}
  BorderPosition( ALIGNANDUM_POSITION row, bool top, TYPE_FRAGMENT fragment) : 
    mRow(row), mTop(top), mFragment(fragment) {}
  ALIGNANDUM_POSITION mRow;
  bool mTop;
  TYPE_FRAGMENT mFragment;
};

  class BorderComparator {
  public:
    int operator() (const BorderPosition & x, const BorderPosition & y) const {
      return x.mRow < y.mRow;
    }
  };


/*---------------------factory functions ---------------------------------- */

    /** make an alignator object, which does a dot-alignment. The default version can be given an AlignataMatrix-
	object */
Alignator * makeAlignatorFragmentsSquared(CALCTYPE gop, 
					  CALCTYPE gep, 
					  Fragmentor * fragmentor, 
					  const SubstitutionMatrix * subst_matrix) {

  if (!subst_matrix) 
    return new ImplAlignatorFragmentsSquared( getDefaultSubstitutionMatrix(), 
					      gop, gep, gop, gep, fragmentor );
  else
    return new ImplAlignatorFragmentsSquared( subst_matrix, 
					      gop, gep, gop, gep, fragmentor);
  
}

//----------------------------------------------------------------------------------------------------------------------------------------
    /** constructors and destructors */
ImplAlignatorFragmentsSquared::ImplAlignatorFragmentsSquared( const SubstitutionMatrix * subst_matrix,
				      CALCTYPE row_gop, CALCTYPE row_gep, 
				      CALCTYPE col_gop, CALCTYPE col_gep,
				      Fragmentor * fragmentor) :
    ImplAlignatorFragments( subst_matrix, row_gop - row_gep, row_gep, col_gop - col_gep, col_gep, fragmentor) {
}

//------------------------------------------------------------------------------------------
ImplAlignatorFragmentsSquared::ImplAlignatorFragmentsSquared( const ImplAlignatorFragmentsSquared & src ) : 
  ImplAlignatorFragments( src ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplAlignatorFragmentsSquared::ImplAlignatorFragmentsSquared( const ImplAlignatorFragmentsSquared &) " << endl;
#endif
}

//------------------------------------------------------------------------------------------
ImplAlignatorFragmentsSquared::~ImplAlignatorFragmentsSquared() {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplAlignatorFragmentsSquared::~ImplAlignatorFragmentsSquared()" << endl;
#endif
}
//------------------------------------------------------------------------------------------
//------------------------------------------------------------------------------------------
CALCTYPE ImplAlignatorFragmentsSquared::getGapCost( TYPE_DOT x1, TYPE_DOT x2 ) const {

  ALIGNANDUM_POSITION r1 = (*mFragments)[x1]->getRowTo();
  ALIGNANDUM_POSITION r2 = (*mFragments)[x2]->getRowFrom();  
  ALIGNANDUM_POSITION c1 = (*mFragments)[x1]->getColTo();
  ALIGNANDUM_POSITION c2 = (*mFragments)[x2]->getColFrom();  

  CALCTYPE gap_cost = 0;
  ALIGNANDUM_POSITION d;

  if ((d = (r2 - r1)) > 1)
      gap_cost += mRowGop + d * mRowGep;

  if ((d = (c2 - c1)) > 1)
      gap_cost += mColGop + d * mColGep;

  return gap_cost;
}

//-------------------------------------------< Alignment subroutine >----------------------------------------------
void ImplAlignatorFragmentsSquared::performAlignment( const Alignandum * prow, const Alignandum * pcol, Alignata * ali) {

  /**
     Overview over the algorithm:

     - proceed row-wise.
     - update score of path up to current fragment,
     if left border is touched.
     - add fragment to search-region, if bottom is
     higher than the current row.

  */
// #define DEBUG
// #define DEBUG2

  TYPE_FRAGMENT global_best_fragment = NOFRAGMENT;
  CALCTYPE global_best_score = 0;

  // fill borders with fragments
  vector<BorderPosition> borders;
  typedef vector<BorderPosition>::const_iterator BorderIterator;
  for (TYPE_FRAGMENT i = 0; i < mNFragments; i++) {
    borders.push_back(BorderPosition((*mFragments)[i]->getRowFrom(), true, i));
    borders.push_back(BorderPosition((*mFragments)[i]->getRowTo(), false, i));    
  }

  // sort borders according to row
  sort( borders.begin(), borders.end(), BorderComparator());
  
  // create data structure for search region
  // sort fragments in search region by increasing column
  typedef multimap <ALIGNANDUM_POSITION, TYPE_FRAGMENT> MyFragmentSet;
  MyFragmentSet search_region;
  
  // array with scores of fragments
  vector<CALCTYPE> scores(mNFragments,0);

  // array with fragments in current row
  vector<TYPE_FRAGMENT> fragment_stack;

  ALIGNANDUM_POSITION last_row = 0;

  //----------------------------------> main alignment loop <----------------------------------------------------
  BorderIterator it(borders.begin()), it_end(borders.end());
  for ( ; it!=it_end; ++it ) {	   

    TYPE_FRAGMENT current_fragment = it->mFragment;
    ALIGNANDUM_POSITION current_row = it->mRow;                         
    bool is_top      = it->mTop;

#ifdef DEBUG
    cout << "current_fragment=" << current_fragment
	 << " current_row=" << current_row
	 << " is_top=" << is_top 
	 << endl;
#endif

    /* if a new row is entered, enter fragments from stack to search-area */
    if (current_row != last_row) {
      
      while (!fragment_stack.empty()) {
	TYPE_FRAGMENT fragment = fragment_stack.back();
	fragment_stack.pop_back();
	search_region.insert(pair<ALIGNANDUM_POSITION, TYPE_FRAGMENT>((*mFragments)[fragment]->getColTo(), fragment));
      }
      last_row = current_row;
    }

    if (is_top) {
      /* search search-area: always lookup starting at col 1 until current_col - 1. Try to find
	 a positive trace leading to current fragment. If it were negative, it would not be part of
	 the optimum alignment up to current_fragment. */
      TYPE_FRAGMENT search_best_fragment   = NOFRAGMENT;
      ALIGNANDUM_POSITION current_col = (*mFragments)[current_fragment]->getColFrom();
      CALCTYPE search_best_score = 0;

#ifdef DEBUG2
      cout << "SEARCH_AREA" << endl;
      for (MyFragmentSet::iterator search_it = search_region.begin(); search_it != search_region.end(); ++search_it)
	cout << "  [" << (*search_it).first << ", " << (*search_it).second << "]" 
	     << "=" << scores[(*search_it).second] 
	     << " " << mTrace[(*search_it).second] 
	     << endl;
#endif
    
      MyFragmentSet::const_iterator search_it(search_region.begin()), search_it_end(search_region.end());
      while (search_it != search_it_end && ((*search_it).first < current_col)) {
	
	TYPE_FRAGMENT const search_fragment = (*search_it).second;
	CALCTYPE search_score = scores[search_fragment];
	
	if (search_score > 0) {
	  search_score += getGapCost( search_fragment, current_fragment);
	  if (search_score >= search_best_score) {
	    search_best_score = search_score;
	    search_best_fragment   = search_fragment;
	  }
	}
	
	++search_it;
      }
      
      /* no positive trace found, new trace starts at current fragment */
      if (search_best_fragment == NOFRAGMENT)
	search_best_score = (*mFragments)[current_fragment]->getScore();
      else
	search_best_score += (*mFragments)[current_fragment]->getScore();

#ifdef DEBUG
      cout << "current_fragment=" << current_fragment 
	   << " current_row=" << current_row 
	   << " current_col=" << current_col 
	   << endl;
      cout << "search_best_fragment=" << search_best_fragment 
	   << " search_best_score=" << search_best_score 
	   << endl;
#endif
      
      /* do local alignment, traces with score <= 0 are skipped */
      if (search_best_score < 0)
	continue;
      
      scores[current_fragment] = search_best_score;
      mTrace[current_fragment] = search_best_fragment;
      
      /* remember end point of best trace */
      if (search_best_score > global_best_score) {
	global_best_score = search_best_score;
	global_best_fragment   = current_fragment;
      }
      
    } else {
      /* add fragment to list of fragments, that are to be added to the
	 search area, once the current row has finished */
      fragment_stack.push_back(current_fragment);    
    }
  } 
  /* ---->end of alignment loop<----- */
    
  mLastFragment= global_best_fragment;
  mScore  = global_best_score;

#ifdef DEBUG
  cout << "global_best_fragment=" << global_best_fragment << " global_best_score=" << global_best_score << endl;
#endif

  //--------------> cleaning up <---------------------------------------------------------------
  //--------------------------------------------------------------------------------------------------

}

} // namespace AlignLib






