/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream>
#include <iomanip>
#include "ImplAlignatum.h"
#include "ImplTranslator.h"
#include "Alignata.h"
#include "AlignataIterator.h"
#include "Alignandum.h"
#include "Renderer.h"
#include "HelpersTranslator.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

  //-------------------------------------------------------------------------------------------------
  /** factory functions */
  Alignatum * makeAlignatum() {
    return new ImplAlignatum();
  }

  Alignatum * makeAlignatumFromString( const std::string & src, 
				       ALIGNANDUM_POSITION from, 
				       ALIGNANDUM_POSITION to ) {
    return new ImplAlignatum( src, from, to);
  }
  Alignatum * makeAlignatumFromString( const char * src, 
				       ALIGNANDUM_POSITION from, 
				       ALIGNANDUM_POSITION to ) {
    std::string s( src );
    return new ImplAlignatum( s, from, to);
  }
  
  Alignatum * makeAlignatum(const Alignandum * src, 
			    const Alignata * map_this2new,
			    const ALIGNANDUM_POSITION max_length) {
    
    std::string s( src->asString());
    Alignatum * result = new ImplAlignatum( s );
  
    ALIGNANDUM_POSITION length;
    
    if (map_this2new) {
      if (max_length == 0)
	length = map_this2new->getColTo();
      else
	length = max_length;
      result->mapOnAlignment( map_this2new, length );
    }
    
    return result;
  }
    

//---------------------------------------------------------< constructors and destructors >--------
ImplAlignatum::ImplAlignatum() :   
  mRepresentation(""),
  mFrom(0), 
  mTo (0),
  mLength(0), 
  mGapChar (getDefaultTranslator()->getGapChar()),
  mSeparator('\t') {
}


ImplAlignatum::ImplAlignatum (const std::string & representation, 
			      ALIGNANDUM_POSITION from, 
			      ALIGNANDUM_POSITION to) : 
  mRepresentation(representation), 
  mFrom( from), mTo( to ),
  mGapChar (getDefaultTranslator()->getGapChar()),
  mSeparator('\t') {

  mLength = mRepresentation.length();
  
  if (mTo == 0) 
    mTo = mFrom + mLength - countGaps() - 1;
}

//--------------------------------------------------------------------------------------------
/* create a newly aligned object from ImplAlignatum, but map using ali in alignment 
   note, that the residues in the alignment are called 1..length while the residues
   in the aligned strings are 0..length-1
*/
ImplAlignatum::ImplAlignatum (const ImplAlignatum & src ):
    mRepresentation(src.mRepresentation),
    mFrom( src.mFrom), 
    mTo ( src.mTo ),
    mLength( src.mLength), 
    mGapChar (getDefaultTranslator()->getGapChar()),
    mSeparator( src.mSeparator ) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignatum::ImplAlignatum( const ImplAlignatum & src)" << endl;
#endif
}	
	       
//--------------------------------------------------------------------------------------------
ImplAlignatum::~ImplAlignatum () {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplAlignatum::~ImplAlignatum ()" << endl;
#endif
}

//--------------------------------------------------------------------------------------------
void ImplAlignatum::mapOnAlignment(const Alignata * map_old2new,
				   const ALIGNANDUM_POSITION new_length, 
				   const bool unaligned_chars) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignatum::mapOnAlignment( const Alignata *)" << endl;
#endif

  std::string new_representation = "";
  
  ALIGNANDUM_POSITION length = new_length;
  if (new_length == 0)
    length = map_old2new->getColTo();

  new_representation.append( length, mGapChar );
  
  // get alignment start positions
  int row_from = map_old2new->getRowFrom();
  int row_to = map_old2new->getRowTo();

  // get amino-acid start positions and save them in from/to
  mFrom = getResidueNumberNext(row_from-1); 
  mTo   = getResidueNumberPrevious(row_to-1); 

  // substitute new characters for aligned positions:
  {
    AlignataIterator_const it = map_old2new->begin();
    AlignataIterator_const it_end = map_old2new->end();
    
    for (; it != it_end; ++it) 
      new_representation[it->mCol-1] = mRepresentation[it->mRow-1];
  }

  // add unaligned characters
  if (unaligned_chars) {

    AlignataIterator_const it = map_old2new->begin();
    AlignataIterator_const it_end = map_old2new->end();

    ALIGNANDUM_POSITION last_old = it->mRow;
    ALIGNANDUM_POSITION last_new = it->mCol;
    
    ++it;
    // substitute new characters for aligned positions:
    for (; it != it_end; ++it) {
      ALIGNANDUM_POSITION old = it ->mRow-1;
      ALIGNANDUM_POSITION nnew = it ->mCol-1;      
      while (old - last_old > 0 && nnew - last_new > 0) {
	old--;
	nnew--;
	if (mRepresentation[old] >= 'A' &&
	    mRepresentation[old] <= 'Z' )
	  new_representation[nnew] = mRepresentation[old] - 'A' + 'a';
      }
      last_old = it->mRow;
      last_new = it->mCol;
    }
  }
  
  mRepresentation = new_representation;
  mLength = mRepresentation.length();
  
}
//------------------------------------------------------------------------------------------------------
ImplAlignatum * ImplAlignatum::getClone() const {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplAlignatum * ImplAlignatum::getClone()" << endl;
#endif
  return new ImplAlignatum(*this ); 
}

//-------------------------------------------------------------------------------------------------------
ImplAlignatum * ImplAlignatum::getNew() const {
  return new ImplAlignatum();
}

//-------------------------------------------------------------------------------------------------------
std::string ImplAlignatum::getString() const {
    return mRepresentation;
}

//-------------------------------------------------------------------------------------------------------
const std::string & ImplAlignatum::getStringReference() const {
    return mRepresentation;
}
       
//-------------------------------------------------------------------------------------------------------
ALIGNANDUM_POSITION ImplAlignatum::getFrom() const {
    return mFrom;
}
    
//-------------------------------------------------------------------------------------------------------
ALIGNANDUM_POSITION ImplAlignatum::getTo() const {
    return mTo;
}

//-------------------------------------------------------------------------------------------------------
void ImplAlignatum::writeRow( std::ostream & output, 
			      ALIGNANDUM_POSITION segment_start, 
			      ALIGNANDUM_POSITION segment_end,
			      const Renderer * renderer) const {

    if (segment_end >= mLength || segment_end <= segment_start) 
	segment_end = mLength;
  
    ALIGNANDUM_POSITION left  = getResidueNumberNext( segment_start );
    ALIGNANDUM_POSITION right = getResidueNumberPrevious( segment_end - 1);

    if (renderer) 
      output << left << getFieldSeparator() 
	     << renderer->Render(mRepresentation, segment_start, segment_end) << getFieldSeparator()
	     << right;
    else
      output << left << getFieldSeparator()
      	     << mRepresentation.substr( segment_start, segment_end - segment_start) << getFieldSeparator()
	     << right;

}

//---------------------------------------------------------------------------------------------------
void ImplAlignatum::readRow( std::istream & input ) {
}

/** get represenation */
const std::string & ImplAlignatum::getRepresentation() const { 
  return mRepresentation; 
} 

/** set representation */
void ImplAlignatum::setRepresentation( std::string & representation, 
				       ALIGNANDUM_POSITION first_res,
				       ALIGNANDUM_POSITION last_res) { 

  // set first residue number to 1 if not given
  if (first_res < 1) 
    first_res = 1;

  mFrom = first_res;

  mRepresentation = representation; 
  mLength = mRepresentation.length();

  if (last_res < 1)
    mTo = mLength - countGaps();

}


//---------------------------------------------------------------------------------------------------
/** return the length of the line */
ALIGNANDUM_POSITION ImplAlignatum::getAlignedLength() const { 
    return mLength;
}

//---------------------------------------------------------------------------------------------------
/** return the length of the line without gaps */
ALIGNANDUM_POSITION ImplAlignatum::getTrueLength() const {
    return mTo - mFrom + 1;
}
    
//---------------------------------------------------------------------------------------------------
/** add the specified number of gaps in the front and in the back */
void ImplAlignatum::addGaps(int before, int after) {
    int i = 0;
    std::string x  = "";
    
    for (i = 0; i < before; i++) x += mGapChar;
    x+= mRepresentation;
    for (i = 0; i < after; i++)  x += mGapChar;

    mRepresentation = x;
    mLength = mRepresentation.length();
}


//---------------------------------------------------------------------------------------------------
/** add one or more gaps in the middle */
void ImplAlignatum::insertGaps( int position, ALIGNANDUM_POSITION count ) {

  std::string insertion = "";
  for (int i = 0; i < count; i++) 
    insertion += mGapChar;

  mRepresentation.insert( position, insertion );
  mLength = mRepresentation.length();
}

//------------------------------------------------------------------------------------------------
/** remove leading/or trailing gaps */
void ImplAlignatum::removeEndGaps() {
  mRepresentation.erase( 0, mRepresentation.find_first_not_of( mGapChar ));
  mRepresentation.erase( mRepresentation.find_last_not_of( mGapChar ) + 1, mRepresentation.length());
  mLength = mRepresentation.length();
}


//------------------------------------------------------------------------------------------------
/** remove one or more positions from the aligned object */
void ImplAlignatum::removeColumns( int position, ALIGNANDUM_POSITION count ) {
  mRepresentation.erase( position, position + count );
}

//---------------------------------------------------------------------------------------
int ImplAlignatum::countGaps() {
  int ngaps = 0;
  ALIGNANDUM_POSITION length = mRepresentation.length();
  ALIGNANDUM_POSITION i = 0;
 
  for (; i < length; i++) 
    if (mRepresentation[i] == mGapChar)
      ngaps ++;
  
  return ngaps;
}

//------------------------------------------------------------------------------------------
// Calculate the amino acid residue number of residue in string-position pos. Start counting
// from the left
ALIGNANDUM_POSITION ImplAlignatum::getResidueNumberNext( ALIGNANDUM_POSITION pos ) const {

    ALIGNANDUM_POSITION i = 0;
    ALIGNANDUM_POSITION result = mFrom;
    while (i < pos && mRepresentation[i] == mGapChar) i++;
    
    for (; i < pos; i++) 
	if (mRepresentation[i] != mGapChar) 
	    result++;
    
  return (result);
}
//------------------------------------------------------------------------------------------
// Calculate the amino acid residue number of residue in string-position pos. Start counting
// from the right
ALIGNANDUM_POSITION ImplAlignatum::getResidueNumberPrevious( ALIGNANDUM_POSITION pos ) const {

    ALIGNANDUM_POSITION i = mRepresentation.length() - 1;
    ALIGNANDUM_POSITION result = mTo;
    while (i >= pos && mRepresentation[i] == mGapChar) i--;
    
    for (; i > pos; i--) 
	if (mRepresentation[i] != mGapChar) 
	    result--;
    
  return (result);
}

/** write into stream */
void ImplAlignatum::Write( std::ostream & output ) const {
  writeRow( output );
}

/** read from stream */
void ImplAlignatum::Read( std::istream & input ) {

  input >> mFrom;
  input >> mRepresentation;
  input >> mTo;

  mLength = mRepresentation.length();

};

//-------------------------------------------------------------------------------------------

} // namespace AlignLib
