/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef IMPL_ALIGNATUM_H
#define IMPL_ALIGNATUM_H 1

#include <iosfwd>
#include <string>
#include "alignlib.h"
#include "Alignatum.h"

namespace AlignLib {

/** Objects of this type represent an aligned object. This class is responsible for
    producing a nice, real-world representation of the aligned object. To this end, 
    this class' provides several methods for modifying an aligned string.

    There is no method for inserting residues here. Do this, by modifying alignments
    before creating an Alignatum-object.

    Adding Gaps and deleting columns works on the string, i.e. the numbering starts
    at 0. These functions are intended to make multiple alignments look nice, not
    to modify alignments.
    
    This class is a protocol class and as such defines only the general interface

    @author Andreas Heger
    @version $Id: ImplAlignatum.h,v 1.3 2003/08/16 09:09:26 aheger Exp $
    @short protocol class for aligned objects
*/

class Alignata;
class Alignandum;
class Renderer;

class ImplAlignatum : public Alignatum {
 public:
    // constructors and desctructors
    /** when building a new Alignatum sequence by cloning and an alignment, this specifies,
	if the new sequence is in row or column
    */

    /** constructor */
    ImplAlignatum  ();
    
    /** copy constructor */
    ImplAlignatum  (const ImplAlignatum & src);
    /** destructor */
    virtual ~ImplAlignatum ();

    ImplAlignatum (const std::string & representation, 
		   ALIGNANDUM_POSITION from = 1, 
		   ALIGNANDUM_POSITION to = 0);

    /*-----> accessors <----------------------------------------------------- */
    /** write object into stream nicely formatted. Segments are addresed by [from,to)
	@param output stream for result
	@param segment_start beginning of segment 
	@param segment_end end of segment 
    */
    virtual void writeRow( std::ostream & output, 
			   ALIGNANDUM_POSITION segment_start = 0, 
			   ALIGNANDUM_POSITION segment_end = 0,
			   const Renderer * = NULL)  const;

    /** readline */
    virtual void readRow( std::istream & input );

    /** return a copy of the string representation */
    virtual std::string getString() const;

    /** return a reference to the string representation */
    virtual const std::string & getStringReference() const;

    /** return the length of the line */
    virtual ALIGNANDUM_POSITION getAlignedLength() const;

    /** return the length of the line without gaps */
    virtual ALIGNANDUM_POSITION getTrueLength() const;
    
    /** return the number of the first residue */
    virtual ALIGNANDUM_POSITION getFrom() const;

    /** return the number of the last residue */
    virtual ALIGNANDUM_POSITION getTo() const;

    /** add the specified number of gaps in the front and in the back */
    virtual void addGaps(int before, int after);		
    
    /** add one or more gaps in the middle 
	@param position where gap(s) should be inserted 
	@param count the number of gaps to be inserted
    */
    virtual void insertGaps( int position, ALIGNANDUM_POSITION count = 1);
    
    /** remove leading/or trailing gaps */
    virtual void removeEndGaps();

    /** remove one or more positions from the aligned object */
    virtual void removeColumns( int position, ALIGNANDUM_POSITION count = 1); 

    /** remap the current alignatum object using ali. 
	If unaligned chars is true, lower case unaligned characters will be
	put before the next aligned character (as much as fit) 
    */
    virtual void mapOnAlignment(const Alignata * map_old2new,
				const ALIGNANDUM_POSITION new_length = 0,
				const bool unaligned_chars = false );

    /** return a copy of the object, mapped if so desired 
	@param ali alignment to be used for the mapping. If empty, return identical copy
	@param skip_gaps true, if gaps in other sequence shall be skipped 
	@param is_in_row true, if alignment is in row
    */
    virtual ImplAlignatum * getClone() const;
    
    /** return an empty copy of this object */
    virtual ImplAlignatum * getNew() const;

    /** write into stream */
    virtual void  Write( std::ostream & output ) const;

    /** read from stream */
    virtual void  Read( std::istream & input );

 protected:

    /** calculate number of gaps in sequence */
    virtual int countGaps();

    /** calculate residue number of residue in position pos, or, if this is a gap, the residue number
	of the next non-gap character. */
    virtual ALIGNANDUM_POSITION getResidueNumberNext( ALIGNANDUM_POSITION pos ) const;

    /** calculate residue number of residue in pos, or, if this is a gap, the residue number
	of the previous non-gap character. */
    virtual ALIGNANDUM_POSITION getResidueNumberPrevious( ALIGNANDUM_POSITION pos ) const;

    /** accessor functions */
    
    /** get represenation */
    const std::string & getRepresentation() const;

    /** set representation */
    void setRepresentation( std::string & representation,
			    ALIGNANDUM_POSITION first_res = 0,
			    ALIGNANDUM_POSITION last_res = 0);

    inline char getFieldSeparator() const { return mSeparator; }

 private:
    
    /** the string representation used for single line output */
    std::string mRepresentation;

    /** number of first residue in aligned string */
    ALIGNANDUM_POSITION mFrom;
    
    /** number of last residue in aligned string */
    ALIGNANDUM_POSITION mTo;
    
    /** length of aligned string */
    ALIGNANDUM_POSITION mLength;

    /** gap-character */
    char mGapChar;

    /** field separator */
    char mSeparator;
};

}

#endif /* ALIGNATUM_H */

