/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream>
#include <iomanip>
#include "ImplAlignatumFasta.h"
#include "ImplTranslator.h"
#include "Alignata.h"
#include "AlignataIterator.h"
#include "Alignandum.h"
#include "Renderer.h"
#include "HelpersTranslator.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

//--------------------------------------------------------------------------------------------
  /** factory functions */
  Alignatum * makeAlignatumFasta() {
    return new ImplAlignatumFasta();
  }

  Alignatum * makeAlignatumFasta( const std::string & description,
				  const std::string & src, 
				  ALIGNANDUM_POSITION from, 
				  ALIGNANDUM_POSITION to ) {

    return new ImplAlignatumFasta( description, src, from, to);
  }



//---------------------------------------------------------< constructors and destructors >--------
ImplAlignatumFasta::ImplAlignatumFasta() :   
  ImplAlignatum(),
  mDescription("") {
}

ImplAlignatumFasta::ImplAlignatumFasta (const std::string & description,
					const std::string & representation, 
					ALIGNANDUM_POSITION from, 
					ALIGNANDUM_POSITION to ) :
					
  ImplAlignatum( representation, from, to ),
  mDescription( description ) {
}

//--------------------------------------------------------------------------------------------
/* create a newly aligned object from ImplAlignatumFasta, but map using ali in alignment 
   note, that the residues in the alignment are called 1..length while the residues
   in the aligned strings are 0..length-1
*/
ImplAlignatumFasta::ImplAlignatumFasta (const ImplAlignatumFasta & src ):
  ImplAlignatum(src), 
  mDescription(src.mDescription) {
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplAlignatumFasta::ImplAlignatumFasta( const ImplAlignatumFasta & src)" << endl;
#endif
}	
	       
//--------------------------------------------------------------------------------------------
ImplAlignatumFasta::~ImplAlignatumFasta () {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplAlignatumFasta::~ImplAlignatumFasta ()" << endl;
#endif
}

//------------------------------------------------------------------------------------------------------
ImplAlignatumFasta * ImplAlignatumFasta::getClone() const {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplAlignatumFasta * ImplAlignatumFasta::getClone()" << endl;
#endif
  return new ImplAlignatumFasta(*this ); 
}

//-------------------------------------------------------------------------------------------------------
ImplAlignatumFasta * ImplAlignatumFasta::getNew() const {
  return new ImplAlignatumFasta();
}


//-------------------------------------------------------------------------------------------------------
void ImplAlignatumFasta::writeRow( std::ostream & output, 
			      ALIGNANDUM_POSITION segment_start, 
			      ALIGNANDUM_POSITION segment_end,
			      const Renderer * renderer) const {
  
  output << setw(30) << mDescription;

  ImplAlignatum::writeRow( output, segment_start, segment_end, renderer);
}

/** write into stream */
void ImplAlignatumFasta::Write( std::ostream & output ) const {
  output << ">" << mDescription << endl << getRepresentation() << endl;
}

/** read from stream */
void ImplAlignatumFasta::Read( std::istream & input ) {

#define MAX_CHUNK 10000

  char * buffer = new char[MAX_CHUNK];
  
  while ( (input.peek() != '>') && 
	  !input.eof() ) {
    input.getline( buffer, MAX_CHUNK);
  }

  if (input.eof())
    return;
    
  input.get();
  input.getline(buffer, MAX_CHUNK);
    
  mDescription = buffer;

  // erase end of line
  mDescription.erase(mDescription.size(),1); 
  
  std::string representation("");
   
  // build the sequence character-wise
  while ( (input.peek() != '>') && 
	  !input.eof() ) {
      
    input.getline( buffer, MAX_CHUNK);
      
    for (unsigned int i = 0; i < strlen(buffer); i++) 
      if (getDefaultTranslator()->isValidChar( buffer[i] )) 
	representation += buffer[i];
  }
  
  setRepresentation(representation);

  delete [] buffer;

}

//-------------------------------------------------------------------------------------------

} // namespace AlignLib






