/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream>
#include "alignlib.h"

#include "Fragmentor.h"

#include "Alignata.h"
#include "HelpersAlignata.h"

#include "AlignataIterator.h"

#include "Alignandum.h"
#include "AlignException.h"

#include "SubstitutionMatrix.h"
#include "HelpersSubstitutionMatrix.h"

#include "Alignator.h"

#include "ImplFragmentorDiagonals.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

/*---------------------factory functions ---------------------------------- */

  /** make an alignator object, which does a dot-alignment. The default version can be given an AlignataMatrix-
      object */
Fragmentor * makeFragmentorDiagonals(CALCTYPE gop, 
				     CALCTYPE gep, 
				     Alignator * dottor ) {
  
  return new ImplFragmentorDiagonals( gop, gep, gop, gep, dottor );
}


//---------------------------------------------------------------------------------------------------
inline TYPE_DIAGONAL calculateDiagonal( const RESIDUEPAIR & p) { return (p.mCol - p.mRow); }

//----------------------------------------------------------------------------------------------------
  
ImplFragmentorDiagonals::ImplFragmentorDiagonals( CALCTYPE row_gop, CALCTYPE row_gep, 
						  CALCTYPE col_gop, CALCTYPE col_gep,
						  Alignator * dottor) :
  ImplFragmentor(), 
  mRowGop(row_gop), 
  mRowGep( row_gep), 
  mColGop(col_gop), 
  mColGep( col_gep), 
  mDottor( dottor ) {
}


ImplFragmentorDiagonals::~ImplFragmentorDiagonals() {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplFragmentorDiagonals::~ImplFragmentorDiagonals()" << endl;
#endif
}

ImplFragmentorDiagonals::ImplFragmentorDiagonals( const ImplFragmentorDiagonals & src ) : 
   ImplFragmentor(src), 
   mRowGop( src.mRowGop), 
   mRowGep( src.mRowGep), 
   mColGop( src.mColGop), 
   mColGep( src.mColGep), 
   mDottor( src.mDottor) {
}

//------------------------------------------------------------------------------------------
/** CleanUp is empty and does not call base class method, as no shifting is necessary */
void ImplFragmentorDiagonals::cleanUp(const Alignandum * row, const Alignandum *col, Alignata * ali) {
#ifdef DEBUG
    if (verbose > LL3) 
      cout << "ImplFragmentorDiagonals::cleanUp(const Alignandum *, const Alignandum *, Alignata *) " << endl;
#endif
}

//------------------------------------------------------------------------------------------
CALCTYPE ImplFragmentorDiagonals::getGapCost( const RESIDUEPAIR & p1, const RESIDUEPAIR & p2 ) const {

  ALIGNANDUM_POSITION r1 = p1.mRow;
  ALIGNANDUM_POSITION r2 = p2.mRow;  
  ALIGNANDUM_POSITION c1 = p1.mCol;
  ALIGNANDUM_POSITION c2 = p2.mCol;

  CALCTYPE gap_cost = 0;
  ALIGNANDUM_POSITION d;

  if ((d = (r2 - r1 - 1)) > 0)
      gap_cost += mRowGop + d * mRowGep;

  if ((d = (c2 - c1 - 1)) > 0)
      gap_cost += mColGop + d * mColGep;

  return gap_cost;
}
	
//------------------------------------------------------------------------------------------------
void ImplFragmentorDiagonals::performFragmentation( const Alignandum * row, 
						    const Alignandum * col, 
						    const Alignata * sample) {

    // create dot-matrix (sorted by diagonal)
    Alignata * matrix = makeAlignataMatrixDiagonal();
    
    // dots are automatically moved to correct position,
    // if only segments are used in row/col
    mDottor->Align( row, col, matrix );
    
    // rescore dots (later: use plugin Scoror?), set score to 1
    rescoreAlignment( matrix, 1.0);
    
    TYPE_DIAGONAL last_diagonal = MAX_DIAGONAL;
    
    AlignataIterator it(matrix->begin()), it_end(matrix->end());
    
    CALCTYPE last_score = 0;
    Alignata * current_ali = makeAlignataSet();
    
    const RESIDUEPAIR * last_p = NULL;

    unsigned int length = 0;
    
    for (; it != it_end; ++it) {
	
	const RESIDUEPAIR * p = &(*it);
	
	TYPE_DIAGONAL current_diagonal = calculateDiagonal( *p );
	
	// save alignment if diagonal has been changed
	if (last_diagonal != current_diagonal) {
	    // save fragment with positive score
	    if (last_score > 0 && length > 1) {
		Alignata * new_ali = sample->getNew();
		copyAlignata( new_ali, current_ali);
		new_ali->setScore( last_score );
		mFragments->push_back( new_ali );
	    }
	    
	    // erase alignment
	    current_ali->Clear();
	    last_p = NULL;
	    last_diagonal = current_diagonal;
	    last_score = 0;
	    length = 0;
	}
	
	// calculate score along diagonal
	CALCTYPE new_score;
	if (last_p != NULL)
	    new_score = last_score + p->mScore + getGapCost( *last_p, *p);
	else
	    new_score = last_score + p->mScore;

	// save fragment, if score drops below zero from positive score
	if (new_score <= 0) {
	  // stop existing alignment
	  // save
	  if (last_score > 0 && length > 1) {
	    Alignata * new_ali = sample->getNew();
	    copyAlignata( new_ali, current_ali);
	    new_ali->setScore( last_score );
	    mFragments->push_back( new_ali );
	    length = 0;
	  }

	  current_ali->Clear();
	  // start new alignment
	  if (p->mScore > 0) {
	    last_p = p;
	    last_score = p->mScore;
	    current_ali->addPair( new RESIDUEPAIR( *p ));
	    length = 1;
	  } else {
	    last_p = NULL;
	    last_score = 0;
	    length = 0;
	  }	    
	} else {
	  // continue existing alignment
	  last_score = new_score;
	  last_p = p;
	  current_ali->addPair( new RESIDUEPAIR( *p ));
	  length++;
	}
    }

    if (last_score > 0 && length > 1) {
	Alignata * new_ali = sample->getNew();
	copyAlignata( new_ali, current_ali);
	new_ali->setScore( last_score );
	mFragments->push_back( new_ali );
    }
    
}
  

} // namespace AlignLib



