/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#if HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef IMPL_MULTIPLE_ALIGNMENT_H
#define IMPL_MULTIPLE_ALIGNMENT_H 1

#include <iosfwd>
#include <string>
#include <list>

#include "alignlib.h" 
#include "MultipleAlignment.h"

namespace AlignLib {

/** 
    Multiple alignments are collection of aligned sequences (more specifically: objects of type
    @ref Alignata). A multiple alignment
    receives submitted aligned objects, or unaligned objects together with a multiple
    alignment, etc. Ownership belongs to the multiple alignment.

    The multiple alignment is a sort of container. Output-formatting, etc. is done
    by the @ref Alignata-objects.
    
    The interface for this class is quite fat, because multiple alignments are used
    in a variety of contexts.

    This implementation is basically complete.
    
    @author Andreas Heger
    @version $Id: ImplMultipleAlignment.h,v 1.3 2003/04/04 08:33:41 aheger Exp $
    @short an implementation for multiple alignments
*/

class Alignandum;
class Alignatum;
class Alignata;	
class Renderer;

class ImplMultipleAlignment : public MultipleAlignment {

  // class member functions
 public:

    // constructors and desctructors
    /** empty constructor 
     */
    ImplMultipleAlignment  ();

    /** copy constructor */
    ImplMultipleAlignment  (const ImplMultipleAlignment &);

    /** destructor */
    virtual ~ImplMultipleAlignment ();

    //---------------------------------------------------------------------------------------
    /*------- accessors --------------------------------------------------------------------*/
    /** returns the length of the multiple alignment. All objects in a multiple alignment have
	the same length */
    virtual ALIGNANDUM_POSITION getLength() const;

    /** sets the length of the multiple alignment. Raises an exception, if the mali is not empty
	*/
    virtual void setLength( ALIGNANDUM_POSITION length);

    /** returns the width of the multiple alignment, i.e., the number of objects in this multiple
	alignment */
    virtual int getWidth() const;
    
    /** returns a const reference to the object at row in the multiple alignment. This allows treating
	the multiple alignment as a two-dimensional matrix. Since string does define operator[] as well, you
	can access the symbol in column c and row r by calling 
	symbol =  multiple_alignment[r][c]
    */
    virtual const std::string & operator[]( int row ) const;		

    /** returns a pointer to the object at row from the multiple alignment. The pointer is not const,
	so you are free to do all sort of ugly stuff.*/
    virtual Alignatum * getRow( int row ) const;
    
    /** erases an entry form the multiple alignment */
    virtual void eraseRow( int row );

    /* ------------------ mutators ----------------------------------------------------------- */

    /*------------------- functions for adding new members to the multiple alignment---------*/

    /** add an aligned object to the multiple alignment. Ownership of this object is transferred to
	the multiple alignment. 
	@param src	       pointer to the aligned object to be added.
	@param alignment pointer to the alignment used for combining these two objects. If it is
		       not supplied, then it is assumed, that it is the identity alignment. In
		       that case src has to have the same length the multiple alignment. Note, the
		       multiple alignment is in col, the src is in row of the multiple alignment, so
		       when calling the member-function Alignata::Map() with a residue from
		       src, you get the correct position in the multiple alignment.
	@param skip_gaps false, if gaps in existing multiple alignment shall be introduced
    */
    virtual void Add( Alignatum * src,
		      const Alignata * alignment = NULL,
		      bool mali_is_in_row = true,
		      bool insert_gaps_mali = true,
		      bool insert_gaps_alignatum= true,
		      bool use_end_mali = false,
		      bool use_end_alignatum = false);

    /** add the contents of a multiple alignment to the multiple alignment by mapping it through an alignment
	@param src	       pointer to the unaligned object to be added.
	@param alignment pointer to the alignment used for combining these two objects. If it is
		       not supplied, then it is assumed, that it is the identity alignment. In
		       that case src has to have the same length the multiple alignment. Note, the
		       multiple alignment is in col, the src is in row of the multiple alignment, so
		       when calling the member-function Alignata::Map() with a residue from
		       src, you get the correct position in the multiple alignment.
          @param skip_gaps false, if gaps in existing multiple alignment shall be introduced
			   
     */
    virtual void Add( const MultipleAlignment * src,
		      const Alignata * alignment = NULL,
		      bool mali_is_in_row = true,
		      bool insert_gaps_mali = true,
		      bool insert_gaps_alignatum= true,
		      bool use_end_mali = false,
		      bool use_end_alignatum = false);
    
    /** returns the consensus string for the multiple alignment */
    virtual std::string getConsensusString () const ;			

    /** register a new renderer */
    virtual void registerRenderer( const Renderer * renderer );

    /** returns true, if there are no aligned objects in this alignment */
    virtual bool isEmpty() const;

    /** clears the multiple alignment */
    virtual void Clear();

    /** returns a clone of this object */
    virtual MultipleAlignment * getClone() const;
    
    /** returns an empty version of this object */
    virtual MultipleAlignment * getNew() const;

    /** Write the multiple alignment to a stream
     */
    virtual void Write( std::ostream & output, 
			ALIGNANDUM_POSITION segment_from = 0, 
			ALIGNANDUM_POSITION segment_to = 0) const ;

    /** Read a multiple alignment form a stream. I am not quite sure
	how this will work, because there might be different @ref Alignata
	objects in there.
     */
    virtual void Read( std::istream & );
    
 protected:
    /** free all memory. Tell all stored objects to destruct themselves */
    void freeMemory();

 private:
    /** the length of the multiple alignment */
    mutable int mLength;                       
    
    /** I store an array of vectors. The pointers can not be const, because sequences are told to rescale. */
    std::vector<Alignatum *> mRows;             

    /** the Renderer */
    const Renderer * mRenderer;
    
};


}

#endif /* IMPL_MULTIPLE_ALIGNMENT_H */

