/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream>
#include <iomanip>
#include <stdio.h>

#include "alignlib.h"
#include "AlignException.h"
#include "ImplProfile.h"
#include "Translator.h"
#include "HelpersTranslator.h"
#include "LogOddor.h"
#include "Regularizor.h"
#include "MultipleAlignment.h"
#include "Alignata.h"
#include "AlignataIterator.h"
#include "HelpersAlignandum.h"

/** default objects */

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

//---------------------------------------> constructors and destructors <--------------------------------------
ImplProfile::ImplProfile( const Regularizor * regularizor, 
			  const LogOddor * logoddor ) : 
  mRegularizor( regularizor ),
  mLogOddor( logoddor ),
  mCounts(NULL), mFrequencies(NULL), mProfile(NULL) {
}
  
const AlignandumDataProfile & ImplProfile::getData() const {
  mData.mCountsPointer		= &mCounts[getFrom() - 1];
  mData.mFrequenciesPointer	= &mFrequencies[getFrom() -1 ];
  mData.mProfilePointer		= &mProfile[getFrom() - 1];
  return mData;
}

//---------------------------------------------------------------------------------------------------------------
ImplProfile::ImplProfile(const ImplProfile & src ) : ImplAlignandum( src ), 
    mRegularizor( src.mRegularizor ),
    mLogOddor( src.mLogOddor),  
    mCounts(NULL), 
    mFrequencies(NULL), 
    mProfile(NULL) {

#ifdef DEBUG
    if (verbose > LL3) cout << "ImplProfile::ImplProfile( const ImplProfile &) " << endl;
#endif

    delete [] mCounts;
    delete [] mProfile;
    delete [] mFrequencies;
  
    // the first column is empty, so copy one more column
    ALIGNANDUM_POSITION copy_length = getTrueLength() + 1; 
  
    if (src.mCounts) {
	allocateCounts();
	memcpy( mCounts, src.mCounts, sizeof(TYPE_COUNTS_COLUMN) * copy_length);
    }
    if (src.mFrequencies) {
	allocateFrequencies();
	memcpy( mFrequencies, src.mFrequencies, sizeof(TYPE_FREQUENCIES_COLUMN) * copy_length);
    }
    
    if (src.mProfile) {
	allocateProfile();
	memcpy( mProfile, src.mProfile, sizeof(TYPE_PROFILE_COLUMN) * copy_length);
    }

}

//---------------------------------------------------------------------------------------------------------------
ImplProfile::~ImplProfile() {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplProfile::~ImplProfile()" << endl;
#endif

  // deleting NULL-pointer is save, or so I have heard.
  delete [] mCounts;	  mCounts = NULL;
  delete [] mFrequencies; mFrequencies = NULL;
  delete [] mProfile;	  mProfile = NULL;
}

//--------------------------------------------------------------------------------------
ImplProfile * ImplProfile::getClone() const {
    return new ImplProfile( *this );
};


//---------------------------------------------------------------------------------------------------------------
void ImplProfile::allocateCounts() {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplProfile::allocateCounts()" << endl;
#endif

  delete [] mCounts;
  mCounts = NULL;
    
  ALIGNANDUM_POSITION length = getTrueLength() + 1;

  mCounts = new TYPE_COUNTS_COLUMN[length];
  
  int i, j;
  for (i = 0; i < length; i++ )
    for (j = 0; j < PROFILEWIDTH; j++)
      mCounts[i][j] = 0;
    
}              

//---------------------------------------------------------------------------------------------------------------
void ImplProfile::allocateProfile() const {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplProfile::allocateProfile()" << endl;
#endif
    delete [] mProfile;

    ALIGNANDUM_POSITION length = getTrueLength() + 1;

    mProfile = new TYPE_PROFILE_COLUMN[length];
    int i, j;
    for (i = 0; i < length; i++ )
	for (j = 0; j < PROFILEWIDTH; j++)
	    mProfile[i][j] = 0;

}

//---------------------------------------------------------------------------------------------------------------
void ImplProfile::allocateFrequencies() const {
#ifdef DEBUG
    if (verbose > LL3) cout << "ImplProfile::allocateFrequencies()" << endl;
#endif

    delete [] mFrequencies;
    
    ALIGNANDUM_POSITION length = getTrueLength() + 1;

    mFrequencies = new TYPE_FREQUENCIES_COLUMN[length];

    int i, j;
    for (i = 0; i < length; i++ )
      for (j = 0; j < PROFILEWIDTH; j++)
	mFrequencies[i][j] = 0;
}

//---------------------------------------------------------------------------------------------------------------
RESIDUE ImplProfile::getMaximumCounts( const ALIGNANDUM_POSITION column ) const {
    int i;

    TYPE_COUNTS max   = 0;
    RESIDUE   max_i   = getDefaultTranslator()->getGapCode();

    for (i = 0; i < PROFILEWIDTH; i++) {
	if (mCounts[column][i] > max) {
	    max   = mCounts[column][i];
	    max_i = i;
	}
    }  
    return max_i;
}

//---------------------------------------------------------------------------------------------------------------
RESIDUE ImplProfile::getMaximumFrequencies( const ALIGNANDUM_POSITION column ) const {
    int i;

    TYPE_FREQUENCIES max   = 0;
    RESIDUE   max_i   = getDefaultTranslator()->getGapCode();

    for (i = 0; i < PROFILEWIDTH; i++) {
	if (mFrequencies[column][i] > max) {
	    max   = mFrequencies[column][i];
	    max_i = i;
	}
    }  
    return max_i;
}

//---------------------------------------------------------------------------------------------------------------
RESIDUE ImplProfile::getMaximumProfile( const ALIGNANDUM_POSITION column ) const {
    int i;

    TYPE_PROFILE max   = 0;
    RESIDUE   max_i   = getDefaultTranslator()->getGapCode();

    for (i = 0; i < PROFILEWIDTH; i++) {
	if (mProfile[column][i] > max) {
	  max   = mProfile[column][i];
	  max_i = i;
	}
    }  
    return max_i;
}

//---------------------------------------------------------------------------------------------------------------
void ImplProfile::Mask( ALIGNANDUM_POSITION column) {
    int i;

    if (mCounts) 
      for (i = 0; i < PROFILEWIDTH; i++) 
	mCounts[column][i] = 0;
    
    if (mFrequencies)
      for (i = 0; i < PROFILEWIDTH; i++) 
	mFrequencies[column][i] = 0;

    if (mProfile) 
      for (i = 0; i < PROFILEWIDTH; i++) 
	mProfile[column][i] = MASK_VALUE;

}
    
//---------------------------------------------------------------------------------------------------------------
RESIDUE ImplProfile::asResidue( const ALIGNANDUM_POSITION column ) const {
    
  if (mCounts) 
    return getMaximumCounts( getOffset(column) );
  
  return getDefaultTranslator()->getGapCode();
}

//--------------------------------------------------------------------------------------
void ImplProfile::Prepare() const {
#ifdef DEBUG
  if (verbose > LL3) cout << "void ImplProfile::Prepare() const" << endl;
#endif
    // do nothing, when a profile and frequencies already exist.
  if (!mFrequencies) {
    allocateFrequencies();
    mRegularizor->fillFrequencies( mFrequencies, mCounts, getTrueLength() );
  }

  if (!mProfile) {
    allocateProfile();
    mLogOddor->fillProfile( mProfile, mFrequencies, getTrueLength() );
  }
  setPrepared( true );
}



//--------------------------------------------------------------------------------------
void ImplProfile::Release() const {
  delete [] mFrequencies;
  mFrequencies = NULL;
  delete [] mProfile;
  mProfile = NULL;
  setPrepared(false);
}

//--------------------------------------------------------------------------------------
void ImplProfile::Shuffle( unsigned int num_iterations,
			   ALIGNANDUM_POSITION window_size) {

  if (window_size == 0)
    window_size = getLength();
  
  ALIGNANDUM_POSITION first_from = getFrom();
  TYPE_COUNTS buffer[PROFILEWIDTH];
  
  for (unsigned x = 0; x < num_iterations; x++) { 

    ALIGNANDUM_POSITION i,j;
    ALIGNANDUM_POSITION to = getTo();

    while (to > first_from ) {
      ALIGNANDUM_POSITION from = to - window_size;

      if (from < 1) {
	from = 1;
	window_size = to;
      }

      for (i = to; i > from; i--) {
	j = to - GetRandomPosition(window_size) + 1;
	memcpy(buffer, mCounts[i], PROFILEWIDTH * sizeof(mCounts));
	memcpy(mCounts[i], mCounts[j], PROFILEWIDTH * sizeof(mCounts));
	memcpy(mCounts[j], buffer, PROFILEWIDTH * sizeof(mCounts));
      }

      to -= window_size;
    }
  }


  /*
  ALIGNANDUM_POSITION i,j;
  ALIGNANDUM_POSITION from = getFrom();
  ALIGNANDUM_POSITION from_1 = from - 1;
  ALIGNANDUM_POSITION to = getTo();
  ALIGNANDUM_POSITION length = getLength();

  TYPE_COUNTS buffer[PROFILEWIDTH];

  for (i = to; i >= from; i--) {
    j = GetRandomPosition(length) + from_1;
    memcpy(buffer, mCounts[i], PROFILEWIDTH * sizeof(mCounts));
    memcpy(mCounts[i], mCounts[j], PROFILEWIDTH * sizeof(mCounts));
    memcpy(mCounts[j], buffer, PROFILEWIDTH * sizeof(mCounts));
  }
  */
  this->Release();

}
 

//--------------------------------------> I/O <------------------------------------------------------------
void ImplProfile::Write( std::ostream & output ) const {
    
    ALIGNANDUM_POSITION i;
    ALIGNANDUM_POSITION length = getOffset(getLength());
    int j;

    output.setf( ios::fixed );

    if (mCounts) {
	output << "----------->counts<----------------------------------------" << endl;
	for (i = getOffset(1); i <= length; i++) {
	    output << setw(2) << i << "\t";
	    for (j = 0; j < PROFILEWIDTH; j++) 
		output << setw(6) << setprecision(2) << mCounts[i][j];
	    output << endl;
	  
	}
    } else {
	output << "----------->no counts available<---------------------------" << endl;
    }

    if (mFrequencies) {
	output << "----------->frequencies<-----------------------------------" << endl;
	for (i = getOffset(1); i <= length; i++) {
	    output << setw(2) << i << "\t";
	    for (j = 0; j < PROFILEWIDTH; j++) 
		output << setw(6) << setprecision(2) << mFrequencies[i][j];
	    output << endl;
	}
    } else {
	output << "----------->no frequencies available<----------------------" << endl;
    }
    
    if (mProfile) {
      output << "----------->profile<---------------------------------------" << endl;
	for (i = getOffset(1); i <= length; i++) {
	    output << setw(2) << i << "\t";
	    for (j = 0; j < PROFILEWIDTH; j++) 
	      output << setw(6) << setprecision(2) << mProfile[i][j];
	    output << endl;
	}
    } else {
	output << "----------->no profile available<--------------------------" << endl;
    }
	
}

//--------------------------------------> I/O <------------------------------------------------------------
void ImplProfile::Read( std::istream & input ) {
    //!! to implement
}

} // namespace AlignLib
