/*
  alignlib - a library for aligning protein sequences

  $Id$

  Copyright (C) 2004 Andreas Heger
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream>
#include <fstream>
#include <string>

#include "HelpersAlignandum.h"
#include "HelpersSequence.h"
#include "ImplSequence.h" 
#include "AlignException.h"
#include "Translator.h"
#include "HelpersTranslator.h"

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif

using namespace std;

namespace AlignLib {

  //---------------------------------< implementation of factory functions >--------------

  //----------------------------------------------------------------------------------
  /** create a sequence from a NULL-terminated string */
  Alignandum * makeSequence( const char * sequence ) {
    return new ImplSequence( sequence );
  }

  //----------------------------------------------------------------------------------
  /** create a sequence from a string */
  Alignandum * makeSequence( const std::string & sequence ) {
    return new ImplSequence( sequence.c_str() );
  }

//--------------------------------------------------------------------------------------
ImplSequence::ImplSequence() : mSequence(NULL) {
}

//--------------------------------------------------------------------------------------
ImplSequence::ImplSequence( const char * src ) : ImplAlignandum(), mSequence(NULL) {
    ALIGNANDUM_POSITION length = strlen( src );

    //!! check for correct translation?
    setTrueLength( length );
    useFullLength();
    mSequence = getDefaultTranslator()->Encode( src, length );
    setPrepared(true );
}

//--------------------------------------------------------------------------------------
ImplSequence::ImplSequence( const ImplSequence & src ) : ImplAlignandum( src ), mSequence(NULL) {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplSequence::ImplSequence( const ImplSequence & )" << endl;
#endif

  if (mSequence != NULL) delete [] mSequence;
  //!! make exception safe
  mSequence = new RESIDUE[src.getTrueLength() + 1];
  memcpy( mSequence, src.mSequence, src.getTrueLength() + 1);
}


//--------------------------------------------------------------------------------------
ImplSequence::~ImplSequence() {
#ifdef DEBUG
  if (verbose > LL3) cout << "ImplSequence::~ImplSequence() " << endl;
#endif
  if (mSequence != NULL) 
    delete [] mSequence;
}

//--------------------------------------------------------------------------------------
Alignandum * ImplSequence::getClone() const {
  return new ImplSequence( *this );
};


//--------------------------------------------------------------------------------------
RESIDUE	ImplSequence::asResidue(ALIGNANDUM_POSITION n) const { 
  return mSequence[getOffset(n)]; 
}

//--------------------------------------------------------------------------------------
const AlignandumDataSequence & ImplSequence::getData() const {
    mData.mSequencePointer = &mSequence[getFrom() -1 ];
    return mData;
}

//--------------------------------------------------------------------------------------
void ImplSequence::Prepare() const {
}

//--------------------------------------------------------------------------------------
void ImplSequence::Release() const {
}

//--------------------------------------------------------------------------------------
void ImplSequence::Mask( ALIGNANDUM_POSITION x) {
  mSequence[ x ] = getDefaultTranslator()->getMaskCode();
}


//--------------------------------------------------------------------------------------
void ImplSequence::Shuffle( unsigned int num_iterations, ALIGNANDUM_POSITION window_size ) {

  if (window_size == 0)
    window_size = getLength();
  
  ALIGNANDUM_POSITION first_from = getFrom();
  
  for (unsigned x = 0; x < num_iterations; x++) { 

    ALIGNANDUM_POSITION i,j;
    ALIGNANDUM_POSITION to = getTo();
  
    while (to > first_from ) {
      ALIGNANDUM_POSITION from = to - window_size;

      if (from < 1) {
	from = 1;
	window_size = to;
      }

      for (i = to; i > from; i--) {
	j = to - GetRandomPosition(window_size) + 1;
	RESIDUE x = mSequence[j];
	mSequence[j] = mSequence[i];
	mSequence[i] = x;
      }

      to -= window_size;
    }
  }
}
 

//--------------------------------------------------------------------------------------
void ImplSequence::Write( std::ostream & output ) const {
  const char * result = getDefaultTranslator()->Decode( mSequence, getTrueLength() );
  output << result;
  delete [] result;
}
 
//--------------------------------------------------------------------------------------
void ImplSequence::Read( std::istream & input ) {
}



} // namespace AlignLib
